##PROCEDURE(help) DifferentialAlgebra0[Attributes]
##HALFLINE	returns the attributes of a regular differential chain
##AUTHOR        F. Boulier
##DATE          August 2009
##
##CALLINGSEQUENCE
##- `Attributes`('ideal')
##
##PARAMETERS
##- 'ideal' : a differential ideal or ring
##
##DESCRIPTION
##- The function call `Attributes`('ideal') returns the list of the 
##  attributes of 'ideal', if 'ideal' is a regular differential chain.
##  It returns a list of list of attributes if 'ideal' is a list.
##
##- The function call `Attributes`('R') returns the list of the
##  attributes of the zero ideal of 'R', if 'R' is a differential ring.
##  
##- The returned list involves some of the seven following
##  possible keywords, which provide some precision on the regular 
##  differential chains:
##  `differential`, `prime`, `primitive`, `squarefree`, `coherent`,
##  `autoreduced`, `normalized`. See "DifferentialAlgebra0" for
##  more details.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Attributes(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Attributes](...)`.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [u]);
##
##- The attributes of the (0) differential ideal. This ideal is not prime.
##> Attributes(R);
##< [differential, autoreduced, primitive, squarefree, coherent, normalized]
##
##- The non-differential ideal generated by u. It is prime.
##> S := DifferentialRing(derivations = [], blocks = [u]);
##
##> ideal := RosenfeldGroebner([u], S);
##
##> Attributes(ideal[1]);
##< [prime, autoreduced, primitive, squarefree, normalized]
##
##SEEALSO
##- "DifferentialAlgebra0", "PretendRegularDifferentialChain", "RosenfeldGroebner", "Pardi"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "PretendRegularDifferentialChain" : Help:DifferentialAlgebra0[PretendRegularDifferentialChain]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "Pardi" : Help:DifferentialAlgebra0[Pardi]

# EXPORTED

Attributes := proc (drideal)
    local result;
    if not IsIdealDRing (drideal) then
        error _ERRIDRNG_, "first", convert (drideal, ':-string')
    end if;
    try
        if IsDifferentialRing (drideal) then
	    result := AttributesSafe 
				(PretendRegularDifferentialChain ([], drideal))
	elif type (drideal, 'list') then
            result := map (AttributesSafe, drideal)
        else
            result := AttributesSafe (drideal);
        end if
    catch:
        error
    end try;
    return result;
end proc:

# Subfunction of Attributes

AttributesSafe := proc (a_table)
    local arguments, result;
    arguments := "Attributes" (a_table);
    arguments := arguments, ToBMIOptionSequence
                    (':-_JET_', ':-_JET_', 0, 0, ':-_SMALL_');
    result := BMIBladEvalSafe ([], arguments);
    return result;
end proc:

