##PROCEDURE(help) DifferentialAlgebra0[BelongsTo]
##HALFLINE        decides membership in differential ideals
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- BelongsTo ('p', 'ideal', 'opts')
##- BelongsTo ('L', 'ideal', 'opts')
##
##PARAMETERS
##
##- 'p' : a differential polynomial
##- 'L' : a list or a set of differential polynomials
##- 'ideal' : a differential ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `BelongsTo` ('p','ideal') returns `true`
##  if the differential polynomial 'p' belongs to the differential
##  ideal defined by 'ideal', else
##  it returns `false`. The differential polynomial 'p' is regarded
##  as a differential polynomial of the embedding differential
##  polynomial ring of 'ideal'.
##
##- If 'ideal' is a list of regular differential chains, the function
##  returns `true` if and only if 'p' belongs to each component of
##  the list.
##
##- The function call `BelongsTo` ('L', 'ideal') returns a list
##  or a set of ::boolean::.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `BelongsTo(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[BelongsTo](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'ideal' is used (see "Notation").
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for input. This option is overridden by the 
##  option above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with (DifferentialAlgebra0):
##> R := DifferentialRing (derivations = [t], blocks = [u]);
##
##- Every differential polynomial belongs to the unit differential ideal
##
##> BelongsTo (u[t], []);
##< true
##
##> ideal := RosenfeldGroebner ([u[t]^2-4*u], R);
##
##- The two first differential polynomials do not belong to 'ideal' but
##  their product does.
##
##> BelongsTo ([u[t], u[t,t]-2, u[t]*(u[t,t] - 2)], ideal);
##< [false, false, true]
##
##SEEALSO
##- "DifferentialAlgebra0", "NormalForm", "ReducedForm", "DifferentialPrem"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]
##- "ReducedForm" : Help:DifferentialAlgebra0[ReducedForm]
##- "DifferentialPrem" : Help:DifferentialAlgebra0[DifferentialPrem]

# EXPORTED

BelongsTo := proc 
    (eqns::{algebraic,equation,list,set},
     ideal,
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

     local result, leqns, innot;

     if not IsIdeal (ideal) then
	error _ERRIDEAL_, "second", convert (ideal, ':-string')
     elif ideal = [] then
# Everything belongs to the unit ideal
# Cannot check the syntax because we do not have the ring
	if type (eqns, '{:-list,:-set}') then
	    result := map (x -> true, eqns);
	else
	    result := true;
	end if
     else
# The general case
	try
            leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "first", convert (eqns, ':-string')
	end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (ideal));
	try
            if type (ideal, ':-table') then
                result := BelongsToSafe (leqns, ideal, innot, 
					 _TIMEOUT_, _MEMOUT_, _CELLSZ_)
	    else
		result := map (AndBelongsToSafe, leqns, ideal, innot,
					_TIMEOUT_, _MEMOUT_, _CELLSZ_)
	    end if
	catch:
	    error
	end try;
	result := FromExpandedRatfracList (result, eqns)
    end if;
    return result;
end proc:

# Local to BelongsTo
# eqns is a list of differential polynomials
# Returns a list of booleans

BelongsToSafe := proc (eqns, regchain, innot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "BelongsTo" 
	(ListOrTableToBMI (regchain [':-_FREEZESEQ_'], eqns), regchain);
    arguments := arguments,
	ToBMIOptionSequence (innot, ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe ([], arguments);
end proc:

# Local to BelongsTo
# Returns true if oneq belongs to each component of ideal
# oneq is one differential polynomial
# ideal is non empty list

AndBelongsToSafe := proc (oneq, ideal, innot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local i, result;
    for i from 1 to nops (ideal) do
	result := BelongsToSafe 
		([oneq], ideal [i], innot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	if not result [1] then
	    return false;
	end if
    end do;
    return true;
end proc:
