##PROCEDURE(help) DifferentialAlgebra0[Coeffs]
##HALFLINE        extracts the coefficients of a rational differential fraction
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `Coeffs`('q','v','R','opts')
##
##PARAMETERS
##- 'q' : a rational differential fraction
##- 'v' : a variable
##- 'R' : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Coeffs`('q','v','R') decomposes 'q' as a sum
##  of coefficients multiplied by monomials. The subexpressions of 'q'
##  which only involve *derivatives* greater than or equal to 'v' appear 
##  among the monomials. The ones which only involve *derivatives* lower
##  than 'v' appear among the coefficients. The variable 'v' is regarded
##  as a *derivative* of 'R', or a *derivative* of its embedding polynomial
##  ring, if 'R' is an ideal.
##
##- The denominator of 'q' appears either in each coefficient, if all 
##  the *derivatives* it depends on are lower than 'v', or, in each monomial, 
##  if all the *derivatives* it depends on, are greater than or equal to 'v'.
##  The function does not apply if the denominator of 'q' involves 
##  *derivatives* lower than 'v' and *derivatives* greater than or equal to 'v'.
##
##- The function call `Coeffs`('q','v','R') returns a sequence of two
##  lists. The first list is the list of coefficients. The second one
##  is the list of monomials (in the same order). Monomials are sorted 
##  by decreasing rank of their numerators.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Coeffs(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Coeffs](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'q' and 'v'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [[d(v),d(u)],[v,u]]);
##
##- The denominator only depends on derivatives greater than or equal to 'u'.
##  Thus it appears among the monomials.
##
##>(quote) Coeffs(1/4*(u[x]*u[y]*u-u[x]*u[y])/u, u, R);
##< [1/4, -1/4], [u[x]*u[y], u[x]*u[y]/u]
##
##- All the derivatives are less than _u[x,x]_ 
##
##>(quote) Coeffs(1/4*(u[x]*u[y]*u-u[x]*u[y])/u, u[x,x], R);
##< [1/4*(u[x]*u[y]*u-u[x]*u[y])/u], [1]
##
##> nf := 1/4*(d(u)[y,y]*u[x]*u^2-d(u)[y,y]*u[x]*u-d(u)[y]*u[x]*u[y]*u+d(u)[y]*u[x]*u[y]-d(u)*u[x])/u^2;
##< 1/4*(d(u)[y,y]*u[x]*u^2-d(u)[y,y]*u[x]*u-d(u)[y]*u[x]*u[y]*u+d(u)[y]*u[x]*u[y]-d(u)*u[x])/u^2
##
##- A differential of the differential rational fraction is considered.
##  Its coefficients are extracted.
##
##>(quote) Coeffs(nf, d(u), R);
##< [1/4*(u[x]*u-u[x])/u, 1/4*(-u[x]*u[y]*u+u[x]*u[y])/u^2, -1/4*u[x]/u^2], [d(u)[y,y], d(u)[y], d(u)]
##
##SEEALSO
##- "DifferentialAlgebra0", "Indets"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Indets" : Help:DifferentialAlgebra0[Indets]


# EXPORTED

Coeffs := proc (ratfrac, v::algebraic, drideal,
    {collected::boolean:=false}, 
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local result, arguments, a_table, leqns, innot, outnot;

    if not IsIdealDRing (drideal) then
        error _ERRIDRNG_, "third", convert (drideal, ':-string')
    elif type (drideal, ':-list') then
        a_table := drideal [1]
    else
        a_table := drideal
    end if;
    try
        leqns := ToExpandedRatfracList (ratfrac)
    catch:
	error _ERRRATF_, "first", convert (ratfrac, ':-string')
    end try;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    arguments := "Coeffs" 
            (ConvertString (a_table [':-_FREEZESEQ_'], leqns[1]), 
	     ConvertString (a_table [':-_FREEZESEQ_'], v), 
	     a_table, collected
        );
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'],arguments);
    catch:
        error
    end try;
    return result;
end proc:
