##PROCEDURE(help) DifferentialAlgebra0[DeltaPolynomial]
##HALFLINE        returns a Delta-polynomial
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- DeltaPolynomial ('p', 'q', 'R','opts')
##
##PARAMETERS
##
##- 'p' : a differential polynomial
##- 'q' : a differential polynomial
##- 'R' : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `DeltaPolynomial` ('p', 'q', 'R') returns the
##  _Delta_-polynomial generated by 'p' and 'q', which are regarded
##  as differential polynomials of 'R', or, of its embedding ring, if 'R'
##  is an ideal. 
##  See "DifferentialAlgebra0" for the definition of _Delta_-polynomials.
##
##- The numeric coefficients of the returned _Delta_-polynomial are
##  normalized: their gcd is equal to _1_, and, the leading one
##  is positive.
##  It is required that the leading derivatives of 'p' and 'q' are
##  derivatives of some same dependent variable.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `DeltaPolynomial(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[DeltaPolynomial](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'q'.
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [u,v]);
##
##- The *triangular* case: the least common derivative of the two
##  leading derivatives is different from both of them.
##
##> DeltaPolynomial(u[x]-v, u[y], R);
##< v[y]
##
##- The *non-triangular* case: the leading derivative of the second
##  argument is a derivative of the leading derivative of the first one.
##
##> DeltaPolynomial(u[x]^2-4*u, u[x,x], R);
##< u[x]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]

# EXORTED

DeltaPolynomial := proc 
   (p0::{algebraic,equation},
    q0::{algebraic,equation},
    drideal,
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local a_table, arguments, result, innot, outnot, p, q;

    if not IsIdealDRing (drideal) then
	error _ERRIDRNG_, "third", convert (drideal, ':-string');
    elif type (drideal, ':-list') then
	a_table := drideal [1];
    else
	a_table := drideal;
    end if;
    try
	p := ToExpandedRatfracList (p0, ':-no_denom' = true) [1]
    catch:
	error _ERRPOLY_, "first", convert (p0, ':-string')
    end try;
    try
	q := ToExpandedRatfracList (q0, ':-no_denom' = true) [1]
    catch:
	error _ERRPOLY_, "second", convert (q0, ':-string')
    end try;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    try
	arguments := "DeltaPolynomial"
		(ConvertString (a_table [':-_FREEZESEQ_'], p), 
		 ConvertString (a_table [':-_FREEZESEQ_'], q), 
		 a_table);
	arguments := arguments,
		ToBMIOptionSequence 
		    (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'],arguments)
    catch:
	error
    end try;
    return result;
end proc:
