##PACKAGE(help)  DifferentialAlgebra0
##HALFLINE       Package dedicated to the simplification of systems of polynomial differential equations
##AUTHOR         F. Boulier
##DATE           September 2009
##
##DESCRIPTION
##
##- The `DifferentialAlgebra0` package is dedicated to the simplification
##  problem of systems of differential polynomials.
##  The underlying theory is the *differential algebra* of Joseph Fels Ritt
##  and Ellis Robert Kolchin [R50,K73] (see the "References" page).
##  The main functionality is provided by the "RosenfeldGroebner" function,
##  which implements an algorithm for representing the radical of
##  the differential ideal generated by an input polynomial system as
##  an intersection of differential ideals presented by regular differential
##  chains. The "Examples" page illustrates the main functionalities
##  of the package, over examples. 
##  More details are provided in the "Glossary" page.
##  Here is the list of the package functions.
##  
##TABLE(border="0")
##ROW "Attributes"          | "BelongsTo"           | "Coeffs"
##ROW "DeltaPolynomial"     | "DifferentialPrem"    | "DifferentialRing"
##ROW "Differentiate"       | "Display"             | "Equations"
##ROW "EssentialComponents" | "FactorDerivative"    | "FieldElement"
##ROW "Indets"              | "Inequations"         | "Initial"
##ROW "IsDifferentialRing"  | "IsOrthonomic"        | "IsReduced"
##ROW "IsRegularDifferentialChain" | "LeadingCoefficient" | "LeadingDerivative" 
##ROW "LeadingRank"         | "MaxRankElement"      | "MinRankElement"
##ROW "NormalForm"          | "Notation"            | "Parameters"
##ROW "Pardi"               | "PowerSeriesSolution" | "PowerSeriesSystem"
##ROW "PreparationEquation" | "PretendRegularDifferentialChain" | "Ranking"
##ROW "ReducedForm"         | "RewriteRules"        | "RosenfeldGroebner"
##ROW "Separant"            | "SortByRank"          | "Tail"
##ENDTABLE
##
##XREFMAP
##- "Examples" : Help:DifferentialAlgebra0[Examples]
##- "Glossary" : Help:DifferentialAlgebra0[Glossary]
##- "References" : Help:DifferentialAlgebra0[References]
##- "Attributes" : Help:DifferentialAlgebra0[Attributes]
##- "BelongsTo" : Help:DifferentialAlgebra0[BelongsTo]
##- "Coeffs" : Help:DifferentialAlgebra0[Coeffs]
##- "DeltaPolynomial" : Help:DifferentialAlgebra0[DeltaPolynomial]
##- "DifferentialPrem" : Help:DifferentialAlgebra0[DifferentialPrem]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]
##- "Differentiate" : Help:DifferentialAlgebra0[Differentiate]
##- "Display" : Help:DifferentialAlgebra0[Display]
##- "Equations" : Help:DifferentialAlgebra0[Equations]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]
##- "FactorDerivative" : Help:DifferentialAlgebra0[FactorDerivative]
##- "FieldElement" : Help:DifferentialAlgebra0[FieldElement]
##- "Indets" : Help:DifferentialAlgebra0[Indets]
##- "Inequations" : Help:DifferentialAlgebra0[Inequations]
##- "Initial" : Help:DifferentialAlgebra0[Initial]
##- "IsDifferentialRing" : Help:DifferentialAlgebra0[IsDifferentialRing]
##- "IsOrthonomic" : Help:DifferentialAlgebra0[IsOrthonomic]
##- "IsReduced" : Help:DifferentialAlgebra0[IsReduced]
##- "IsRegularDifferentialChain" : Help:DifferentialAlgebra0[IsRegularDifferentialChain]
##- "LeadingCoefficient" : Help:DifferentialAlgebra0[LeadingCoefficient]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "MaxRankElement" : Help:DifferentialAlgebra0[MaxRankElement]
##- "MinRankElement" : Help:DifferentialAlgebra0[MinRankElement]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Parameters" : Help:DifferentialAlgebra0[Parameters]
##- "Pardi" : Help:DifferentialAlgebra0[Pardi]
##- "PowerSeriesSolution" : Help:DifferentialAlgebra0[PowerSeriesSolution]
##- "PowerSeriesSystem" : Help:DifferentialAlgebra0[PowerSeriesSystem]
##- "PreparationEquation" : Help:DifferentialAlgebra0[PreparationEquation]
##- "PretendRegularDifferentialChain" : Help:DifferentialAlgebra0[PretendRegularDifferentialChain]
##- "Ranking" : Help:DifferentialAlgebra0[Ranking]
##- "ReducedForm" : Help:DifferentialAlgebra0[ReducedForm]
##- "RewriteRules" : Help:DifferentialAlgebra0[RewriteRules]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "Separant" : Help:DifferentialAlgebra0[Separant]
##- "SortByRank" : Help:DifferentialAlgebra0[SortByRank]
##- "Tail" : Help:DifferentialAlgebra0[Tail]

# DifferentialAlgebra0
# Author: Franois Boulier, University Lille I, LIFL.
# First version written in 2008.

# This is the main source file of the DifferentialAlgebra0 package.
#
# The _ERRsomething_ identifiers are macros defined in mesgerr.mm
# The other _IDENTIFIERS_ are keywords defined in indices.mm
#
# Beware to the fact that indices.mm is processed by some sed command
# in order to produce a header file for the C part of the library. 
# 
# Functions of the same "family" share the same interface.
# Any modification brought to one member should be brought to the other
# family members too.
#
# - All exported functions have their own file.mm
# - The local functions which are strictly related to one exported function
# are incorporated in the file.mm of the exported function.
# - The other local functions have their own file.mm

$include <DifferentialAlgebra0/src/indices.mm>
$include <DifferentialAlgebra0/src/mesgerr.mm>

DifferentialAlgebra0 := module ()
    description 
	"The DifferentialAlgebra0 package";
    option package;
    local  BMIBladEval, 
	   BMIBladEvalSafe,
	   AttributesSafe,
	   BaseField,
	   BaseFieldGeneratorsSafe,
	   BelongsToSafe,
	   AndBelongsToSafe,
	   ConvertString,
	   DifferentialPremSafe,
	   DifferentialPrem2Safe,
	   DifferentiateSafe,
	   DisplayNotation,
	   DisplayOrderedList,
	   DisplayParameters,
	   DisplayRanking,
	   DisplayRegchainSafe,
	   DisplayRingSafe,
	   EquationsCriterion,
	   EquationsSafe,
	   EquationsWithCriterionSafe,
	   EvaluateProcessExpansionPoint,
	   EvaluateProcessExpansionPointDiff,
	   EvaluateSafe,
	   ToPolynomialEquationsAndInequations,
	   FieldElementSafe,
	   FixNotation,
	   FromExpandedRatfracList,
	   FrozenSymbols,
	   LeadingCoefficientSafe,
	   IndetsSafe,
	   InitialsAndSeparants,
	   IsCompositeBlock,
	   IsConstantSafe,
	   IsIdeal,
	   IsIdealDRing,
	   IsOrthonomicSafe,
	   IsReducedSafe,
	   IsReduced2Safe,
	   LeadingDerivativeSafe,
	   LeadingRankSafe,
	   ListOrTableToBMI,
           ModuleLoad,
	   NormalFormExtToRegchain,
	   NotationSafe,
	   RankingSafe,
	   IntegrateSafe,
	   ParametersSafe,
	   PowerSeriesSolutionDerivatives,
	   PowerSeriesSolutionEval,
	   PowerSeriesSolutionFactor,
	   PowerSeriesSolutionSafe,
	   PowerSeriesSystemSafe,
	   PowerSeriesSystemSafe2,
	   RewriteRules,
	   RewriteRulesSafe,
	   RewriteRulesWithCriterionSafe,
	   SeparantSafe,
	   Simplify,
	   TailSafe,
	   ToBMIOptionSequence,
	   ToDRing,
	   ToExpandedRatfrac,
	   ToExpandedRatfracList,
	   ToRegchain, 
	   ToString;
    export Attributes,
	   BelongsTo,
	   Coeffs,
	   DeltaPolynomial,
	   DifferentialPrem,
	   DifferentialRing,
	   Differentiate,
	   Display,
	   Equations, 		# Family of Equations
	   EssentialComponents,
	   FactorDerivative,
	   FieldElement,
	   Indets,
	   Inequations,
	   Initial,		# Almost the family of Separant
	   IsConstant,
	   IsDifferentialRing,
	   IsOrthonomic,
	   IsReduced,
	   IsRegularDifferentialChain,
	   LeadingCoefficient,	# Family of Separant
	   LeadingDerivative,	# Family of LeadingDerivative
	   LeadingRank,		# Family of LeadingDerivative
	   MaxRankElement,	# Family of MaxRankElement
	   MinRankElement,	# Family of MaxRankElement
	   Notation,
	   NormalForm, 
	   Parameters,
	   Pardi,
	   PowerSeriesSolution,
	   PowerSeriesSystem,
	   PreparationEquation,
	   PretendRegularDifferentialChain,
	   Ranking, 
	   Integrate,
	   ReducedForm,
	   RosenfeldGroebner,
	   Separant,		# Family of Separant
	   SortByRank,
	   Tail;		# Family of Separant

$include <DifferentialAlgebra0/src/Attributes.mm>
$include <DifferentialAlgebra0/src/BaseField.mm>
$include <DifferentialAlgebra0/src/BelongsTo.mm>
$include <DifferentialAlgebra0/src/BMIBladEval.mm>
$include <DifferentialAlgebra0/src/Coeffs.mm>
$include <DifferentialAlgebra0/src/ConvertString.mm>
$include <DifferentialAlgebra0/src/DeltaPolynomial.mm>
$include <DifferentialAlgebra0/src/DifferentialPrem.mm>
$include <DifferentialAlgebra0/src/DifferentialRing.mm>
$include <DifferentialAlgebra0/src/Differentiate.mm>
$include <DifferentialAlgebra0/src/Display.mm>
$include <DifferentialAlgebra0/src/Equations.mm>
$include <DifferentialAlgebra0/src/EssentialComponents.mm>
$include <DifferentialAlgebra0/src/Evaluate.mm>
$include <DifferentialAlgebra0/src/ToPolynomialEquationsAndInequations.mm>
$include <DifferentialAlgebra0/src/FactorDerivative.mm>
$include <DifferentialAlgebra0/src/FieldElement.mm>
$include <DifferentialAlgebra0/src/FixNotation.mm>
$include <DifferentialAlgebra0/src/FromExpandedRatfracList.mm>
$include <DifferentialAlgebra0/src/FrozenSymbols.mm>
$include <DifferentialAlgebra0/src/Indets.mm>
$include <DifferentialAlgebra0/src/Inequations.mm>
$include <DifferentialAlgebra0/src/Initial.mm>
$include <DifferentialAlgebra0/src/IsDifferentialRing.mm>
$include <DifferentialAlgebra0/src/IsIdeal.mm>
$include <DifferentialAlgebra0/src/IsIdealDRing.mm>
$include <DifferentialAlgebra0/src/IsConstant.mm>
$include <DifferentialAlgebra0/src/IsOrthonomic.mm>
$include <DifferentialAlgebra0/src/IsReduced.mm>
$include <DifferentialAlgebra0/src/IsRegularDifferentialChain.mm>
$include <DifferentialAlgebra0/src/LeadingCoefficient.mm>
$include <DifferentialAlgebra0/src/LeadingDerivative.mm>
$include <DifferentialAlgebra0/src/LeadingRank.mm>
$include <DifferentialAlgebra0/src/ListOrTableToBMI.mm>
$include <DifferentialAlgebra0/src/MaxRankElement.mm>
$include <DifferentialAlgebra0/src/MinRankElement.mm>
$include <DifferentialAlgebra0/src/NormalForm.mm>
$include <DifferentialAlgebra0/src/Notation.mm>
$include <DifferentialAlgebra0/src/Parameters.mm>
$include <DifferentialAlgebra0/src/Pardi.mm>
$include <DifferentialAlgebra0/src/PowerSeriesSolution.mm>
$include <DifferentialAlgebra0/src/PowerSeriesSystem.mm>
$include <DifferentialAlgebra0/src/PreparationEquation.mm>
$include <DifferentialAlgebra0/src/PretendRegularDifferentialChain.mm>
$include <DifferentialAlgebra0/src/Ranking.mm>
$include <DifferentialAlgebra0/src/Integrate.mm>
$include <DifferentialAlgebra0/src/ReducedForm.mm>
$include <DifferentialAlgebra0/src/RewriteRules.mm>
$include <DifferentialAlgebra0/src/RosenfeldGroebner.mm>
$include <DifferentialAlgebra0/src/Separant.mm>
$include <DifferentialAlgebra0/src/Simplify.mm>
$include <DifferentialAlgebra0/src/SortByRank.mm>
$include <DifferentialAlgebra0/src/Tail.mm>
$include <DifferentialAlgebra0/src/ToBMIOptionSequence.mm>
$include <DifferentialAlgebra0/src/ToDRing.mm>
$include <DifferentialAlgebra0/src/ToExpandedRatfracList.mm>
$include <DifferentialAlgebra0/src/ToRegchain.mm>

end module:
#savelib('DifferentialAlgebra0'):
