##PROCEDURE(help) DifferentialAlgebra0[DifferentialPrem]
##HALFLINE        the Ritt reduction algorithm
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- DifferentialPrem ('p','regchain','opts')
##- DifferentialPrem ('p','redset','R','opts')
##
##PARAMETERS
##
##- 'p' : a differential polynomial
##- 'regchain' : a regular differential chain
##- 'redset' : a polynomial or a list or a set of differential polynomials
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `DifferentialPrem` ('p','regchain') returns a
##  sequence _h,r_ such that _h_ is a power product of initials
##  and separants of 'regchain', _r_ is a differential polynomial
##  fully reduced (see below) with respect to 'regchain', and,
##  _h*p = r_ modulo the differential ideal *generated* by the
##  regular differential chain.
##
##- The function call `DifferentialPrem` ('p','redset','R','opts') returns 
##  a sequence _h,r_ such that _h_ is a power product of initials
##  and separants of 'redset', _r_ is a differential polynomial
##  fully reduced with respect to each element of 'redset',
##  and, _h*p = r_ modulo the differential ideal generated by 'redset'.
##  The elements of 'redset' must depend on at least, one *derivative*
##  and have integer coefficients. All the differential polynomials are
##  regarded as elements of 'R', or, of its embedding ring, if 'R' is an ideal.
##
##- A differential polynomial 'p' belongs to the differential ideal
##  defined by 'regchain' if and only if, the function call 
##  `DifferentialPrem` ('p','regchain') returns a sequence _h,0_
##  whose second component is zero.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `DifferentialPrem(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[DifferentialPrem](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `reduction` = `full`, `partial` or `algebraic`.
##  Indicates the type of reduction to be performed.
##  Depending on the type of reduction, the differential polynomial _r_
##  satisfies the following properties:
##-- `algebraic`. For each leading rank _v^d_ of 'regchain', or, of 'redset',
##  _r_ has degree less than _d_ in _v_.
##-- `partial`. For each leading derivative _v_ of 'regchain', or, of 'redset',
##  no *proper* derivative of _v_ occurs in _r_.
##-- `full`. For each leading rank _v^d_ of 'regchain', or, of 'redset',
##  no *proper* derivative of _v_ occurs in _r_, and, moreover,
##  _r_ has degree less than _d_ in _v_.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'redset'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [[v,u]]);
##> syst := [u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]];
##< [u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]]
##> ideal := RosenfeldGroebner(syst, R)[1];
##
##> p := u[x,y];
##< u[x,y]
##
##>(quote) h, r := DifferentialPrem(p, ideal);
##< 2*u[y], 2*u[x]
##
##>(quote) h, r := DifferentialPrem(p, syst, R);
##< 2*u[x], 4*u[y]
##
##- The ratio _r_/_h_ is equivalent to _p_ modulo the differential
##  ideal defined by the regular differential chain, but, it is not 
##  the normal form of _p_ 
##
##>(quote) NormalForm(r/h, ideal), NormalForm (p, ideal);
##< 1/2*u[x]*u[y]/u, 1/2*u[x]*u[y]/u
##
##- Different modes of reduction are available
##
##>(quote) h, r := DifferentialPrem(u[x]^3, syst, R, reduction=partial);
##< 1, u[x]^3
##
##>(quote) h, r := DifferentialPrem(u[x,y] + u[x]^3, syst, reduction=algebraic, R);
##< 1, u[x,y]+4*u[x]*u
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative", "LeadingRank", "NormalForm", "IsReduced", "BelongsTo"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]
##- "IsReduced" : Help:DifferentialAlgebra0[IsReduced]
##- "BelongsTo" : Help:DifferentialAlgebra0[BelongsTo]

# EXPORTED

DifferentialPrem := proc
    (F::{algebraic,equation},
     {reduction::identical(:-_FULLY_,:-_PARTIALLY_,:-_ALGBRCALLY_):=:-_FULLY_},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local result, leqns, innot, outnot, a_table, redset;

    if _nrest = 1 then
	a_table := _rest[_nrest];
        if not IsRegularDifferentialChain (a_table) then
	    error _ERRREGC_, "second", convert (a_table, ':-string')
        else
	    try
	        leqns := ToExpandedRatfracList (F, ':-no_denom' = true)
	    catch:
                error _ERRPOLY_, "first", convert (F, ':-string')
            end try;
	    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
	    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
	    try
	        result := DifferentialPremSafe
				(leqns[1], reduction, a_table, innot,
                                 outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
            catch:
                error
            end try;
        end if
    elif _nrest <> 2 then
	error _ERRNARGS_, "Two or three", convert (_nrest+1, ':-string')
    else
	try
	    redset := ToExpandedRatfracList (_rest[1], ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "second", convert (redset, ':-string')
	end try;
	if not IsIdealDRing (_rest[_nrest]) then
	    error _ERRIDRNG_, "third", convert (_rest[_nrest], ':-string')
	elif type (_rest[_nrest], ':-list') then
	    a_table := _rest[_nrest] [1]
	else
	    a_table := _rest[_nrest]
	end if;
	try
	    leqns := ToExpandedRatfracList (F, ':-no_denom' = true)
	catch:
	    error _ERRPOLY_, "first", convert (F, ':-string')
	end try;
	innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
	outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
	try
	    result := DifferentialPrem2Safe
				(leqns[1], reduction, redset, a_table, innot,
				 outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
	catch:
	    error
	end try;
    end if;
    return result;
end proc:

# Local to DifferentialPrem

DifferentialPremSafe := proc (one_eq, reduction, regchain, innot, outnot,
					_TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "DifferentialPrem" 
			(ConvertString (regchain [':-_FREEZESEQ_'], one_eq), 
			 ConvertString (regchain [':-_FREEZESEQ_'], reduction), 
			 regchain);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
end proc:

# Local to DifferentialPrem

DifferentialPrem2Safe := proc (one_eq, reduction, redset, a_table, innot,
				outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "DifferentialPrem"
			(ConvertString (a_table [':-_FREEZESEQ_'], one_eq), 
			 ConvertString (a_table [':-_FREEZESEQ_'], reduction), 
			 ConvertString (a_table [':-_FREEZESEQ_'], redset), 
			 a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:
