##PROCEDURE(help) DifferentialAlgebra0[DifferentialRing]
##HALFLINE        constructs a differential polynomial ring
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `DifferentialRing` ('ders', 'blks', 'pars', 'nota')
##- `DifferentialRing` ('R', 'pars', 'nota')
##
##PARAMETERS
##- 'ders' : a description of the derivations
##- 'blks' : a description of dependent variables and the ranking
##- 'pars' (optional) : a description of the *parameters*
##- 'nota' (optional) : a description of the default notation
##- 'R' : a differential polynomial ring or ideal
##
##DESCRIPTION
##- The function call `DifferentialRing` ('ders', 'blks', 'pars', 'nota') 
##  constructs a differential polynomial ring and returns it. 
##-- The argument 'ders' has the form `derivations` = 'D' where 'D'
##   is a list of independent variables, and provides the derivations. 
##-- The arguments 'blks' has the form `blocks` = 'B' where 'B' is a 
##   list of *blocks*, and provides the set of the dependent variables
##- Together, the lists 'D' and 'B' provide the *ranking* of 'R'. 
##  For more detailed explanations on the syntax and the interpretation
##  of these two lists, see "Ranking" and "Display". 
##
##- The optional argument 'pars' has the form `parameters` = 'P' where 'P'
##  is a list of *parameters* (see "Parameters"). 
##  The *parameters* of the ring, which must also be listed in 'blks', are 
##  dependent variables which do not need to depend on all the independent 
##  variables. 
##  Most often, *parameters* are dependent
##  variables which do not depend on any independent variable and can thus
##  serve as parameters (in the usual sense) to design parametric differential 
##  systems. 
##
##- Independent and dependent variables can be denoted using quite
##  general expressions, which include ::name::, ::indexed:: and
##  unevaluated functions calls. The general syntax is that of
##  a VARIDENT (below), where dots-separated sequences are allowed to be empty.
##  The `DifferentialRing` function forbids, however, the possibly ambiguous
##  definitions that might occur because of the last form
##  of INDICES and the `jet` notation for *derivatives*. 
##
##TABLE(width="50%",border="0",colwidth="1|1|3")
##ROW VARIDENT | `::=` | ::name:: INDICES `...` INDICES
##ROW INDEXPR  | `::=` | VARIDENT
##ROW          | `::=` | ::integer:: INDICES `...` INDICES
##ROW INDICES  | `::=` | (INDEXPR, `...`, INDEXPR)
##ROW          | `::=` | [INDEXPR, `...`, INDEXPR]
##ENDTABLE
##
##- Independent and dependent variables which do not obey the above 
##  grammar are allowed also but get renamed before being processed.
##  The renamed form may appear in error messages.
##
##- The optional argument 'nota' has the form `notation` = 'keyword' where
##  'keyword' is `jet`, `tjet`, `diff` or `Diff`. It provides the default
##  notation of the ring. If not specified, the `jet` notation is chosen.
##  For more information on the different notations, see "Notation".
##
##- The function call `DifferentialRing` ('R', 'pars', 'nota') permits
##  to construct a new differential polynomial ring from an existing one, 'R',
##  or from the embedding ring of 'R' if 'R' is an ideal. The new
##  differential polynomial ring has the same ranking as 'R' but may have
##  a different set of *parameters* or a different default notation.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `DifferentialRing(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[DifferentialRing](...)`.
##
##SECTION(nohelp) Other options
##- Some further optional arguments are available.
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##
##- This example comes from chemistry. The list 'species' contains
##  four dependent variables, which denote the concentrations of 
##  four chemical species. The list 'params' contains three rate
##  constants of chemical reactions. These rate constants are actually
##  *parameters* of the differential ring.
##
##> species := [C,E,P,S]:
##> params := [k[1],k[-1],k[2]]:
##> R := DifferentialRing(derivations = [t], blocks = [F[1], species, params], parameters = params, notation = jet);
##
##- The considered system is a differential-algebraic system, arising from
##  some quasi-steady state approximation of a basic enzymatic reaction
##  system. The dependent variable _F[1]_ denotes the unknown contribution
##  of a fast revertible reaction.
##
##> syst := [F[1]+E[t]-C*k[2], F[1]+S[t], -F[1]+C[t]+C*k[2], P[t]-C*k[2], -C*k[-1]+E*S*k[1]];
##< [F[1]+E[t]-C*k[2], F[1]+S[t], -F[1]+C[t]+C*k[2], P[t]-C*k[2], -C*k[-1]+E*S*k[1]]
##
##- Still the same system, with the `diff` notation.
##
##> Equations(syst, R, output=diff);
##< [diff(E(t),t)+F[1](t)-k[2]*C(t), diff(S(t),t)+F[1](t), diff(C(t),t)-F[1](t)+k[2]*C(t), diff(P(t),t)-k[2]*C(t), k[1]*E(t)*S(t)-k[-1]*C(t)]
##
##- The derivatives of the parameters are zero.
##
##> Differentiate(syst, t, R);
##< [F[1][t]+E[t,t]-C[t]*k[2], F[1][t]+S[t,t], -F[1][t]+C[t,t]+C[t]*k[2], P[t,t]-C[t]*k[2], -C[t]*k[-1]+E[t]*S*k[1]+S[t]*E*k[1]]
##
##- The differential ring 'Rbar' has the same ranking as 'R' but the
##  rate constants are no more considered as parameters.
##
##> Rbar := DifferentialRing(R, parameters = []);
##
##- The derivatives of the rate constants are not zero.
##
##> Differentiate(syst, t, Rbar);
##< [F[1][t]+E[t,t]-C[t]*k[2]-C*k[2][t], F[1][t]+S[t,t], -F[1][t]+C[t,t]+C[t]*k[2]+C*k[2][t], P[t,t]-C[t]*k[2]-C*k[2][t], -C[t]*k[-1]+E[t]*S*k[1]+S[t]*E*k[1]-C*k[-1][t]+E*S*k[1][t]]
##
##SEEALSO
##- "DifferentialAlgebra0", "Ranking", "Parameters", "Notation"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Ranking" : Help:DifferentialAlgebra0[Ranking]
##- "Parameters" : Help:DifferentialAlgebra0[Parameters]

# EXPORTED
# Builds a Differential Polynomial Ring
# The call to BMI checks the syntax and returns the ordering.

DifferentialRing := proc 
       ({derivations::{list,:-_U_}:=':-_U_'},
        {blocks::{list,:-_U_}:=':-_U_'},
	{ranking::{list,:-_U_}:=':-_U_'},
        {parameters::{list,set,:-_U_}:=':-_U_'},
        {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_TIMEOUT_::nonnegative:=0},
        {_MEMOUT_::nonnegative:=0},
        {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})
    local ordering, arguments, prmtrs, a_table, ionot, 
		result, freezeseq, unnumberseq, actual_blocks;
    if _nrest = 0 then
# The basic form
        if derivations = ':-_U_' or 
				(blocks = ':-_U_' and ranking = ':-_U_') then
            error _ERRMDB_
	elif blocks <> ':-_U_' and ranking <> ':-_U_' then
	    error _ERRRKBL_
	elif blocks <> ':-_U_' then
	    actual_blocks := blocks
	else
	    actual_blocks := ranking
	end if;
        if parameters = ':-_U_' then
            prmtrs := []
        else
            prmtrs := convert (parameters, ':-list');
        end if;
	try
	    freezeseq, unnumberseq := 
				FrozenSymbols (derivations, actual_blocks);
	catch:
	    error
	end try;
        arguments := "DifferentialRing"
                    (ConvertString (freezeseq, derivations),
                     ConvertString (freezeseq, actual_blocks),
                     ConvertString (freezeseq, prmtrs));
	ionot := FixNotation (_IONOT_);
        arguments := arguments,
            ToBMIOptionSequence 
		(ionot, ionot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
#		(':-_JET_', ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        try
            ordering := BMIBladEvalSafe (unnumberseq, arguments);
        catch:
            error
        end try;
        result := ToDRing (ordering, _IONOT_, freezeseq, unnumberseq);
    elif _nrest = 1 then

# DifferentialRing (Dring)
# DifferentialRing (regchain)

        if not IsIdealDRing (_rest[1]) then
            error _ERRIDRNG_, "first", convert (_rest[1], ':-string')
        elif type (_rest[1], ':-list') then
            a_table := _rest[1,1];
        else
            a_table := _rest[1];
        end if;
        if parameters = ':-_U_' then
            prmtrs := ParametersSafe (a_table);
        else
            prmtrs := convert (parameters, ':-list');
        end if;
        ordering := RankingSafe (a_table);
	freezeseq := a_table [':-_FREEZESEQ_'];
	unnumberseq := a_table [':-_UNNUMBERSEQ_'];
        arguments := "DifferentialRing"
            (ConvertString (freezeseq, rhs (op (1, ordering))),
             ConvertString (freezeseq, rhs (op (2, ordering))),
             ConvertString (freezeseq, prmtrs));
        arguments := arguments,
            ToBMIOptionSequence 
		(':-_JET_', ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        try
            ordering := BMIBladEvalSafe (unnumberseq, arguments);
        catch:
            error
        end try;
        ionot := FixNotation (_IONOT_, NotationSafe (a_table));
        result := ToDRing (ordering, ionot, freezeseq, unnumberseq);
    else
        error _ERRNARGS_, "At most one", _nrest
    end if;
    return result;
end proc:

