##PROCEDURE(help)  DifferentialAlgebra0[Differentiate]
##HALFLINE         differentiates a differential rational fraction
##AUTHOR           F. Boulier
##DATE             July 2009
##
##CALLINGSEQUENCE
##- `Differentiate`('p', 'theta', 'R', 'opts')
##- `Differentiate`('L', 'theta', 'R', 'opts')
##- `Differentiate`('ideal', 'theta', 'opts')
##
##PARAMETERS
##- 'p'     : a differential rational fraction
##- 'L'     : a list or a set of differential polynomials or rational fractions
##- 'theta' : a sequence of derivation operators
##- 'R'     : a differential polynomial ring or ideal
##- 'ideal' : a differential ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Differentiate`('p', 'theta', 'R') returns the
##  derivative of 'p' with respect to 'theta'.
##  The parameter 'p' is regarded as a differential polynomial or a 
##  differential rational fraction of 'R', or of its embedding ring
##  if 'R' is an ideal. 
##
##- The parameter 'theta' is a possibly empty sequence of *differential 
##  operators*. See "DifferentialAlgebra0" for more details. 
##
##- The function call `Differentiate`('L', 'theta', 'R') returns the
##  list or the set of the derivatives of the elements of 'L' with respect
##  to 'theta'.
##
##- If 'ideal' is a regular differential chain, the function call
##  `Differentiate`('ideal', 'theta') returns the list of the
##  derivatives of the chain elements. If 'ideal' is a list of regular 
##  differential chains, the function call returns a list of lists
##  of derivatives.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Differentiate(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Differentiate](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call
##   `Differentiate`('ideal','theta'), applies the function also over 
##   the differential polynomials which state that the derivatives of 
##   the parameters are zero. Default value is `false`.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##
##> R := DifferentialRing(derivations = [t], blocks = [u]);
##
##> Differentiate(u[t]^2-4*u, t, R);
##< 2*u[t,t]*u[t]-4*u[t]
##
##> Differentiate([u, 1/u], t^2, R);
##< [u[t,t],(-u[t,t]*u^2+2*u[t]^2*u)/u^4]
##
##- No differential operator is provided. The function acts as the identity.
##
##> Differentiate(u[t]^2-4*u, output=diff, R);
##< diff(u(t),t)^2-4*u(t)
##
##SEEALSO
##- "DifferentialAlgebra0", "FactorDerivative"
##
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "FactorDerivative" : Help:DifferentialAlgebra0[FactorDerivative]


# EXPORTED
#
# Differentiate (polynomial | list (polynomial), sequence of derivations, ring)
# Differentiate (ideal, sequence of derivations)

Differentiate := proc 
    ({fullset::boolean:=false},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local ideal, eqns, result, lders, a_table, innot, outnot;

    if _nrest = 0 then
	error _ERRNARGS_, "At least one", _nrest
    elif IsIdeal (_rest[1]) then
# The first parameter is an ideal
	ideal := _rest[1];
	if ideal = [] then
	    return [];
	end if;
	lders := [_rest [2..-1]];
	if not type (lders, ':-list(:-algebraic)') then
            error _ERRDERS_, convert (lders, ':-string');
	end if;
	innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (ideal));
	outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (ideal));
        try
            if type (ideal, ':-list') then
                result := map (DifferentiateSafe, ideal, lders,
                            ideal [1], fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := DifferentiateSafe (ideal, lders,
                            ideal, fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if;
	catch:
	    error
	end try
# else the last parameter is an ideal or a ring
    elif not IsIdealDRing (_rest[_nrest]) then
        error _ERRIDRNG_, "last", convert (_rest[_nrest], ':-string')
    else
        lders := [_rest[2.._nrest-1]];
        if not type (lders, ':-list(:-algebraic)') then
            error _ERRDERS_, convert (lders, ':-string');
        elif type (_rest[_nrest], ':-list') then
            a_table := _rest[_nrest,1];
        else
            a_table := _rest[_nrest];
        end if;
	try
            eqns := ToExpandedRatfracList (_rest[1])
	catch:
	    error _ERRRLRT_, "first", convert (_rest[1], ':-string')
	end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
	try
	    result := DifferentiateSafe (eqns, lders, a_table, fullset,
			innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	catch:
	    error
	end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    end if;
    return result;
end proc:

# Local to Differentiate
# Also called by NormalForm

DifferentiateSafe := proc (list_or_table, lders, a_table, fullset, 
			   innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "Differentiate"
		(ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table),
		 fullset, 
		 ConvertString (a_table [':-_FREEZESEQ_'], lders), 
		 a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:


