##PROCEDURE(help) DifferentialAlgebra0[Display]
##HALFLINE
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- Display ('R')
##
##PARAMETERS
##
##- 'R' : a differential polynomial ring, or, ideal
##
##DESCRIPTION
##
##- The function call `Display`('R') prints a text description of 'R'. 
##  In particular, the ranking of 'R' is detailed.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Display(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Display](...)`.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R :=  DifferentialRing(derivations = [x,y], blocks = [[v,u], lex[w]]);
##> Display(R);
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]], R);
##> Display(ideal[1]);
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing", "Ranking"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]
##- "Ranking" : Help:DifferentialAlgebra0[Ranking]

# EXPORTED

Display := proc
   (R::table,
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'},$)

    if not IsDifferentialRing (R) and not IsRegularDifferentialChain (R) then
        error _ERRRCRNG_, "first", convert (R, ':-string');
    end if;

    if IsDifferentialRing (R) then
        DisplayRingSafe (R);
    else
        DisplayRegchainSafe (R);
    end if;
    return NULL
end proc:

# Local to Display

DisplayRegchainSafe := proc (regchain)
    local rnks, att;

    printf ("%s\n", "Data structure: regular differential chain");
    rnks := LeadingRank (regchain, ':-fullset'=true, ':-_OUTPUT_'=':-_JET_');
    if rnks = [] then
	printf ("%s\n", "Differential ideal: (0)");
    else
	printf ("%s  %s\n", "Leading ranks:", DisplayOrderedList (rnks, ` < `));
    end if;
    att := Attributes (regchain);
    printf ("%s     %a\n", "Attributes:", att);
end proc:

# Local to Display

DisplayRanking := proc (R)
    local rkng, ders, kwblks, blks, deps;

    rkng := Ranking (R);
    ders := subs (op (rkng), ':-derivations');
    kwblks := subs (op (rkng), ':-blocks');
    blks := map (convert, kwblks, ':-list');
    deps := map (op, blks);

    if nops (ders) = 0 then
	if nops (deps) = 0 then
	    printf ("%s\n", "No variable.");
	else
	    printf ("%s %s\n", "Ordering on variables:", 
					DisplayOrderedList (deps, ` > `));
	end if;
	printf ("%s\n", "No derivation.");
    elif nops (ders) = 1 then
	if nops (deps) = 0 then
	    printf ("%s\n", "No dependent variable.");
	else
	    printf ("%s        %s\n", "Ranking:", 
					DisplayOrderedList (blks, ` >> `));
	end if;
	printf ("%s     %s\n", "Derivation:", ToString (ders[1]));
    else
	if nops (deps) = 0 then
	    printf ("%s\n", "No dependent variable.");
	else
	    printf ("%s        %s\n", "Ranking:", 
					DisplayOrderedList (kwblks, ` >> `));
	end if;
	printf ("%s    %s\n", "Derivations:", 
					DisplayOrderedList (ders, ` > `))
    end if;
end proc:

# Local to Display

DisplayNotation := proc (R)
    printf ("%s       %s\n", "Notation:", ToString (Notation (R)));
end proc:

# Local to Display

DisplayParameters := proc (R)
    printf ("%s     %s\n", "Parameters:", ToString (Parameters (R)));
end proc:

# Local to Display

DisplayRingSafe := proc (R)
    printf ("%s\n", "Data structure: differential ring");
    DisplayRanking (R);
    DisplayNotation (R);
    DisplayParameters (R);
end proc:

# Local to Display.
# Pretty-prints (?) the list L, by inserting operator between its elements.
# Operator is `>` or `>>`.

DisplayOrderedList := proc (L, operator)
    local i;
    return cat 
	(seq (op ([ToString (L[i]), ToString (operator)]), 
		i = 1 .. nops (L) - 1), 
	 ToString (L[-1]));
end proc:
