##PROCEDURE(help) DifferentialAlgebra0[Equations]
##HALFLINE        selects equations
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `Equations`('ideal','criterion','opts')
##- `Equations`('L','criterion','R','opts')
##
##PARAMETERS
##- 'ideal' : a differential polynomial ideal
##- 'criterion' (optional) : a selection criterion
##- 'L' : a differential polynomial, a list or a set of differential polynomials
##- 'R' : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Equations`('ideal') returns the list of the
##  differential polynomials which constitute the chain if 'ideal' is
##  a regular differential chain. If 'ideal' is a list of regular
##  differential chains, the function call returns a list of lists
##  of differential polynomials.
##
##- The function call `Equations`('ideal','criterion') returns only 
##  the differential polynomials which satisfy 'criterion'.
##
##- The function call `Equations`('L','criterion','R') returns the list
##  or the set of the differential polynomials of 'L' which satisfy
##  'criterion'. The elements of 'L' should be differential polynomials
##  of 'R', or of its embedding ring if 'R' is an ideal.
##  If 'criterion' is omitted, 'L' is returned.
##
##- The possible values of 'criterion' are listed below.
##
##-- `order`  'relop' 'n'
##  where 'relop' stands for one of the relational operators 
##  `>`, `>=`, `<`, `<=`, `=`, `<>` and 'n' 
##  is a nonnegative integer. 
##  In this case, the function call only returns differential
##  polynomials whose orders satisfy the criterion. 
##  Differential polynomials which do not depend on any *derivative* 
##  are not selected.
##
##-- `rank` 'relop' 'rk'
##  where 'relop' stands for one of the relational operators
##  `>`, `>=`, `<`, `<=`, `=`, `<>` 
##  and 'rk' is a rank.
##  The special ranks _0_ and _1_ are allowed.
##  In this case, the function call only returns differential
##  polynomials whose leading ranks satisfy the criterion. 
##
##-- `leader` 'relop' 'v'
##  where 'relop' stands for one of the relational operators 
##  `>`, `>=`, `<`, `<=`, `=`, `<>` and 'v' is a *derivative*. 
##  In this case, the function call 
##  only returns differential polynomials whose leading derivatives
##  satisfy the criterion. 
##  This form applies also to differential
##  polynomials which depend on independent variables only.
##
##-- `leader` 'relop' `derivative`('v') 
##  where 'relop' is either `=` or `<>` and 'v' is a *derivative*. 
##  In this case, the function call only returns differential polynomials 
##  whose leading derivatives are equal to (or different from) any derivative
##  of 'v'.
##  This form applies also to differential
##  polynomials which depend on independent variables only.
##
##-- `leader` 'relop' `proper`('v') 
##  where 'relop' is either `=` or `<>` and 'v' is a *derivative*.
##  In this case, the function call only returns differential polynomials
##  whose leading derivatives are equal to (or different from) any proper
##  derivative of 'v'.
##  This form applies also to differential
##  polynomials which depend on independent variables only.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Equations(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Equations](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call
##  `Equations`('ideal'), applies the function also over
##  the differential polynomials which state that the derivatives of
##  the parameters are zero. Default value is `false`.
##  This option is incompatible with the `diff` notation.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'L'.
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [[u,v],p], parameters = [p]);
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+p,v[x,x]-u[x]], R);
##
##- The two equations stating that the derivatives of _p_ are zero, are omitted.
##
##> Equations(ideal);
##< [[2*v[y]^2-u^2+2*p*u-p^2, u[y]^2-2*u, u[x]*u-u[x]*p-2*u[y]*v[y], v[x,x]*u-v[x,x]*p-2*u[y]*v[y]], [p, u, v[x,x]]]
##
##- The two equations stating that the derivatives of _p_ are zero, are present.
##
##> Equations(ideal[1], fullset=true);
##< [p[y], p[x], 2*v[y]^2-u^2+2*p*u-p^2, u[y]^2-2*u, u[x]*u-u[x]*p-2*u[y]*v[y], v[x,x]*u-v[x,x]*p-2*u[y]*v[y]]
##
##- Selects the equations of *order* greater than _1_ only.
##
##> Equations(ideal, order>1);
##< [[v[x,x]*u-v[x,x]*p-2*u[y]*v[y]], [v[x,x]]]
##
##- Selects the equations of *rank* less than or equal to the special
##  *rank* _1_, i.e., all the numeric differential polynomials.
##
##> Equations([0, 17, u[x]], rank<=1, R);
##< [0, 17]
##
##- Selects the equations whose leading derivatives are derivatives of _u_.
##
##> Equations(ideal[1], leader=derivative(u));
##< [u[y]^2-2*u, u[x]*u-u[x]*p-2*u[y]*v[y]]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative", "LeadingRank", "RewriteRules"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "RewriteRules" : Help:DifferentialAlgebra0[RewriteRules]

# EXPORTED
#
# Equations (regchain | intersection of regchains)
# Equations (regchain, criterion)
# Equations (list of equations, criterion, differential ring)
#
# The first form returns the list of equations of a regular chain.
# Applies to either a regular chain or a non-empty intersection of 
# regular chains. The result is either a list or a list of lists.
#
# If fullset then the lists involve also the equations which express
# that the derivatives of the parameters are zero
#
# The two other forms permit to select some elements in a list of equations.

Equations := proc 
       ({fullset::boolean:=false},
	{solved::boolean:=false},
        {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_TIMEOUT_::nonnegative:=0},
        {_MEMOUT_::nonnegative:=0},
        {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local result, ideal, a_table, keyword, relop, modifier, derv, eqns, 
                                            innot, outnot;
    if solved then
	return RewriteRules (args)
    end if;
    if _nrest = 1 then
# Equations (ideal)
# The result is a list
        if not IsIdeal (_rest[1]) then
            error _ERRIDEAL_, "first", convert (_rest[1], ':-string')
        end if;
        ideal := _rest[1];
        if ideal = [] then 
            return [];
        elif type (ideal,':-list') then
            a_table := ideal[1];
        else
            a_table := ideal;
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        if outnot = ':-_DIFF_' and fullset then
            error _ERRFSDF_
        end if;
        try
            if type (ideal, ':-list') then
                    result := map (EquationsSafe, ideal, fullset, outnot, 
                             _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := EquationsSafe (ideal, fullset, outnot,
                             _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if
        catch:
            error
        end try;
    elif _nrest = 2 and IsIdeal (_rest[1]) then
# Equations (ideal, criterion)
# The result is a list
        ideal := _rest[1];
        if ideal = [] then
            return [];
        elif type (ideal,':-list') then
            a_table := ideal[1];
        else
            a_table := ideal;
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
	if outnot = ':-_DIFF_' and fullset then
            error _ERRFSDF_
        end if;
        try
            keyword, relop, modifier, derv :=
                                    EquationsCriterion (_rest[2]);
        catch:
            error _ERRCRIT_, "second", convert (_rest[2], ':-string');
        end try;
        try
            if type (ideal, ':-list') then
                result := map (EquationsWithCriterionSafe, ideal,
                    keyword, relop, modifier, derv, a_table, fullset,
                    innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := EquationsWithCriterionSafe (ideal, keyword, 
                    relop, modifier, derv, a_table, fullset,
                    innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if
        catch:
            error
        end try;
    elif _nrest = 3 or _nrest = 2 then
# Equations (equations, criterion, dring)
# The result is a sequence
        if not IsIdealDRing (_rest[_nrest]) then
            error _ERRIDRNG_, "last", convert (_rest[_nrest], ':-string');
        elif type (_rest[_nrest],':-list') then
            a_table := _rest[_nrest,1];
        else
            a_table := _rest[_nrest];
        end if;
	try
            eqns := ToExpandedRatfracList (_rest[1], ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "first", convert (_rest[1], ':-string')
	end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        try
            if _nrest = 2 then
                keyword, relop, modifier, derv :=
                            ':-_RANK_', ':-_GE_', ':-_IDENTICAL_', 0
            else
                keyword, relop, modifier, derv :=
                                    EquationsCriterion (_rest[2]);
            end if
        catch:
            error _ERRCRIT_, "second", convert (_rest[2], ':-string');
        end try;
        try
            result := EquationsWithCriterionSafe 
                        (eqns, keyword, relop, modifier, derv, a_table, 
                            fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        catch:
            error
        end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    else
        error _ERRNARGS_, "From one to three", _nrest
    end if;
    return result;
end proc:

# Subfunction of EQUATIONS
#
# Process the criterion which has the form
#
# e.g. leader=v[y], order>3, leader=derivative(u), rank=x^2...
#
# more generally:
# * leader relop var (with relop: =, <>, <, <=, >=, >)
# * leader relop derivative(var) (with relop: =, <>)
# * leader relop proper(var)     (with relop: =, <>)
# * rank   relop v^d (with relop: =, <>, <, <=, >=, >)
# * order relop nonnegative (with relop: =, <, <=, >=, >)
#
# The reverted expression (switching lhs and rhs) is understood too.
#
# Return a sequence: keyword, relop, modifier, derv
#   keyword  = leader | rank | order
#   relop    = the relational operator
#   modifier = derivative | proper | identical
#   derv     = a derivative, a rank or an integer number
#
# Raise an error otherwise.

EquationsCriterion := proc (expr)
    local Lopernd, Lkw_opernd, Lkw_rev_opernd, gauche, droite, 
                            opernd, keyword, relop, modifier, derv, i;
    Lopernd := [`<`, `<=`, `=`, `<>`, `>=`, `>`];
    Lkw_opernd := 
            [':-_LT_', ':-_LE_', ':-_EQ_', ':-_NE_', ':-_GE_', ':-_GT_'];
    Lkw_rev_opernd := 
            [':-_GT_', ':-_GE_', ':-_EQ_', ':-_NE_', ':-_LE_', ':-_LT_'];

    gauche := lhs (expr);
    droite := rhs (expr);
    opernd := op (0, expr);
    if not member (opernd, Lopernd, 'i') then
        error
    end if;
    if gauche = ':-_LEADER_' or gauche = ':-_RANK_' then
        keyword := gauche;
        relop := Lkw_opernd [i];
        if type (droite, ':-specfunc(:-algebraic,:-_DERIV_)') or
                    type (droite, ':-specfunc(:-algebraic,:-_PROPER_)') then
            if keyword <> ':-_LEADER_' or 
			(relop <> ':-_EQ_' and relop <> ':-_NE_') then 
		error 
	    end if;
            modifier := op (0, droite);
            derv := op (droite);
        else
            modifier := ':-_IDENTICAL_';
            derv := droite
        end if
    elif gauche = ':-_ORDER_' and type (droite, ':-nonnegative') then
        keyword := gauche;
        relop := Lkw_opernd [i];
        derv := droite;
    elif droite = ':-_LEADER_' or droite = ':-_RANK_' then
        keyword := droite;
        relop := Lkw_rev_opernd [i];
        if type (gauche, ':-specfunc(:-algebraic,:-_DERIV_)') or
                    type (gauche, ':-specfunc(:-algebraic,:-_PROPER_)') then
            if keyword <> ':-_LEADER_' or
			(relop <> ':-_EQ_' and relop <> ':-_NE_') then 
		error 
	    end if;
            modifier := op (0, gauche);
            derv := op (gauche);
        else
            modifier := ':-_IDENTICAL_';
            derv := gauche
        end if
    elif droite = ':-_ORDER_' and type (gauche, ':-nonnegative') then
        keyword := droite;
        relop := Lkw_rev_opernd [i];
        derv := gauche;
    else
        error
    end if;
    return keyword, relop, modifier, derv;
end proc:

# Subfunction of EQUATIONS
# Equations (ideal)
#
# If ideal is a regchain, returns the list of the equations.
# If ideal is an intersection of regchains, returns the list of 
# the lists of equations of the regchains.

EquationsSafe := proc 
		(regchain, fullset, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local result, arguments;

    arguments := "Equations" (regchain, fullset);
    arguments := arguments, 
        ToBMIOptionSequence (':-_JET_', outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    result := BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
    return result;
end proc:

# Subfunction of EQUATIONS
# Equations (list of equations, leader = derivative(u), ideal)
#
# Returns the sequence of the matching equations

EquationsWithCriterionSafe := proc (list_or_table, 
                            keyword, relop, modifier, derv, 
                            a_table, fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "EquationsWithCriterion"
	    (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table),
	     keyword, relop, modifier, 
	     ConvertString (a_table [':-_FREEZESEQ_'], derv),
	     fullset, a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:
