##PROCEDURE(help) DifferentialAlgebra0[EssentialComponents]
##HALFLINE        returns an irredundant decomposition of the radical of the differential ideal generated by a differential polynomial
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##- EssentialComponents ('p', 'R', 'opts')
##
##PARAMETERS
##- 'p' : a differential polynomial
##
##- 'R' : a differential polynomial ring
##
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `EssentialComponents` ('p', 'R') returns an irredundant
##  representation of the radical of the differential ideal generated
##  by 'p', as an intersection of radical differential ideals 
##  presented by regular differential chains. The differential polynomial
##  'p' is regarded as an element of 'R'. The coefficient field
##  involves the rational numbers and the independent variables.
## 
##- Each regular differential chain of the computed representation
##  is itself made of a unique differential polynomial.
##
##- If some *parameters* are involved in 'p', they must be moved to the 
##  base field (see below), else, the implicit equations which define 
##  them are automatically inserted in the processed lists of differential
##  polynomials, and, make the function raise an error.
##
##- This function can be called through "RosenfeldGroebner", by providing
##  the option `singsol` = `essential`.
##
##- This function, which applies the Low Power Theorem 
##  [R50, chapter III; K73, chapter IV, section 15],
##  relies on the analysis of preparation equations (see "PreparationEquation").
##
##- By default, the built regular differential chains are assumed
##  to hold the attributes: `differential`, `autoreduced`, `primitive`,
##  `squarefree`, `normalized` and `coherent`. 
##  For more details on these attributes, see "DifferentialAlgebra0".
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `attributes` = 'L' where 'L' is a list involving some of the
##  keywords `differential`, `prime`, `primitive`, `squarefree`, `coherent`,
##  `autoreduced`, `normalized`. This option permits to customize the
##  list of attributes of the built regular differential chains.
##  The presence of some of the attributes may imply, automatically,
##  the presence of some other ones. For more details, see 
##  "DifferentialAlgebra0".
##
##- `basefield` = `field` (`generators` = 'G', `relations` = 'regchain').
##  This option permits to perform the computation, regarding 'p'
##  as a differential polynomial with coefficients in the differential 
##  field _k_ presented by  the list of *derivatives* 'G' and the regular 
##  differential chain 'regchain'.
##  The field _k_ contains the rational numbers.
##  Its set of generators is made of the independent variables, plus
##  all the dependent variables occuring in 'G' or in the differential
##  polynomials of 'regchain'.
##  Every polynomial expression belonging to the differential ideal
##  defined by 'regchain', is zero in _k_. Every other polynomial
##  expression between the generators of _k_, is invertible in _k_.
##  Notes:
##-- Any of the arguments `generators` = 'G', and, `relations` = 'regchain'
##   can be omitted.
##-- It is required that the generators of _k_ appear at the bottom of
##   the ranking of 'R', and, that any *block* which involves a generator
##   of _k_, purely consists of generators of _k_.
##  
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [t], blocks = [y]);
##
##> p := y[t]^3-4*y[t]*y*t+8*y^2;
##< y[t]^3-4*y[t]*y*t+8*y^2
##
##- The RosenfeldGroebner algorithm returns three regular differential chains: 
##  the first one corresponding to the general solution, and, the two other 
##  ones corresponding to singular(i.e. non-general) solutions. 
##
##> ideal := RosenfeldGroebner([p], R);
##> Equations(ideal);
##< [[y[t]^3-4*y[t]*y*t+8*y^2], [27*y-4*t^3], [y]]
##
##- However, the regular differential chain corresponding to the case _y = 0_
##  is redundant and can be omitted in the representation. 
##
##> idealmin := EssentialComponents(p, R);
##> Equations(idealmin);
##< [[y[t]^3-4*y[t]*y*t+8*y^2], [27*y-4*t^3]]
##
##- The following example illustrates the `basefield` argument.
##  One introduces two dependent variables 's' and 'c', to encode
##  the sine and the cosine functions.
##
##> R := DifferentialRing(derivations = [t], blocks = [y, [s, c]]);
##> field_rels := PretendRegularDifferentialChain([c^2+s^2-1, c[t] + s], R, pretend = false);
##
##> p := c*y[t]^3-4*y[t]*y*s+8*s*y^2;
##< c*y[t]^3-4*y[t]*y*s+8*s*y^2
##
##- The RosenfeldGroebner algorithm returns three cases. One of them
##  introduces a spurious relation between 's' and 'c'.
##
##> ideal := RosenfeldGroebner([p, c^2+s^2-1, c[t] + s], R);
##
##> Equations(ideal);
##< [[c^2+s^2-1, c[t]+s, c*y[t]^3-4*y[t]*y*s+8*s*y^2], [c^2+s^2-1, c[t]+s, y], [c^2-1, s, y[t]]]
##
##- By moving 's' and 'c' into the base field of the equation, it is
##  not only possible to avoid the spurious relation between 's' and 'c'
##  but also to apply the Low Power Theorem. The field defining relations
##  appear at the bottom of the regular differential chain.
##
##> idealmin := EssentialComponents(p, basefield = field (relations = field_rels), R);
##> Equations(idealmin);
##< [[c^2+s^2-1, c[t]+s, c*y[t]^3-4*y[t]*y*s+8*s*y^2]]
##
##SEEALSO
##- "DifferentialAlgebra0", "RosenfeldGroebner", "PreparationEquation"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Indets" : Help:DifferentialAlgebra0[Indets]
##- "FieldElement" : Help:DifferentialAlgebra0[FieldElement]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "PreparationEquation" : Help:DifferentialAlgebra0[PreparationEquation]


# EXPORTED

EssentialComponents := proc (
    F, 
    dring,
    {basefield:=':-_U_'},
    {attributes::list(identical(:-differential,:-prime,:-autoreduced,
                    :-primitive,:-squarefree,:-normalized,:-coherent)):=
            [':-differential',':-autoreduced',':-primitive',':-squarefree',
             ':-normalized',':-coherent']},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'},$)

    local gens, rels, eqns, arguments, innot, outnot, result, singsol, dimlb;

    if not IsDifferentialRing (dring) then
        error _ERRDRNG_, "last", convert (dring, ':-string');
    end if;
    try
	eqns := ToExpandedRatfracList (F, ':-no_denom' = true)
    catch:
	error _ERRPOLY_, "first", convert (F, ':-string')
    end try;
    try
        gens, rels := BaseField (basefield);
    catch:
	error
    end try;
    singsol := ':-_ESSENTIAL_';
    dimlb := ':-_SAFECASE_';
    arguments := "RosenfeldGroebner"
			(ConvertString (dring [':-_FREEZESEQ_'], eqns),
			 "[]",
			 ConvertString (dring [':-_FREEZESEQ_'], gens),
			 ConvertString (dring [':-_FREEZESEQ_'], rels),
			 ConvertString (dring [':-_FREEZESEQ_'], attributes),
			 dring,
			 convert (singsol, ':-string'),
			 convert (dimlb, ':-string'),
			 false); # The call comes from EssentialComponent

    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (dring));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (dring));
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (dring [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    result := map (ToRegchain, result, outnot, 
			dring [':-_FREEZESEQ_'], dring [':-_UNNUMBERSEQ_']);
    return result;
end proc:
