# Commented out. This function is no more exported

#PROCEDURE(help) DifferentialAlgebra0:-Evaluate
#HALFLINE        Evaluates a rational differential fraction
#AUTHOR          F. Boulier
#DATE            September 2009
#
#CALLINGSEQUENCE
#
#- Evaluate ('F','subst','R','opts')
#- Evaluate ('L','subst','R','opts')
#
#PARAMETERS
#- 'F' : a rational differential fraction
#- 'L' : a list or a set of rational differential fractions
#- 'subst' : a list or a set of equations
#- 'R' : a differential polynomial ring or ideal
#- 'opts' (optional) : a sequence of options
#
#DESCRIPTION
#
#- The function call `Evaluate`('F','subst','R') evaluates 'F' at 'subst'
#  and returns the result. The rational differential fraction 'F' is
#  regarded as an element of 'R', or, of its embedding ring, if 'R'
#  is an ideal. 
#
#- The 'subst' argument is a list or a set of equations of the form
#  _v[i] = alpha[i]_, where the _v[i]_ are *derivatives* or independent
#  variables and the _alpha[i]_ are general expressions. The _v[i]_
#  must be pairwise distinct.
#
#- The returned expression is either in the `diff` or in the `Diff`
#  notation. The default output notation is `diff`.
#
#- The function call `Evaluate`('L','subst','R') returns a list or
#  a set of evaluated rational differential fractions.
#
#- The evaluation rules are as follows. 
#  If _u_ is an independent variable occuring in 'F', then it is
#  replaced by some _alpha[i]_ if and only if the corresponding _v[i]_ 
#  is equal to _u_. Let _u_ be a *derivative*, occuring in 'F'.
#-- If _u_ is not the derivative of any _v[i]_, then _u_ is left
#   unevaluated.
#-- If _u_ is the derivative of exactly one _v[i]_, then, _u_ is
#   replaced by `diff` (_alpha[i]_, _theta_), or `Diff` (_alpha[i]_, _theta_),
#   depending on the output notation, where _theta_ denotes the
#   derivation operator such that _u = theta*v[i]_.
#-- If _u_ is the derivative of many different _v[i]_, then, an equation 
#   such that _v[i]_ has maximal *order* is selected for performing
#   the evaluation.
#
#- This function is used to compute power series solutions of
#  differential ideals presented by regular differential chains.
#  See "PowerSeriesSolution".
#
#- This command is part of the `DifferentialAlgebra0` package. It can be
#  called using the form `Evaluate(...)` after executing the command
#  `with`('DifferentialAlgebra0'). It can also be directly called using the
#  form `DifferentialAlgebra0[Evaluate](...)`.
#
#OPTIONS
#- The 'opts' arguments may contain one or more of the options below.
#
#- `input` = `jet`, `tjet`, `diff` or `Diff`. 
#  Specifies the notation used for the first argument and the
#  left handsides of 'subst'. 
#  If not specified, the notation of 'R' is used (see "Notation").
#
#- `output` = `diff` or `Diff`. Specifies the 
#  notation used for the result of the function call. 
#  If not specified, the notation of 'R' is used,
#  provided that it is `diff` or `Diff`.
#
#- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
#  notation used for both input and output. 
#  This option is overriden by the two options above.
#
#SECTION(nohelp) Other options
#
#- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
#  computation. Default is zero (no time out).
#
#- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
#  computation. Default is zero (no memory out).
#
#- `cellsize` = `small` or `large`. Specifies the size of the cells that
#  must be allocated by the BLAD libraries.
#
#EXAMPLES
#
#> with(DifferentialAlgebra0):
#> R := DifferentialRing(derivations = [t,x], blocks = [u,v]);
#< R := DRing
#
#- Simple evaluations
#
#> Evaluate(u[x,x]+u, [u=sin(x)], R);
#< 0
#
#> Evaluate(u[x] + v, [u=sin(x),v=exp(1/t)], R);
#<  cos(x) + exp(1/t)
#
#- The derivatives _u_ and _u[x]_ are left unevaluated.
#
#> Evaluate(u + u[x] + u[x,x], [u[x,x]=0], R);
#< diff(u(t,x),x)+u(t,x)
#
#- The following derivatives occur in generic power series solutions
#  of the heat equation _u[t,t] = u[x]_. The values of these *derivatives*, 
#  at (say) the origin, are often viewed as given by two functions _f(x)_
#  and _g(x)_. 
#  The `Evaluate` function is used by "PowerSeriesSolution",
#  to perform this renaming.
#
#> L := [ seq(u[x$i], i=0..3), seq(u[t,x$i], i=0..3) ];
#< [u[], u[x], u[x, x], u[x, x, x], u[t], u[t, x], u[t, x, x], u[t, x, x, x]]
#
#> Evaluate(L, [u=f(x), u[t]=g(x)], R);
#< [f(x), diff(f(x),x), diff(diff(f(x),x),x), diff(diff(diff(f(x),x),x),x), g(x), diff(g(x),x), diff(diff(g(x),x),x), diff(diff(diff(g(x),x),x),x)]
#
#> PowerSeriesSolution(PretendRegularDifferentialChain([u[t,t]-u[x]],R), 2, [u=f(x), u[t]=g(x)]) [1];
#< u(t,x) = f(0)+eval(diff(f(x),x),{x = 0})*x+g(0)*t+1/2*eval(diff(diff(f(x),x),x),{x = 0})*x^2+eval(diff(g(x),x),{x = 0})*t*x+1/2*eval(diff(f(x),x),{x = 0})*t^2
#
#SEEALSO
#- "DifferentialAlgebra0", "PowerSeriesSolution"

# Evaluate := proc 
#     (ratfracs,
#      subst::{list(equation),set(equation)},
#      drideal,
#      {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
#      {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
#      {_OUTPUT_::identical(:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
#      {_TIMEOUT_::nonnegative:=0},
#      {_MEMOUT_::nonnegative:=0},
#      {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)
# 
#     local result;
#     try
# 	result := Evaluate2 (ratfracs, subst, drideal, 
# 			     _IONOT_, _INPUT_, _OUTPUT_, 
# 			     _TIMEOUT_, _MEMOUT_, _CELLSZ_);
#     catch:
# 	error
#     end try;
#     return result;
# end proc:
# 
# # Local to Evaluate.
# # Needed in order to call other DifferentialAlgebra0 exported functions
# 
# Evaluate2 := proc (ratfracs, subst, drideal, 
# 		   _ionot_, _input_, _output_, _timeout_, _memout_, _cellsz_)
# 
#     local R, innot, outnot, result, indepsubs, freeders, 
# 				eqns, lhs_subst, new_subst, i;
# 
# # Force the input notation to be tjet.
#     if not IsIdealDRing (drideal) then
# 	error _ERRIDRNG_, "third", convert (drideal, ':-string');
#     elif IsDifferentialRing (drideal) then
# 	R := DifferentialRing (drideal, _IONOT_ = ':-_TJET_');
#     elif type (drideal, list) then
# 	R := DifferentialRing (drideal [1], _IONOT_ = ':-_TJET_');
#     else
# 	R := DifferentialRing (drideal, _IONOT_ = ':-_TJET_');
#     end if;
# 
# # Force the output notation to be diff or Diff
#     innot := FixNotation (_input_, _ionot_, NotationSafe (drideal));
#     outnot := FixNotation (_output_, _ionot_, NotationSafe (drideal));
#     if outnot = ':-_JET_' or outnot = ':-_TJET_' then
# 	outnot := ':-_DIFF_';
#     end if;
# 
# # Convert ratfracs and rhs(subst) to be in the tjet notation
# # Checks that the right handsides of the substitutions are pairwise
# # distinct.
#     eqns := ToExpandedRatfracList (ratfracs);
#     eqns := NormalForm (eqns, R, _INPUT_=innot);
# 
#     lhs_subst := map (lhs, subst);
#     lhs_subst := NormalForm (lhs_subst, R, _INPUT_=innot);
#     if nops (lhs_subst) <> nops (convert (lhs_subst, set)) then
# 	error _ERREVAL_
#     end if;
#     new_subst := [ seq (lhs_subst [i] = rhs (subst [i]), 
# 					i = 1 .. nops (subst)) ];
# # go
#     result, indepsubs := EvaluateSafe (eqns, new_subst, R, outnot);
#     result := FromExpandedRatfracList (result, ratfracs);
#     return result;
# end proc:

# Local to Evaluate2.
# Called also by PowerSeriesSolution and PowerSeriesSystem
#
# ratfracs, rhs (subst) and R are in tjet notation
# outnot = diff or Diff
#
# Returns a sequence: the result of the evaluation, indepsubs
# where indepsubs = the part of subst which applies to independent variables

EvaluateSafe := proc (eqns, subst, R, outnot)
    local indepsubs, dersubs, newsubs, freeders, koeffs, 
					terms, u, result, i, j;

    indepsubs, dersubs, freeders := 
			EvaluateProcessExpansionPoint (eqns, subst, R, outnot);
    if freeders <> [] then
	newsubs := zip ((x,y) -> x=y, 
			freeders, NormalForm (freeders, R, '_OUTPUT_'=outnot));
        dersubs := [ op (dersubs), op (newsubs) ];
    end if;
    result := NULL;
    u := MinRankElement (Indets (R), R);
    for i from 1 to nops (eqns) do
	koeffs, terms := Coeffs (eqns [i], u, R);
	if indepsubs <> [] then
	    koeffs := subs (op (indepsubs), koeffs);
        end if;
	if dersubs <> [] then
	    terms := subs (op (dersubs), terms);
	end if;
	result := result, add (koeffs[j]*terms[j], j = 1 .. nops(terms));
    end do;

    return [ result ], indepsubs;
end proc:

# Local to EvaluateSafe
# 
# Expansion is a list of substitutions.
# This function determines all the derivatives of the right handsides
# of expansion, which occur in the list ratfracs of rational differential
# fractions. 
# 
# It returns a sequence indepsubs, dersubs, freeders
# where
# 
# - indepsubs is the part of expansion whose right handsides are
#   independent variables
# 
# - dersubs is the complete list of substitutions, determined
#   from expansion, needed to evaluate all the rational differential
#   fractions
# 
# - freeders is the list of the derivatives occuring in the
#   rational differential fractions, which are not concerned by
#   the substitutions.
# 
# The arguments ratfracs, lhs (expansion) and R are in the tjet notation.
# The output notation outnot is used to distinguish diff from Diff

EvaluateProcessExpansionPoint := proc (ratfracs, expansion, R, outnot)
    local lhs_expansion, arguments, 
          indepsubs, dersubs, vars, indices, operators, freeders, i;

    arguments := "ProcessExpansionPoint" 
	(ConvertString (R [':-_FREEZESEQ_'], ratfracs), 
	 ConvertString (R [':-_FREEZESEQ_'], map (lhs, expansion)),
	 R);
    arguments := arguments,
		    ToBMIOptionSequence 
			(':-_TJET_', ':-_TJET_', 0, 0, ':-_SMALL_');
    indepsubs, vars, indices, operators, freeders := 
			BMIBladEvalSafe (R [':-_UNNUMBERSEQ_'], arguments);
    indepsubs := [ seq (expansion [indepsubs [i]], i = 1 .. nops(indepsubs)) ];
    dersubs := [ seq (vars [i] = EvaluateProcessExpansionPointDiff 
			(rhs (expansion [indices [i]]), operators [i], outnot),
		    i = 1 .. nops (vars)) ];
    return indepsubs, dersubs, freeders;
end proc:

# Local to EvaluateProcessExpansionPoint
# Differentiates expr with respect to theta.
# expr is the right handside of a substitution
# theta is a derivation operator
# outnot is the output notation.

EvaluateProcessExpansionPointDiff := proc (expr, theta, outnot)
    local x, seqn;
    if theta = 1 then
	return expr;
    else
	seqn := seq (x $ degree (theta, x), x in indets (theta));
	if outnot = ':-_DIFF_' then
	    return diff (expr, seqn);
	else
	    return Diff (expr, seqn);
	end if;
    end if;
end proc:
