##TOPIC(help) DifferentialAlgebra0[Examples]
##SECTION Examples
##
##SUBSECTION Study of a differential-algebraic system
##- The following example is borrowed from [HW96].
##  It features a differential-algebraic system of differential
##  index _2_, describing a steering wheel. The dependent variables
##  _x(t)_ and _y(t)_ are space coordinates. The dependent variable
##  _z(t)_ is a command.
##
##> with(DifferentialAlgebra0):
##> syst := [ diff(x(t),t) = 7/10*y(t) + sin(25/10*z(t)), diff(y(t),t) = 14/10*x(t) + cos(25/10*z(t)), x(t)^2 + y(t)^2 = 1 ]; 
##< [ diff(x(t),t) = 7/10*y(t) + sin(25/10*z(t)), diff(y(t),t) = 14/10*x(t) + cos(25/10*z(t)), x(t)^2 + y(t)^2 = 1 ]
##
##- The "RosenfeldGroebner" function permits to transform this
##  system into an equivalent system of differential index _0_.
##  In particular, it permits:
##
##-- To compute a complete set of relations, defining the algebraic
##   variety, over which, the integral curves of the system must remain.
##   These relations are essential for choosing consistent initial values.
##
##-- To compute the missing ODE, which expresses the evolution of _z(t)_.
##
##- First, one transforms the above system into a polynomial differential
##  form, by introducing two new dependent variables _s(t)_ and _c(t)_
##  which encode the sine and the cosine.
##
##> syst := [ diff(x(t),t) = 7/10*y(t) + s(t), diff(y(t),t) = 14/10*x(t) + c(t), x(t)^2 + y(t)^2 = 1, diff(s(t), t) = 25/10*diff(z(t),t)*c(t), diff(c(t),t) = - 25/10*diff(z(t),t)*s(t), s(t)^2 + c(t)^2 = 1 ];
##< [ diff(x(t),t) = 7/10*y(t) + s(t), diff(y(t),t) = 14/10*x(t) + c(t), x(t)^2 + y(t)^2 = 1, diff(s(t), t) = 25/10*diff(z(t),t)*c(t), diff(c(t),t) = - 25/10*diff(z(t),t)*s(t), s(t)^2 + c(t)^2 = 1 ]
##
##- Then, one defines a differential polynomial ring, endowed with an
##  *orderly* ranking, and, one simplifies the system.
##
##> R := DifferentialRing (derivations = [t], blocks = [[z,y,x,s,c]]);
##
##> ideal := RosenfeldGroebner (syst, R, input=diff);
##
##- Because the ranking is *orderly*, the *order* _0_ differential 
##  polynomials of the returned regular differential chain
##  provide the algebraic variety, over which, the integral curves of
##  the system must remain. These differential polynomials form
##  a non-differential squarefree regular chain (see "RegularChains"), 
##  that can be solved from bottom-up.
##  They are important for determining consistent initial values.
##  Indeed, once _z(0)_ is chosen, _s(0)_ and _c(0)_ are fixed, and,
##  _x(0)_ cannot be chosen freely.
##
##> RewriteRules(ideal [1], order = 0);
##< [s^2 = -c^2+1, x^4 = -20/21*x^3*c+341/441*x^2+20/21*x*c+100/441*c^2, y = -(-441*x^3*s-210*x^2*s*c+100*x*s*c^2+341*x*s+210*s*c)/(100*c^3-100*c)]
##
##- The missing ODE, which expresses the evolution of _z(t)_, is the
##  element of the regular differential chain, whose leading
##  derivative is a derivative of _z_.
##
##> RewriteRules(ideal [1], leader = derivative (z));
##< [z[t] = -(58344300*x^3*c^4-54519507*x^3*c^2+10892700*x^3+55566000*x^2*c^5-56831670*x^2*c^3+13461000*x^2*c-43306200*x*c^4+40951407*x*c^2-8422700*x-37044000*c^5+36335670*c^3-7987000*c)/(11025000*c^5-11025000*c^3+2500000*c)]
##
##ENDSUBSECTION
##
##SUBSECTION The Henri-Michaelis-Menten reduction, in chemistry
##
##- This example features a model reduction for chemical reaction
##  systems, based on a *quasi-steady state approximation* [BLLM07]. 
##  The process is illustrated over a famous case, but, has more
##  general applications.
##
##- The following system is derived from a basic enzymatic reaction
##  system, by application of the mass-action law.
##  Four chemical species are involved: an enzyme _E_, a substrate _S_,
##  an intermediate complex _C_, and a product _P_.
##  There are three chemical reactions.
##  The dependent variables represent the concentrations of the four species.
##
##TABLE(width="80%",border="0",colwidth="1|1|1|1|1|1|1|1|1|1|1")
##ROW       | _k[1]_ |      | |       | _k[-1]_ |        | |      | _k[2]_ |
##ROW _E+S_ | `--->` | _C_, | | _C_   | `---->` | _E+S_, | |  _C_ | `--->` | _E+P_.
##ENDTABLE
##
##> with(DifferentialAlgebra0):
##> syst := [ diff(E(t),t) = -k[1]*E(t)*S(t) + k[-1]*C(t) + k[2]*C(t), diff(S(t),t) = -k[1]*E(t)*S(t) + k[-1]*C(t), diff(C(t),t) = k[1]*E(t)*S(t) - k[-1]*C(t)  - k[2]*C(t), diff(P(t),t) = k[2]*C(t)];
##< [ diff(E(t),t) = -k[1]*E(t)*S(t) + k[-1]*C(t) + k[2]*C(t), diff(S(t),t) = -k[1]*E(t)*S(t) + k[-1]*C(t), diff(C(t),t) = k[1]*E(t)*S(t) - k[-1]*C(t)  - k[2]*C(t), diff(P(t),t) = k[2]*C(t)]
##
##- Under some assumptions, this system can be approximated by
##  the so-called Henri-Michaelis-Menten formula. We are going to
##  show how this formula can be automatically recovered.
##
##- The main assumption is: the revertible reaction is *fast*, and,
##  the last one is *slow*. With other words, _k[1], k[-1]_ `>>` _k[2]_.
##
##- Taking into account this main assumption, actually amounts to
##  perform a *quasi-steady state approximation* of the above system.
##  This can be done as follows.
##
##-- First, replace, in the system, the contribution
##  _k[1]*E(t)*S(t) - k[-1]*C(t)_ of the fast reactions, which
##  occurs in the three first ODE, by a new
##  dependent variable _F[1]_. This variable represents the *unknown*
##  contribution, which cannot be assumed to be _0_, of the fast
##  reactions, after the transient step. Somehow, this variable
##  just encodes that *fast reactions are reactions*.
##
##-- Second, add the algebraic equation _k[1]*E(t)*S(t) - k[-1]*C(t) = 0_
##   to the system. This equation forces the dynamics of the system 
##   to stays over the algebraic variety that it defines.
##   Somehow, it encodes the *speed*.
##
##-- Third, eliminate the new dependent variable _F[1]_ from the
##   differential-algebraic system. This elimination, performed
##   using "RosenfeldGroebner", computes the simplest ODE which
##   describes the evolution of _S(t)_, and, which is a consequence
##   of the DAE.
##
##- The two first steps yield the following differential-algebraic 
##  system. The rate constants are considered as *parameters*.
##
##> DAE := [ diff(E(t),t) = -F[1](t) + k[2]*C(t), diff(S(t),t) = - F[1](t), diff(C(t),t) = F[1](t) - k[2]*C(t), diff(P(t),t) = k[2]*C(t), k[1]*E(t)*S(t) - k[-1]*C(t) = 0];
##< [ diff(E(t),t) = -F[1](t) + k[2]*C(t), diff(S(t),t) = - F[1](t), diff(C(t),t) = F[1](t) - k[2]*C(t), diff(P(t),t) = k[2]*C(t), k[1]*E(t)*S(t) - k[-1]*C(t) = 0]
##
##> species := [C,E,P,S];
##< [C,E,P,S]
##
##> params := [k[1],k[-1],k[2]];
##< [k[1],k[-1],k[2]]
##
##> R := DifferentialRing(blocks = [F[1], species, params], notation = diff, derivations = [t], parameters = params);
##
##- Let us perform the third step.
##  A second list, involving one differential polynomial, _C(t)_, is
##  passed to "RosenfeldGroebner" as an argument, to
##  avoid a spurious discussion on the case _C(t) = 0_.
##  The rate constants are moved to the base field of the equations,
##  to avoid generating degenerate cases, by discussing their values.
##
##> ideal := RosenfeldGroebner(DAE, [C(t)], basefield = field (generators = params), R);
##
##- The returned regular differential chain involves a differential
##  polynomial expressing the evolution of _S(t)_, however, this relation
##  is not yet the Henri-Michaelis-Menten formula.
##  Indeed, the minor hypotheses have not yet been taken into account.
##  This formula is just the simplest formula that can be deduced
##  from the DAE, by using the main assumption *only*.
##
##> RewriteRules(ideal[1], leader=diff(S(t),t));
##< [diff(S(t),t) = -(E(t)*S(t)^2*k[1]^2*k[2]+E(t)*S(t)*k[1]*k[-1]*k[2])/(E(t)*k[1]*k[-1]+S(t)*k[1]*k[-1]+k[-1]^2)]
##
##- One introduces a few more parameters, and the linear conservation laws
##  of the system.
##
##> conservation_laws := [E(t) + C(t) = E[0] + C[0], S(t) + C(t) + P(t) = S[0] + C[0] + P[0]];
##< [E(t) + C(t) = E[0] + C[0], S(t) + C(t) + P(t) = S[0] + C[0] + P[0]]
##
##> params := [k[1],k[-1],k[2],E[0],C[0],S[0],P[0]];
##< [k[1],k[-1],k[2],E[0],C[0],S[0],P[0]]
##
##- A minor assumption is: _P(0)_ = _C(0)_ = _0_.
##
##> hypotheses :=  subs(P[0]=0, C[0]=0, conservation_laws);
##< [E(t) + C(t) = E[0], S(t) + C(t) + P(t) = S[0]]
##
##> R :=  DifferentialRing (blocks = [F[1], species, params], notation = diff, derivations = [t], parameters = params);
##
##- One could actually proceed from the formula established above, but, it
##  is simpler to perform the whole simplification anew.
##
##> ideal := RosenfeldGroebner ([ op(DAE), op(hypotheses) ], R, basefield = field (generators = params));
##
##> formula := RewriteRules (ideal[1], leader=diff(S(t),t));
##< [diff(S(t),t) = -(S(t)^2*k[1]^2*k[2]*E[0]+S(t)*k[1]*k[-1]*k[2]*E[0])/(S(t)^2*k[1]^2+2*S(t)*k[1]*k[-1]+k[1]*k[-1]*E[0]+k[-1]^2)]
##
##- Not yet: one still needs to perform some renaming on parameters, and
##  to neglect the term _K*E[0]_, assuming _S(0)_ `>>` _E(0)_.
##
##> formula := subs (k[-1]=K*k[1], formula):
##> formula := normal (formula):
##> formula := algsubs (k[2]*E[0]=V[max], formula):
##
##- Here is the Henri-Michaelis-Menten formula.
##
##> formula := normal (subs (K*E[0]=0, formula));
##
##ENDSUBSECTION
##
##SUBSECTION The pendulum
##
##- Consider the set of differential polynomials describing the motion of a
##  pendulum, that is, a point mass _m_ suspended from a massless rod of 
##  length _l_ under the influence of gravity _g_, in Cartesian coordinates 
##  _(x,y)_. 
##  The Lagrangian formulation leads to two second order differential 
##  equations with an algebraic constraint. _T_ is the Lagrangian multiplier.
##
##> with(DifferentialAlgebra0):
##> params := [m,l,g]:
##> P := [m*diff(y(t),t,t)+T(t)*y(t)+g, m*diff(x(t),t,t)+T(t)*x(t),x(t)^2+y(t)^2-l^2];
##< [m*diff(y(t),t,t)+T(t)*y(t)+g, m*diff(x(t),t,t)+T(t)*x(t),x(t)^2+y(t)^2-l^2]
##
##- To find the lowest order differential polynomials vanishing on the zeros of
##  the system modeling the pendulum and in particular all the algebraic
##  constraints (that is, the differential polynomials of *order* _0_), use an
##  _orderly_ ranking.
##
##  The second component in the output of "RosenfeldGroebner" corresponds to the
##  equilibria of the pendulum. From the first component of the output, it can
##  be seen that the actual motion is described by two first order differential
##  equations with a constraint: the solution depends on only two arbitrary
##  constants.
##
##> R := DifferentialRing (derivations=[t], blocks = [[T,x,y],params], notation = diff, parameters = params);
##
##> ideal := RosenfeldGroebner(P, R, basefield = field (generators = params));
##
##> RewriteRules(ideal[1], output=jet);
##< [x^2 = -y^2+l^2, y[t]^2 = -(T*y^2*l^2-T*l^4+y^3*g-y*l^2*g)/m/l^2, T[t] = -3*y[t]*g/l^2]
##
##> RewriteRules(ideal[2], output=jet);
##< [y^2 = l^2, x = 0, T = -y*g/l^2]
##
##- To obtain the differential system satisfied by _x_ and _y_ alone, 
##  eliminate _T_.
##  Note here that the motion is given by a second order differential equation
##  in _y_. Then the Lagrange multiplier is given explicitly in terms of _y_
##  and its first derivative.
##
##> R := DifferentialRing(derivations=[t], blocks = [T,[x,y],params], notation = diff, parameters = params);
##
##> ideal := RosenfeldGroebner(P, R, basefield = field(generators = params));
##
##> RewriteRules(ideal, output=jet, leader < T(t));
##< [[x^2 = -y^2+l^2, y[t,t] = -(-y[t]^2*y*m*l^2-y^4*g+2*y^2*l^2*g-l^4*g)/(y^2*m*l^2-m*l^4)], [y^2 = l^2, x = 0]]
##
##ENDSUBSECTION
##
##SUBSECTION General and singular solutions of a single differential equation
##
##- In general, when "RosenfeldGroebner" is applied on a single differential
##  polynomial _F_, it returns at least one component involving _F_, and,
##  possibly, some other components. The first component describes the
##  *general* solution of _F_. The other ones describe the *singular* solutions.
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[t], blocks = [y], notation = diff);
##
##> F := diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2;
##< diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2
##
##- Over this example, "RosenfeldGroebner" returns the general component
##  plus two singular ones. As we shall see, one of the singular components
##  is *essential*, while the other one is not [K73,H99].
##
##> ideal := RosenfeldGroebner([F], R);
##
##> Equations(ideal);
##< [[diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2], [27*y(t)-4*t^3], [y(t)]]
##
##- The three components can be solved explicitly.
##  The solutions of the general component writes _y(t)=c*(t-c)^2_,
##  where _c_ denotes an arbitrary constant.
##  The solution of the first singular component writes _y(t)=4/27*t^3_.
##  It is not a particular case of the general solution.
##  This first singular component is said to be essential.
##  The solution of the second singular component writes _y(t)=0_.
##  It is a particular case of the general solution, since it is obtained
##  for _c = 0_. This second singular component is said to be non-essential.
##
##- The function "EssentialComponents" only returns the essential
##  components of _F_:
##
##> ideal := EssentialComponents(F, R);
##
##> Equations(ideal);
##< [[diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2], [27*y(t)-4*t^3]]
##
##ENDSUBSECTION
##
##SUBSECTION Power Series Solutions
##
##- The regular differential chains returned by "RosenfeldGroebner",
##  permit to compute integral formal power series solutions of the
##  input system. By *integral*, it is meant that the exponents are
##  integers. By *formal*, it is meant that convergence issues are
##  not addressed. The following system provides a good example
##  of what a general PDE polynomial system may look like, Its power
##  series solutions turn out to be polynomials.
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [[v,u]]);
##> syst := [u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]];
##< [u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]]
##
##> ideal := RosenfeldGroebner(syst, R);
##> RewriteRules(ideal[1]);
##< [u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y])/u, v[x,x] = u[x]]
##
##- Looking at the leading derivatives of the returned regular differential
##  chain, one sees that the system solutions depend on three constants,
##  or initial values, which can be chosen almost arbitrary
##  (see [DL84] for related undecidability results). Indeed, they 
##  must be solutions of the polynomial system of equations, returned 
##  by "PowerSeriesSystem":
##
##>(quote) PowerSeriesSystem(ideal[1]);
##< [D[2](u)(0,0)^2-2*u(0,0) = 0, D[1](u)(0,0)^2-4*u(0,0) = 0], [u(0,0) <> 0, D[1](u)(0,0) <> 0, D[2](u)(0,0) <> 0]
##
##- Let us choose the following initial values, and, check that they do
##  satisfy the above polynomial system.
##
##> iv := [u=c[0]^2, u[y]=sqrt(2)*c[0], u[x]=2*c[0], v=c[1], v[x]=c[2]];
##< [u=c[0]^2, u[y]=sqrt(2)*c[0], u[x]=2*c[0], v=c[1], v[x]=c[2]]
##
##>(quote) PowerSeriesSystem(ideal[1], iv);
##< [0 = 0, 0 = 0], [c[0]^2 <> 0, 2*c[0] <> 0, 2^(1/2)*c[0] <> 0]
##
##- The "PowerSeriesSolution" function can now compute the 
##  integral formal power series solutions of 'syst, for these initial
##  values.
##
##> sols := PowerSeriesSolution(ideal[1], 3, iv);
##< [v(x,y) = c[1]+(1/2*c[0]^2*2^(1/2)-1/2*2^(1/2))*y+c[2]*x+1/2*c[0]*y^2+2^(1/2)*c[0]*x*y+c[0]*x^2+1/12*2^(1/2)*y^3+1/2*x*y^2+1/2*2^(1/2)*x^2*y+1/3*x^3, u(x,y) = c[0]^2+2^(1/2)*c[0]*y+2*c[0]*x+1/2*y^2+2^(1/2)*x*y+x^2]
##
##> syst := Equations(syst, R, output=diff);
##< [diff(u(x,y),x)^2-4*u(x,y), diff(diff(u(x,y),x),y)*diff(v(x,y),y)-u(x,y)+1, diff(diff(v(x,y),x),x)-diff(u(x,y),x)]
##
##> expand(eval (syst, sols));
##< [0, 0, 0]
##
##ENDSUBSECTION
##
##XREFMAP
##- "Examples" : Help:DifferentialAlgebra0[Examples]
##- "Glossary" : Help:DifferentialAlgebra0[Glossary]
##- "References" : Help:DifferentialAlgebra0[References]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]
##- "PowerSeriesSystem" : Help:DifferentialAlgebra0[PowerSeriesSystem]
##- "PowerSeriesSolution" : Help:DifferentialAlgebra0[PowerSeriesSolution]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]
##- "BelongsTo" : Help:DifferentialAlgebra0[BelongsTo]
##- "Pardi" : Help:DifferentialAlgebra0[Pardi]

