##PROCEDURE(help) DifferentialAlgebra0[FactorDerivative]
##HALFLINE        extracts the derivation operator of a *derivative*
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE 
##- `FactorDerivative`('v', 'R', 'opts')
##
##PARAMETERS
##- 'v' : a *derivative*
##- 'R' : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `FactorDerivative`('v','R') returns a sequence
##  _theta_, _u_ such that _theta_ is the derivation operator, and, 'u'
##  is the dependent variable, associated to _u_ (see "DifferentialAlgebra0").
##  The argument 'v' must be a *derivative* of 'R', or of 
##  its embedding ring if 'R' is an ideal.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `FactorDerivative(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[FactorDerivative](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'v'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[x,y], blocks=[[v,u],p], parameters=[p]);
##
##>(quote) theta, indep := FactorDerivative(u[x,y], R);
##< x*y, u
##
##> Differentiate(indep, theta, R);
##< u[x, y]
##
##>(quote) FactorDerivative(u, R);
##< 1, u
##
##>(quote) FactorDerivative(diff(u(x,y),x), R, notation=diff);
##< x, u(x, y)
##
##SEEALSO
##- "DifferentialAlgebra0", "Differentiate"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Differentiate" : Help:DifferentialAlgebra0[Differentiate]

# EXPORTED
#
# FactorDerivative (derivative, differential ring)
# Returns a sequence: derivation operator, dependent variable

FactorDerivative := proc (derv::algebraic, drideal,
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local result, a_table, innot, outnot, arguments;

    if not IsIdealDRing (drideal) then
        error _ERRIDRNG_, "second", convert (drideal, ':-string');
    elif type (drideal, ':-list') then
        a_table := drideal[1];
    else
        a_table := drideal;
    end if;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    arguments := "FactorDerivative" 
			(ConvertString (a_table [':-_FREEZESEQ_'], derv), 
			 a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    return result;
end proc:
