##PROCEDURE(help) DifferentialAlgebra0[FieldElement]
##HALFLINE        decides membership in differential base fields
##AUTHOR          F. Boulier
##DATE            October 2009
##
##CALLINGSEQUENCE
##
##- FieldElement ('p', 'F', 'R', 'opts')
##- FieldElement ('L', 'F', 'R', 'opts')
##
##PARAMETERS
##
##- 'p' : a differential polynomial
##- 'L' : a list or a set of differential polynomials
##- 'F' : a field description
##- 'R' : a differential ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `FieldElement` ('p','F','R') returns `true`
##  if the differential polynomial 'p' belongs to the differential
##  field _k_ defined 'F' and 'R', else it returns `false`.
##  The differential polynomial 'p' is regarded as a differential 
##  polynomial of 'R', if 'R' is a differential ring, or, of
##  the embedding ring of 'R', if 'R' is an ideal.
##
##- The argument 'F' has the form 
##  `field` (`generators` = 'G', `relations` = 'regchain').
##  It defines a differential field _k_ presented by the
##  list of *derivatives* 'G' and the regular differential chain 'regchain'.
##  The field _k_ contains the rational numbers.
##  Its set of generators is made of the independent variables, plus
##  all the dependent variables occuring in 'G' or in the differential 
##  polynomials of 'regchain'.
##  Every polynomial expression belonging to the differential ideal
##  defined by 'regchain', is zero in _k_. Every other polynomial
##  expression between the generators of _k_, is invertible in _k_.
##  Notes:
##-- Any of the arguments `generators` = 'G', and, `relations` = 'regchain'
##   can be omitted.
##-- It is required that the generators of _k_ appear at the bottom of
##   the ranking of 'R', and, that any *block* which involves a generator
##   of _k_, purely consists of generators of _k_.
##
##- The function call `FieldElement` ('L','F','R') returns a list or
##  a set of ::boolean::.
##
##- Non-trivial base fields are used by the "RosenfeldGroebner" 
##  and the "EssentialComponents" functions.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `FieldElement(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[FieldElement](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p', 'L' and 'G'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for both input and output. This option is overridden by the 
##  option above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with (DifferentialAlgebra0):
##> R := DifferentialRing (derivations = [t], blocks = [u,v,w]);
##
##- With no arguments, the field _k_ is the smallest field involving
##  the rational numbers and the independent variables.
##
##> FieldElement ([1, t, u], field (), R);
##< [true, true, false]
##
##- In this example, the field _k_ is the smallest differential field involving
##  the rational numbers, the independent variables, and, the derivatives 
##  of _v_ and _w_. 
##
##> FieldElement ([u/v[t], w[t,t], 1/(v+w)], field (generators = [v,w]), R);
##< [false, true, true]
##
##- In this example, the field _k_ is presented by generators and relations.
##  The expression _v[t,t]-2_ is _0_ in _k_.
##
##> fieldrels := PretendRegularDifferentialChain ([v[t]^2-4*v], R);
##
##> NormalForm (v[t,t]-2, fieldrels);
##< 0
##
##> FieldElement ((v[t,t]-2)*u+w[t]+1, field (relations = fieldrels, generators = [v,w]), R);
##< true
##
##SEEALSO
##- "DifferentialAlgebra0", "Indets", "RosenfeldGroebner", "EssentialComponents"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Indets" : Help:DifferentialAlgebra0[Indets]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]

# EXPORTED

FieldElement := proc 
    (eqns,
     basefield,
     drideal::table,
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

     local result, leqns, innot, a_table, gens, rels;

    if not IsIdealDRing (drideal) then
	error _ERRIDRNG_, "third", convert (drideal, ':-string')
    elif type (drideal, ':-list') then
	a_table := drideal [1]
    else
	a_table := drideal
    end if;
# Processing the equations
    try
	leqns := ToExpandedRatfracList (eqns)
    catch:
	error _ERRRLRT_, "first", convert (eqns, ':-string')
    end try;
# Processing the basefield parameter
    try
	gens, rels := BaseField (basefield);
    catch:
        error
    end try;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    try
        result := FieldElementSafe (leqns, gens, rels, a_table, innot, 
					 _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    catch:
	error
    end try;
    result := FromExpandedRatfracList (result, eqns);
    return result;
end proc:

# Local to FieldElement
# eqns is a list of differential polynomials
# Returns a list of booleans

FieldElementSafe := 
	proc (eqns, gens, rels, a_table, innot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "FieldElement" 
	(ListOrTableToBMI (a_table [':-_FREEZESEQ_'], eqns), 
	 ConvertString (a_table [':-_FREEZESEQ_'], gens),
	 ConvertString (a_table [':-_FREEZESEQ_'], rels),
	 a_table);
    arguments := arguments,
	ToBMIOptionSequence (innot, ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe ([], arguments);
end proc:
