##PROCEDURE(help) DifferentialAlgebra0[Indets]
##HALFLINE        returns the set of variables its argument depends on
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `Indets`('Ring','select','opts')
##- `Indets`('p','select','R','opts')
##- `Indets`('L','select','R','opts')
##- `Indets`('F','select','R','opts')
##- `Indets`('ideal','select','opts')
##
##PARAMETERS
##- 'Ring' : a differential polynomial ring
##- 'p'    : a differential rational fraction
##- 'L'    : a list or a set of differential rational fractions
##- 'F'    : a field description
##- 'R'    : a differential polynomial ring or ideal
##- 'ideal' : a differential ideal
##- 'select' (optional) : an expression for selecting a subset of the variables
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Indets`('p','R') returns the set
##  of *derivatives* occurring in 'p', viewed as a differential rational
##  fraction of 'R', if 'R' is a ring, else, of the embedding ring of 'R'.
## 
##- The function call `Indets`('L','R') returns the set
##  of *derivatives* occurring in the elements of 'L', viewed as differential
##  rational fractions of 'R'.
##
##- If 'ideal' is a regular differential chain, the function call
##  `Indets`('ideal') returns the set of the *derivatives* occurring in
##  the chain elements. If 'ideal' is a list of regular differential
##  chains, the list of the sets of the *derivatives* occurring in each
##  chain is returned.
##
##- The function call `Indets`('Ring') returns the set of the dependent
##  variables
##  provided at the construction of 'Ring' (see "DifferentialRing").
##  In this case, the two forms, `selection` = `depvars`
##  and `selection` = `derivatives`, of 'select', yield the same result.
##
##- The function call `Indets`('F','R'), where 'F' stands for
##  `field` (`generators` = 'G', `relations` = 'regchain'), 
##  returns the actual set of the generators of the field _k_,
##  presented by the list of *derivatives* 'G' and the regular 
##  differential chain 'regchain'.
##  The field _k_ contains the rational numbers.
##  Its set of generators is made of the independent variables, plus
##  all the dependent variables occuring in 'G' or in the differential 
##  polynomials of 'regchain'.
##  Every polynomial expression belonging to the differential ideal
##  defined by 'regchain', is zero in _k_. Every other polynomial
##  expression between the generators of _k_, is invertible in _k_.
##  Notes:
##-- Any of the arguments `generators` = 'G', and, `relations` = 'regchain'
##   can be omitted.
##-- It is required that the generators of _k_ appear at the bottom of
##   the ranking of 'R', and, that any *block* which involves a generator
##   of _k_, purely consists of generators of _k_.
##
##- The 'select' parameter has the form `selection` = 'keyword' where
##  'keyword' is `depvars`, `indepvars`, `derivatives`, `allvars` or
##  `parameters`. 
##  The default value is `selection` = 'derivatives'.
##  The interpretations of these keywords are listed below.
##
##-- `indepvars`. The independent variables are selected.
##
##-- `derivatives`. The *derivatives* are selected (default).
##
##-- `parameters`. The *derivatives*
##  which are defined as *parameters* are selected (see "Parameters").
##
##-- `depvars`. The dependent variables associated to
##  the *derivatives* are selected.
##
##-- `allvars`. The *derivatives* and the independent
##  variables are selected.
##
##- If 'ideal' is a regular differential chain, the function call 
##  `Indets`('ideal', `selection` = `iniconds` (_n_)) returns the
##  set of the *derivatives* which have *order* less than or equal
##  to _n_, and, which are not the derivative of any leading 
##  derivative of 'ideal'. These *derivatives* correspond to
##  the arbitrary constants which occur in the formal power
##  series solutions of 'ideal'. If 'ideal' is a list, then,
##  a list of sets of derivatives is returned.
##  
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Indets(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Indets](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call
##  `Indets`('ideal'), applies the function also over
##  the differential polynomials which state that the derivatives of
##  the parameters are zero. Default value is `false`.
##  This option is not compatible with the `diff` notation, combined
##  with 'criterion' being equal to `derivatives` or `allvars`.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[x,y], blocks=[[v,u],p,q], parameters = [p,q(x)]);
##
##- The four dependent variables, in the `jet` notation.
##
##SUBSECTION Over a differential ring
##> Indets(R);
##< {p, v, u, q}
##
##- The two independent variables.
##
##> Indets(R, selection = indepvars);
##< {y, x}
##
##- The two *derivatives* which are listed among the *parameters*, in the
##  `diff` notation.
##
##> Indets(R, selection=parameters, output=diff);
##< {p, q(x)}
##ENDSUBSECTION
##
##SUBSECTION Over a rational differential fraction
##> f := u[x]^2-4*u+y/(q[x]+v);
##< u[x]^2-4*u+y/(q[x]+v)
##
##- The *derivatives* occurring in _f_, in the `diff` notation.
##
##> Indets (f, R, output=diff);
##< {v(x,y), u(x,y), diff(u(x,y),x), diff(q(x),x)}
##ENDSUBSECTION
##
##SUBSECTION Over a regular differential chain
##> ideal := RosenfeldGroebner ([u[x]^2-4*u, u[x,y]*v[y]-u+p,v[x,x]-u[x]], R);
##> RewriteRules (ideal);
##< [[u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y]*p)/u, v[x,x] = u[x]], [p = 0, u = 0, v[x,x] = 0]]
##
##- The *derivatives* occurring in the elements of each regular differential
##  chain.
##
##> Indets (ideal);
##< [{p, u, u[y], v[x, x], u[x], v[y]}, {p, u, v[x, x]}]
##
##- The *derivatives* which have *order* less than or equal to _3_, and,
##  which are not derivatives of any leading derivative of each
##  regular differential chain.
##
##> Indets (ideal, selection=iniconds(3));
##< [{p, q, u, v, q[x], q[x,x], q[x,x,x], v[x]}, {q, v, q[x], q[x,x], q[x,x,x], v[x], v[y], v[x,y], v[y,y], v[x,y,y], v[y,y,y]}]
##
##- This function call permits to obtain all the *derivatives* of
##  all the dependent variables, up to *order* _1_.
##
##> Indets (PretendRegularDifferentialChain ([], R), selection=iniconds(1));
##< {p, q, u, v, q[x], u[x], u[y], v[x], v[y]}
##
##ENDSUBSECTION
##
##SUBSECTION Over a base field
##- The dependent variables _q_ and _p_ lie in the field.
##
##> Indets (field (generators = [p, q]), R);
##< {p, q}
##ENDSUBSECTION
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing", "LeadingDerivative", "RosenfeldGroebner", "FieldElement"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "FieldElement" : Help:DifferentialAlgebra0[FieldElement]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]
##- "Parameters" : Help:DifferentialAlgebra0[Parameters]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "PowerSeriesSolution" : Help:DifferentialAlgebra0[PowerSeriesSolution]


# EXPORTED
#
# Returns the indets occurring in a differential polynomial
#
# Indets (differential ring)
# Indets (ideal)
# Indets (polynomial | list of polynomials, differential ring)

Indets := proc
    ({selection::{identical(:-_DEPVARS_,:-_DERIVS_,:-_INDEPVARS_,
			       :-_ALLVARS_, :-_PARAMS_),
		     specfunc(nonnegative,:-iniconds)}:=':-_DERIVS_'},
     {fullset::boolean:=false},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local result, eqns, gens, rels, drideal, a_table, innot, outnot;

    if _nrest = 1 then
# Indets (ideal | dring)
        if not (IsDifferentialRing (_rest[1]) or 
                            IsIdeal (_rest[1], ':-possibly_empty' = true)) then 
            error _ERRIDRNG_, "first", convert (_rest[1], ':-string')
	elif type (selection, ':-specfunc(:-nonnegative,:-iniconds)') and
					IsDifferentialRing (_rest[1]) then
	    error _ERRSEL_, convert (_rest[1], ':-string')
        end if;
        drideal := _rest[1];
        if drideal = [] then
            return {};
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (drideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (drideal));
# The diff notation is not compatible with the fullset option combined
# with selection = derivatives | allvars
	if IsIdeal (drideal) and outnot = ':-_DIFF_' and fullset and
		member (selection, [':-_DERIVS_', ':-_ALLVARS_']) then
	    error _ERRFSDF_;
        end if;
        try
            if type (drideal, ':-table') then
                result := IndetsSafe (drideal, selection, drideal, 
                            fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := map (IndetsSafe, drideal, selection, drideal[1],
                            fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if;
        catch:
            error
        end try;
    elif type (selection, ':-specfunc(:-nonnegative,:-iniconds)') then
	error _ERRSEL_, convert (_rest[1], ':-string');
    elif _nrest = 2 then
# Indets (field (generators = ..., relations = ...), dring)
# Indets (polynomial | list of polynomials, dring)
        if not IsIdealDRing (_rest[2]) then
            error _ERRIDRNG_, "second", convert (_rest[2], ':-string');
        elif type (_rest[2], ':-list') then
            a_table := _rest[2,1];
        else
            a_table := _rest[2];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
# eqns := the list of polynomials to process.
# In the case of a basefield, eqns := the list of generators of the field
	if op (0, _rest[1]) = ':-field' then
	    try
		gens, rels := BaseField (_rest[1]);
	        eqns := BaseFieldGeneratorsSafe 
				(gens, rels, a_table, innot, innot,
				 _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	    catch:
		error
	    end try;
	else
	    try
                eqns := ToExpandedRatfracList (_rest[1])
	    catch:
		error _ERRFRLRT_, "first", convert (_rest[1], ':-string')
	    end try
	end if;
# Let us go.
        try
            result := IndetsSafe (eqns, selection, a_table, 
                            fullset, innot, outnot, 
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_)
        catch:
            error
        end try;
    else
        error _ERRNARGS_, "One or two", _nrest;
    end if;
    return result;
end proc:

# Subfunction of Indets

IndetsSafe := proc (list_or_table, selection, a_table, 
                    fullset, innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_) 
    local arguments, ordre, result;
    if type (selection, ':-specfunc (:-nonnegative, :-iniconds)') then
	ordre := op (1, selection);
	arguments := "AllDerivatives" 
		(ConvertString (a_table [':-_FREEZESEQ_'], ordre), 
		 list_or_table);
    else
	arguments := "Indets" 
                 (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table), 
                  ConvertString (a_table [':-_FREEZESEQ_'], selection), 
                  fullset, 
		  a_table);
    end if;
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:

# Subfunction of Indets

BaseFieldGeneratorsSafe := proc (gens, rels, a_table, innot, outnot,
						_TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "BaseFieldGenerators"
			(ConvertString (a_table [':-_FREEZESEQ_'], gens), 
			 ConvertString (a_table [':-_FREEZESEQ_'], rels), 
			 a_table);
    arguments := arguments,
	ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:
