##PROCEDURE(help) DifferentialAlgebra0[Inequations]
##HALFLINE        returns the initials and separants of a regular differential chain
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- Inequations ('ideal', 'opts')
##
##PARAMETERS
##
##- 'ideal' : a differential polynomial ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- If 'ideal' is a regular differential chain, then, 
##  the function call `Inequations` ('ideal') returns the list of
##  initials and separants of 'ideal'. The returned differential
##  polynomials are primitive: the gcd of their numerical coefficients
##  is equal to _1_.
##  Numeric differential polynomials are removed from the list.
##
##- If 'ideal' is a list, the function call `Inequations` ('ideal') 
##  a list of lists of initials and separants.
##
##- The inequations of a regular differential chain provide constraints
##  that expansion points and initial values of power series solutions 
##  must satisfy. See "PowerSeriesSystem".
##
##- The differential ideal defined by a regular differential chain
##  is the *saturation* of the differential ideal generated by the
##  equations of the chain, by the multiplicative family generated
##  by the inequations.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Inequations(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Inequations](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x, y], blocks = [[v,u]]);
##
##> ideal := RosenfeldGroebner([x*u[x]-u, 3*v[y]^2 - 1], R);
##
##>(quote) Initial(ideal), Separant(ideal);
##< [[x, 3]], [[x, 6*v[y]]]
##
##> Inequations(ideal);
##< [[x, v[y]]]
##
##SEEALSO
##- "DifferentialAlgebra0", "Initial", "Separant", "Equations", "PowerSeriesSolutions"
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "PowerSeriesSystem" : Help:DifferentialAlgebra0[PowerSeriesSystem]
##- "PowerSeriesSolution" : Help:DifferentialAlgebra0[PowerSeriesSolution]
##- "Equations" : Help:DifferentialAlgebra0[Equations]
##- "Initial" : Help:DifferentialAlgebra0[Initial]
##- "Separant" : Help:DifferentialAlgebra0[Separant]

Inequations := proc 
    (ideal,
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local result, independent;

# Once upon a time, independent = yes|no was an option.

    independent := ':-yes';
    if not IsIdeal (ideal) then
	error _ERRIDEAL_, "first", convert (ideal, ':-string')
    elif ideal = [] then
	result := []
    elif type (ideal, ':-table') then
	try
	    result := InitialsAndSeparants 
		(ideal, independent, _OUTPUT_, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	catch:
	    error
	end try;
    else
	try
	    result := map (InitialsAndSeparants, ideal, independent,
				_OUTPUT_, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
        catch:
            error
        end try;
    end if;
    return result;
end proc:

# Local to Inequations

# Returns the list of the initials and separants of regchain
# Independent initials and separants are removed.
# Duplicates are removed.
# Leading coefficients are normalized.

InitialsAndSeparants := proc 
	(regchain, independent, _output_, _timeout_, _memout_, _cellsz_)

    local arguments, X, L, M, i, j, P, koeffs, terms, u;

    X := Indets (regchain, 
		 ':-selection' = ':-_ALLVARS_', ':-_IONOT_'=':-_TJET_');
# The zero ideal
    if X = {} then
	return [];
    end if;
    u := MinRankElement (X, regchain, ':-_IONOT_'=':-_TJET_');

# tjet is the safer notation for polynomial manipulations.

    arguments := ':-_IONOT_' = ':-_TJET_', ':-_TIMEOUT_' = _timeout_,
			':-_MEMOUT_' = _memout_, ':-_CELLSZ_' = _cellsz_;
    L := [ op (Initial (regchain, arguments)),
	   op (Separant (regchain, arguments)) ];

# Normalize the numerical coefficients

    M := NULL;
    for i from 1 to nops (L) do
	if not type (L[i], ':-numeric') then
	    koeffs, terms := Coeffs (L[i], u, regchain, arguments);
	    P := add ((koeffs [j]/koeffs [1])*terms [j], 
						j = 1 .. nops (koeffs));
	    P := numer (P);
	    if not type (P, ':-numeric') then
		M := M, numer (P);
	    end if;
	end if
    end do;

# remove duplicates.

    M := { M };

# Changing the notation and managing the independent parameter.

    arguments := ':-_INPUT_' = ':-_TJET_', ':-_OUTPUT_' = _output_, 
							arguments [2..-1];
    if independent = ':-yes' then
        M := NormalForm ([ op (M) ], regchain, arguments);
    else
	M :=  Equations ([ op (M) ], ':-order' >= 0, regchain, arguments);
    end if;

    return M;
end proc:
