##PROCEDURE(help) DifferentialAlgebra0[Initial]
##HALFLINE        returns the initial of a differential polynomial
##AUTHOR          F. Boulier
##DATE            July 2009
##
##CALLINGSEQUENCE
##- `Initial`('ideal', 'opts')
##- `Initial`('p', 'R', 'opts')
##- `Initial`('L', 'R', 'opts')
##
##PARAMETERS
##- 'ideal' : a differential ideal
##- 'p'     : a differential rational fraction
##- 'L'     : a list or a set of differential rational fractions
##- 'R'     : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Initial`('p','R') returns the initial 
##  of 'p' with respect to the ranking of 'R', or of its embedding
##  polynomial ring, if 'R' is an ideal.
##  It is assumed that 'p' is non-numeric.
##
##- The function call `Initial`('L','R') returns the list or the set of the
##  initials of the elements of 'L' with respect to the ranking of 'R'.
##
##- If 'ideal' is a regular differential chain, the function call 
##  `Initial`('ideal') returns the list of the initials of the chain 
##  elements. If 'ideal' is a list of regular differential chains, 
##  the function call `Initial`('ideal') returns a list of lists of initials.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Initial(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Initial](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call 
##   `Initial`('ideal'), applies the function also over the differential 
##   polynomials which state that the derivatives of the parameters 
##   are zero. Default value is `false`.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with (DifferentialAlgebra0):
##> R := DifferentialRing (derivations=[x,y], blocks=[[v,u],p], parameters=[p]);
##
##> Initial (u[x,y]*v[y]-u+p, R);
##< v[y]
##
##> ideal := RosenfeldGroebner ([u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]], R);
##
##> Equations (ideal[1]);
##< [u[y]^2-2*u, u[x]^2-4*u, 4*v[y]*u-u[x]*u[y]*u+u[x]*u[y]*p, v[x,x]-u[x]]
##
##> Initial (ideal[1]);
##< [1, 1, 4*u, 1]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative", "LeadingCoefficient", "Inequations"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "LeadingCoefficient" : Help:DifferentialAlgebra0[LeadingCoefficient]
##- "Inequations" : Help:DifferentialAlgebra0[Inequations]


# EXPORTED

Initial := proc 
   ({fullset::boolean:=false},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local ideal, a_table, eqns, innot, outnot, result;

    if _nrest = 1 then
# Initial (ideal)
        if not IsIdeal (_rest[1]) then
            error _ERRIDEAL_, "first", convert (_rest[1], ':-string');
        end if;
        ideal := _rest[1];
        if ideal = [] then
            return [];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (ideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (ideal));
        try
            if type (ideal, ':-list') then
                result := map (LeadingCoefficientSafe, ideal, 0,
                            ideal[1], fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := LeadingCoefficientSafe (ideal, 0,
                            ideal, fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if;
        catch:
            error
        end try;
    elif _nrest = 2 then
# Initial (poly, differential ring)
# Initial (list(poly), differential ring)
        if not IsIdealDRing (_rest[_nrest]) then
            error _ERRIDRNG_, "second", convert (_rest[2], ':-string');
        elif type (_rest[_nrest], ':-list') then
            a_table := _rest[_nrest][1];
        else
            a_table := _rest[_nrest];
        end if;
	try
            eqns := ToExpandedRatfracList (_rest[1])
	catch:
	    error _ERRPLPL_, "first", convert (_rest[1], ':-string')
	end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        try
            result := LeadingCoefficientSafe (eqns, 0, 
                            a_table, fullset, innot, outnot, 
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        catch:
            error
        end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    else
        error _ERRNARGS_, "One or two", _nrest;
    end if;
    return result;
end proc:
