##PROCEDURE(help) DifferentialAlgebra0[IsOrthonomic]
##HALFLINE	decides if a regular differential chain is orthonomic
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- IsOrthonomic ('ideal','opts')
##
##PARAMETERS
##
##- 'ideal' : a differential ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- If 'ideal' is a regular differential chain, the function call 
##  `IsOrthonomic` ('ideal') returns `true` if 'ideal' is orthonomic
##  (in the generalized sense, see below) else, it returns `false`.
##
##- An orthonomic regular differential chain defines a *prime*
##  differential ideal. The "Pardi" function automatically recognizes
##  this property.
##
##- If 'ideal' is a list or regular differential chains, the function
##  call `IsOrthonomic` ('ideal')  returns a list of ::boolean::.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `IsOrthonomic(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[IsOrthonomic](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `definition` = `strict` or `generalized`.
##  A regular differential chain is orthonomic, in the *generalized* sense,
##  if its elements have degree _1_ in their leading derivatives.
##  It is orthonomic, in the *strict* sense, if it is orthonomic in
##  the generalized sense, and, the initials of its elements belong
##  to the base field of the equations.
## 
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##
##> R := DifferentialRing(derivations = [x], blocks = [u]);
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u], R);
##
##> Equations(ideal);
##< [[u[x]^2-4*u], [u]]
##
##> IsOrthonomic(ideal);
##< [false, true]
##
##> IsOrthonomic(ideal, definition=strict);
##< [false, true]
##
##> ideal := RosenfeldGroebner([u[x]*u[x,x]+u], R);
##
##> Equations(ideal);
##< [[u[x]*u[x, x] + u], [u]]
##
##> IsOrthonomic(ideal);
##< [true, true]
##
##> IsOrthonomic(ideal, definition=strict);
##< [false, true]
##
##SEEALSO
##- "DifferentialAlgebra0", "Pardi"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Pardi" : Help:DifferentialAlgebra0[Pardi]


# EXPORTED

IsOrthonomic := proc 
    (ideal,
     {definition::identical(:-strict,:-generalized):=':-generalized'},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local result, generalizeddef;

    if not IsIdeal (ideal) then
	error _ERRIDEAL_, "first", convert (ideal, ':-string');
    else
	generalizeddef := evalb (definition = ':-generalized');
	try
	    if type (ideal, ':-table') then
		result := IsOrthonomicSafe 
			(ideal, generalizeddef, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	    else
	        result := map (IsOrthonomicSafe, ideal, generalizeddef, 
					_TIMEOUT_, _MEMOUT_, _CELLSZ_);
	    end if
	catch:
	    error
	end try;
    end if;
    return result;
end proc:

# Local to IsOrthonomic

IsOrthonomicSafe := proc 
		(regchain, generalizeddef, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "IsOrthonomic" (regchain, generalizeddef);
    arguments := arguments,
	ToBMIOptionSequence 
			(':-_JET_', ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
end proc:

