##PROCEDURE(help) DifferentialAlgebra0[IsReduced]
##HALFLINE        decides if a differential polynomial is reduced with respect to a regular differential chain
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- IsReduced ('p','regchain','opts')
##- IsReduced ('L','regchain','opts')
##- IsReduced ('p','redset','R','opts')
##- IsReduced ('L','redset','R','opts')
##
##PARAMETERS
##
##- 'p' : a differential polynomial
##- 'L' : a list or a set of differential polynomials
##- 'regchain' : a regular differential chain
##- 'redset' : a polynomial or a list or a set of differential polynomials
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `IsReduced` ('p','regchain') returns `true`
##  if 'p' is fully reduced (see below) with respect to 'regchain', else `false`.
##  The differential polynomial 'p' is regarded as an element of
##  the embedding differential polynomial ring of 'regchain'.
##
##- The function call `IsReduced` ('L','regchain') returns a list
##  or a set of ::boolean::.
##
##- The function call `IsReduced` ('p','redset','R','opts') returns `true`
##  if 'p' is fully reduced with respect to each element
##  of 'redset', else `false`. 
##  The elements of 'redset' must depend on at least, one *derivative* 
##  and have integer coefficients. All the differential polynomials are 
##  regarded as elements of 'R', or, of its embedding ring, if 'R' is an ideal.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `IsReduced(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[IsReduced](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `reduction` = `full`, `partial` or `algebraic`.
##  Indicates the type of reduction concerned by the test.
##  Depending on the type of reduction, the function returns `true`
##  if _p_ satisfies the following properties:
##-- `algebraic`. For each leading rank _v^d_ of 'regchain', or, of 'redset',
##  _p_ has degree less than _d_ in _v_.
##-- `partial`. For each leading derivative _v_ of 'regchain', or, of 'redset',
##  no *proper* derivative of 'v' occurs in _p_.
##-- `full`. For each leading rank _v^d_ of 'regchain', or, of 'redset',
##  no *proper* derivative of _v_ occurs in _p_, and, moreover,
##  _p_ has degree less than _d_ in _v_.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p', 'L' and 'redset'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x], blocks = [u]);
##
##> ideal := PretendRegularDifferentialChain([u[x]^2-4*u], R);
##
##> IsReduced([u[x,x], u[x]^3, u[x]], ideal);
##< [false, false, true]
##
##> IsReduced([u[x,x], u[x]^3, u[x]], reduction=partial, ideal);
##< [false, true, true]
##
##> IsReduced([u[x,x], u[x]^3, u[x]], reduction=algebraic, ideal);
##< [true, false, true]
##
##> IsReduced(u[x,x,x], [u[x,x], u[x]^2-4*u], R);
##< false
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative", "LeadingRank", "DifferentialPrem", "NormalForm"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "DifferentialPrem" : Help:DifferentialAlgebra0[DifferentialPrem]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]

# EXPORTED

IsReduced := proc
    (eqns,
     {reduction::identical(:-_FULLY_,:-_PARTIALLY_,:-_ALGBRCALLY_):=:-_FULLY_},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

     local result, leqns, innot, a_table, redset;

    if _nrest = 1 then
	a_table := _rest[_nrest];
        if not IsRegularDifferentialChain (a_table) then
	    error _ERRREGC_, "second", convert (a_table, ':-string');
        else
	    try
	        leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
	    catch:
		error _ERRPLPL_, "first", convert (eqns, ':-string')
	    end try;
	    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
	    try
	        result := IsReducedSafe (leqns, reduction, a_table, innot,
                                         _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            catch:
                error
            end try;
            result := FromExpandedRatfracList (result, eqns);
        end if
    elif _nrest <> 2 then
	error _ERRNARGS_, "Two or three", convert (_nrest+1, ':-string');
    else
	redset := _rest [1];
	if not IsIdealDRing (_rest[_nrest]) then
	    error _ERRIDRNG_, "third", convert (_rest[_nrest], ':-string');
	elif type (_rest[_nrest], ':-list') then
	    a_table := _rest[_nrest] [1];
	else
	    a_table := _rest[_nrest];
	end if;
	try
	    leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "first", convert (eqns, ':-string')
	end try;
	try
	    redset := ToExpandedRatfracList (redset, ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "second", convert (redset, ':-string')
	end try;
	innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
	try
	    result := IsReduced2Safe (leqns, reduction, redset, a_table, 
					 innot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	catch:
	    error
	end try;
	result := FromExpandedRatfracList (result, eqns);
    end if;
    return result
end proc:

# Local to IsReduced

IsReducedSafe := proc (eqns, reduction, regchain, innot, 
					_TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "IsReduced" 
			(ListOrTableToBMI (regchain [':-_FREEZESEQ_'], eqns), 
			 ConvertString (regchain [':-_FREEZESEQ_'], reduction), 
			 regchain);
    arguments := arguments,
        ToBMIOptionSequence (innot, ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
end proc:

# Local to IsReduced

IsReduced2Safe := proc (eqns, reduction, redset, a_table, innot,
					 _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "IsReduced"
			(ListOrTableToBMI (a_table [':-_FREEZESEQ_'], eqns), 
			 ConvertString (a_table [':-_FREEZESEQ_'], reduction), 
			 ConvertString (a_table [':-_FREEZESEQ_'], redset), 
			 a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:

