##PROCEDURE(help) DifferentialAlgebra0[LeadingDerivative]
##HALFLINE        returns the leading derivative of a differential polynomial
##AUTHOR          F. Boulier
##DATE            July 2009
##
##CALLINGSEQUENCE
##- `LeadingDerivative`('ideal', 'opts')
##- `LeadingDerivative`('p', 'R', 'opts')
##- `LeadingDerivative`('L', 'R', 'opts')
##
##PARAMETERS
##- 'ideal' : a differential ideal
##- 'p'     : a differential rational fraction
##- 'L'     : a list or a set of differential rational fractions
##- 'R'     : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `LeadingDerivative`('p','R') returns the 
##  leading derivative of 'p' with respect to the ranking of 'R'
##  or the one of its embedding polynomial ring if 'R' is an ideal.
##  The polynomial 'p' is assumed to be non-numeric. It may, however,
##  only depend on independent variables. In this case, the leading
##  independent variable is returned.
##
##- The function call `LeadingDerivative`('L','R') returns the list or 
##  the set of the leading derivatives of the elements of 'L' with respect to 
##  the ranking of 'R'.
##
##- If 'ideal' is a regular differential chain, the function call 
##  `LeadingDerivative`('ideal') returns the list of the 
##  leading derivatives of the chain elements. 
##  If 'ideal' is a list of regular differential chains, 
##  the function call `LeadingDerivative`('ideal') returns a list 
##  of lists of leading derivatives.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `LeadingDerivative(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[LeadingDerivative](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call 
##   `LeadingDerivative`('ideal'), applies the function also over 
##   the differential 
##   polynomials which state that the derivatives of the parameters 
##   are zero. Default value is `false`.
##   This option is incompatible with the `diff` output notation.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[x,y], blocks=[[v,u],p], parameters=[p]);
##
##> LeadingDerivative(u[x,y]*v[y]-u+p, R);
##< u[x,y]
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]], R);
##
##> Equations(ideal[1]);
##< [u[y]^2-2*u, u[x]^2-4*u, 4*v[y]*u-u[x]*u[y]*u+u[x]*u[y]*p, v[x,x]-u[x]]
##
##> LeadingDerivative(ideal[1]);
##< [u[y], u[x], v[y], v[x, x]]
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]


# EXPORTED
#
# LeadingDerivative (poly, differential ring)
# LeadingDerivative (list(poly), differential ring)
# LeadingDerivative (ideal)
#
# Returns the leading derivatives

LeadingDerivative := proc 
       ({fullset::boolean:=false},
        {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_TIMEOUT_::nonnegative:=0},
        {_MEMOUT_::nonnegative:=0},
        {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})
    local ideal, a_table, eqns, innot, outnot, result;
    if _nrest = 1 then
# LeadingDerivative (ideal)
        if not IsIdeal (_rest[_nrest]) then
            error _ERRIDEAL_, "first", convert (_rest[1], ':-string');
        end if;
        ideal := _rest[_nrest];
        if ideal = [] then
            return [];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (ideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (ideal));
	if outnot = ':-_DIFF_' and fullset then
            error _ERRFSDF_;
        end if;
        try
            if type (ideal, ':-list') then
                result := map (LeadingDerivativeSafe, ideal, ideal[1], 
                            fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := LeadingDerivativeSafe (ideal, ideal, 
                            fullset, innot, outnot, 
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if
        catch:
            error
        end try
    elif _nrest = 2 then
# LeadingDerivative (polynomial | list(polynomial), dring)
        if not IsIdealDRing (_rest[_nrest]) then
            error _ERRIDRNG_, "second", convert (_rest[2], ':-string');
        elif type (_rest[_nrest], ':-list') then 
            a_table := _rest[_nrest] [1];
        else
            a_table := _rest[_nrest];
        end if;
        try
            eqns := ToExpandedRatfracList (_rest[1])
        catch:
            error _ERRPLPL_, "first", convert (_rest[1], ':-string')
        end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        try
            result := LeadingDerivativeSafe (eqns, a_table, 
                            false, innot, outnot, 
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        catch:
            error
        end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    else
        error _ERRNARGS_, "One or two", _nrest;
    end if;
    return result
end proc:

LeadingDerivativeSafe := proc (list_or_table, a_table, 
                            fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "LeadingDerivative" 
		(ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table), 
		 fullset, a_table);
    arguments := arguments,
        ToBMIOptionSequence 
            (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:

