##PROCEDURE(help) DifferentialAlgebra0[MinRankElement]
##HALFLINE        returns the element of minimal rank in a list 
##AUTHOR          F. Boulier
##DATE            July 2009
##
##CALLINGSEQUENCE
##- `MinRankElement`('L', 'R', 'pos', 'opts')
##
##PARAMETERS
##- 'L'              : a non-empty list or a set of differential polynomials
##- 'R'              : a differential polynomial ring or ideal
##- 'pos' (optional) : a name
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `MinRankElement`('L','R','pos') returns the 
##  element of minimal rank  
##  occurring in the list 'L'. Ranks are compared with respect to
##  the ranking of 'R', or its embedding polynomial ring, if 'R' is an
##  ideal. If 'pos' is present, it is assigned the index of the element
##  in 'L'.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `MinRankElement(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[MinRankElement](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[x,y], blocks=[[v,u],p], parameters=[p]);
##
##> L := [u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]]:
##< [u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]]
##
##> MinRankElement(L, R, 'pos');
##< u[x]^2-4*u
##
##> L[pos];
##< u[x]^2-4*u
##
##- The following command returns the lowest dependent variable of _R_,
##  including *parameters*.
##
##> MinRankElement(Indets (R), R);
##< p
##
##- The following command returns the lowest dependent variable of _R_,
##  among the ones which are not *parameters*.
##
##> MinRankElement (Indets (R) minus Indets (R, selection = parameters), R);
##< u
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingRank", "MaxRankElement", "SortByRank"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "MaxRankElement" : Help:DifferentialAlgebra0[MaxRankElement]
##- "SortByRank" : Help:DifferentialAlgebra0[SortByRank]

# EXPORTED
#
# MinRankElement (list(polynomials), differential ring)
#
# Returns the element of minimal rank

MinRankElement := proc (eqns::{list,set}, drideal, pos:=':-_U_',
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local result, indx, arguments, a_table, leqns, innot, outnot;

    if not IsIdealDRing (drideal) then
        error _ERRIDRNG_, "second", convert (_rest[2], ':-string');
    elif type (drideal, ':-list') then
        a_table := drideal[1];
    else
        a_table := drideal;
    end if;
    try
        leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
    catch:
        error _ERRNLPOL_, "first", convert (eqns, ':-string')
    end try;
    arguments := "MinRankElement" 
			(ConvertString (a_table [':-_FREEZESEQ_'], leqns), 
			 a_table);
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result, indx := 
		BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    if pos <> ':-_U_' then
        pos := indx;
    end if;
    return result;
end proc:
