##PROCEDURE(help) DifferentialAlgebra0[NormalForm]
##HALFLINE        computes normal forms modulo regular differential chains
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- NormalForm ('p', 'ideal', 'opts')
##- NormalForm ('L', 'ideal', 'opts')
##
##PARAMETERS
##- 'p' : a differential rational fraction
##- 'ideal' : a differential polynomial ideal or ring
##- 'L' : a list or a set of differential rational fractions
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- Assume 'ideal' is a regular differential chain. Denote _R_ its
##  embedding differential polynomial ring and _I_ the differential ideal 
##  that it defines.
##
##- The function call `NormalForm` ('p', 'ideal') returns a rational
##  differential fraction, which is a normal form of 'p' in _R_/_I_.
##  See "DifferentialAlgebra0" for details.
##
##- In the case of rational differential fractions, the computation may fail,
##  making the function raise an error. These cases can completely be 
##  handled by using the `casesplit`=`true` option, documented below.
##
##- The function call `NormalForm` ('L', 'ideal') returns the list or the 
##  set of the normal forms of the elements of 'L' with respect to 'ideal'.
##
##- If 'ideal' is a list of regular differential chains, then the function call
##  `NormalForm` ('p', 'ideal') returns the list of the normal forms of 'p'
##  with respect to all the chains, and, the function call 
##  `NormalForm` ('L', 'ideal') returns the list of the normal forms of 'L'
##  with respect to all the chains.
##
##- If 'ideal' is a differential polynomial ring, then the function call
##  `NormalForm` ('p', 'ideal') returns 'p', and, the function call
##  `NormalForm` ('L', 'ideal') returns 'L'. 
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `NormalForm(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[NormalForm](...)`.
##
##SECTION Handling splittings
##
##- The option `casesplit` = `true` permits to handle the cases which 
##  raise an error. There are, in this case, some restrictions on the 
##  function arguments: 
##
##-- The first argument, 'p', is restricted to a single rational differential 
##   fraction (lists and sets are not allowed).
##-- The second argument, 'ideal' is restricted to a single regular 
##   differential chain (lists of chains are not allowed). 
##
##- The function call `NormalForm` ('p', 'ideal', `casesplit`=`true`)
##  returns a list 
##  [[[_F[1]_, _C[1]_], `...`, [_F[n]_, _C[n]_]],
##   [_C[n+1]_, `...`, _C[n+p]_]].
##  The _C[i]_ are regular differential chains.
##  The intersection of the ideals that they define is equal to _I_, in the 
##  differential case, and, more generally, whenever _I_ is *radical*.
##
##-- The denominator of 'p' is regular modulo the ideals defined by
##   the chains _C[1]_, `...`, _C[n]_ and, for each _i_ in the range _[1,n]_,
##   the rational differential fraction _F[i]_ is the normal form
##   of 'p' modulo _C[i]_.
##-- The denominator of 'p' is zero modulo the ideals defined by
##   the chains _C[n+1]_, `...`, _C[n+p]_: these cases correspond to cases
##   where the normal form does not exist.
##  
##OPTIONS
##
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'ideal' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##TEST
## VerifyTools:-AddVerification(as_global = ((a,b) -> evalb(eval(convert(a-b ,`global`)) = 0)));
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [t], blocks = [u]);
##
##- The normal form, with respect to a differential ring.
##
##> NormalForm((u-1)*(u+1), R);
##< u^2-1
##
##> ideal := RosenfeldGroebner([u[t]^2-4*u], R);
##
##> Equations(ideal);
##< [[u[t]^2-4*u], [u]]
##
##- The normal forms of _u[t]_ with respect to each regular differential chain.
##
##> NormalForm(u[t], ideal);
##< [u[t], 0]
##
##- This example shows that the function can be used to compute
##  normal forms of rational differential fractions, thus, to compute
##  algebraic inverses of rational differential fractions, modulo the
##  differential ideals defined by regular differential chains.
##
##> F := u/(u[t]+u[t,t]);
##< u/(u[t]+u[t,t])
##
##> nf_F := NormalForm (F, ideal[1]);
##< (u[t]*u-2*u)/(4*u-4)
##
##> nf_one_over_F := NormalForm (1/F, ideal[1]);
##< (u[t]+2)/u
##
##> one := normal (nf_F * nf_one_over_F);
##< 1/4*(u[t]-2)*(u[t]+2)/(u-1)
##
##> NormalForm (one, ideal[1]);
##< 1
##
##- This example illustrates the `casesplit` = `true` option.
##
##> ideal := PretendRegularDifferentialChain ([u[t]^2-u^2], R);
##
##- The denominator of the input rational differential fraction 
##  is zero modulo 'ideal'.
##  Only one regular differential chain is returned, in the second list.
##
##> res := NormalForm (1/(u[t]^2-u^2), ideal, casesplit=true);
##
##> Equations (res [2,1]);
##< [u[t]^2-u^2]
##
##- The denominator of the input rational differential fraction
##  is a zero-divisor modulo 'ideal'.
##  The input chain is thus split into two regular differential chains.
##  The normal form can be computed modulo the differential ideal defined by the
##  first one. The denominator of the input rational differential fraction
##  is zero modulo the differential ideal defined by the second one.
##
##> res := NormalForm (1/(u[t]-u), ideal, casesplit=true);
##
##> Equations (res [1, 1, 2]);
##< [u[t] + u]
##
##> Equations (res [2, 1]);
##< [u[t] - u]
##
##SEEALSO
##- "DifferentialAlgebra0", "RosenfeldGroebner", "ReducedForm", "DifferentialPrem"
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "PretendRegularDifferentialChain" : Help:DifferentialAlgebra0[PretendRegularDifferentialChain]
##- "ReducedForm" : Help:DifferentialAlgebra0[ReducedForm]
##- "DifferentialPrem" : Help:DifferentialAlgebra0[DifferentialPrem]


# EXPORTED

NormalForm := proc (eqns,
    drideal,
    {casesplit::boolean:=false},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'},$)

    local arguments, result, leqns, innot, outnot, regchain;

    if IsDifferentialRing (drideal) then
# NormalForm ( ..., ring )
# Just equivalent to Differentiate ( ..., 1, ring )
        try
            leqns := ToExpandedRatfracList (eqns)
        catch:
            error _ERRRLRT_, "first", convert (eqns, ':-string')
        end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (drideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (drideal));
	try
	    result := DifferentiateSafe (leqns, [], drideal, false,
				innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	catch:
	    error
	end try;
	result := FromExpandedRatfracList (result, eqns);
    elif not IsIdeal (drideal, ':-possibly_empty' = true) then
        error _ERRIDRNG_, "second", convert (drideal, ':-string');
    elif casesplit then
# Former NormalFormExt function
# Returns a complicated structure ...
	if type (eqns, '{:-list,:-set}') then
	    error _ERRNFCSE_, convert (eqns, ':-string');
	elif not IsRegularDifferentialChain (drideal) then
	    error _ERRNFCSR_, convert (drideal, ':-string');
	end if;
        try
            leqns := ToExpandedRatfracList (eqns)
        catch:
            error _ERRNFCSE_, "first", convert (eqns, ':-string')
        end try;
	regchain := drideal;
	innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (regchain));
	outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (regchain));
	arguments := "NormalFormHandlingExceptions"
			(ConvertString (regchain [':-_FREEZESEQ_'], leqns[1]),
			 regchain);
	arguments := arguments, ToBMIOptionSequence 
			(innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	try
            result := 
		BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
	catch:
            error
	end try;
	outnot := NotationSafe (regchain);
	result := NormalFormExtToRegchain 
					(result, outnot, 
					 regchain [':-_FREEZESEQ_'], 
					 regchain [':-_UNNUMBERSEQ_']);
    elif drideal = [] then
# The unit ideal
# Normal forms are zero
        if type (eqns, '{:-list,:-set}') then
            return map (x -> 0, eqns)
        else
            return 0
        end if
    else
# The basic case
        try
            leqns := ToExpandedRatfracList (eqns)
        catch:
            error _ERRRLRT_, "first", convert (eqns, ':-string')
        end try;
	innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (drideal));
	outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (drideal));
	if type (drideal, ':-list') then
	    regchain := drideal [1];
            arguments := "NormalForm" 
			(ConvertString (regchain [':-_FREEZESEQ_'], leqns), 
			 op (drideal))
	else
	    regchain := drideal;
            arguments := "NormalForm" 
			(ConvertString (regchain [':-_FREEZESEQ_'], leqns), 
			 drideal)
	end if;
        arguments := arguments, 
            ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        try
            result := BMIBladEvalSafe 
				(regchain [':-_UNNUMBERSEQ_'], arguments);
        catch:
            error
        end try;
        if type (drideal, ':-table') then
            result := map (op, result)
        end if;
        result := FromExpandedRatfracList (result, eqns);
    end if;
    return result
end proc:

# Local to NormalForm
# Function called when casesplit=true only. Twolists has the form:
# [[[nf_1, C_1], ..., [nf_n, C_n]], [C_{n+1}, ..., C_{n+p} ]]
#
# Transforms the C_i into regchains.

NormalFormExtToRegchain := proc (twolists, outnot, freezeseq, unnumberseq, $)
    return
        [ map (proc (pair, fseq, useq) 
		   [pair [1], ToRegchain (pair [2], outnot, fseq, useq)] 
	       end proc,
               twolists [1], freezeseq, unnumberseq),
          map (ToRegchain, twolists [2], outnot, freezeseq, unnumberseq)
        ]
end proc;
