##PROCEDURE(help) DifferentialAlgebra0[Notation]
##HALFLINE        returns the default notation of its argument
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `Notation`('R')
##
##PARAMETERS
##- 'R' : a differential polynomial ring or ideal
##
##DESCRIPTION
##- The function call `Notation`('R') returns the default notation
##  of 'R',
##  which is the one which is applied by the calls to the package
##  functions which have 'R' as an argument, 
##  for parsing arguments and printing results,
##  unless an explicit other notation is provided to the function call
##  (see the `input`, `output` and `notation` options of most of the
##  package functions).
##
##- The returned notation is one of the keywords `jet`, `tjet`, `diff`
##  and `Diff`. Let 'D' = [_x[1]_, `...`, _x[p]_] denote the list of
##  the independent variables, which provide the derivations.
##
##- In the `tjet` notation, a derivative _v_ of a dependent
##  variable _u_, appears as an ::indexed:: having the form: 
##  _u_ [_x[1]_ `$` _i[1]_, `...`, _x[p]_ `$` _i[p]_] where the _i[k]_
##  stand for nonnegative integers.
##  The *derivative* _v_ is then equal to _u_, differentiated _i[1]_ times
##  with respect to _x[1]_, `...`, _i[p]_ times with respect to _x[p]_.
##
##- The `jet` notation is similar to `tjet` except for order zero *derivatives*,
##  which are displayed as plain dependent variables.
##
##-  For parsing arguments, the notations `jet` and `tjet` are equivalent,
##  and, the order of the independent variables in the list does not matter.
##
##- In the `diff` notation, a derivative _v_ of a dependent
##  variable _u_, appears as in the standard MAPLE notation for *derivatives*:
##  `diff` (_u_ (_x[1]_, `...`, _x[p]_), _x[1]_ `$` _i[1]_, 
##					`...`, _x[p]_ `$` _i[p]_).
##
##- The independent variables, _u_ depends on, are listed in the same order
##  as in the list 'D'. Customizing this order is possible
##  by means of the `parameters` argument of the "DifferentialRing"
##  function.
##
##- The `Diff` notation is analogue to the `diff` one.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Notation(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Notation]`.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [u]);
##
##> Notation(R);
##< jet
##
##> Differentiate(u, x, y, x, R);
##< u[x, x, y]
##
##> Differentiate(u, R);
##< u
##
##> Differentiate(u, R, output = tjet);
##< u[]
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u], R, output = diff);
##
##> Notation(ideal);
##< diff
##
##> Equations(ideal);
##< [[diff(u(x,y),x)^2-4*u(x,y)], [u(x,y)]]
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing", "Ranking", "Parameters"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]
##- "Ranking" : Help:DifferentialAlgebra0[Ranking]
##- "Parameters" : Help:DifferentialAlgebra0[Parameters]


# EXPORTED
#
# Returns the _NOTATION_ entry of a differential ring, a regular chain
# or a non-empty list of regular chains.

Notation := proc (drideal)
    local result;
    if IsIdeal (drideal, ':-possibly_empty'=false) 
                            or IsDifferentialRing (drideal) then
        result := NotationSafe (drideal);
    else
        error _ERRIDRNG_, "first", convert (drideal, ':-string');
    end if;
    return result;
end proc:

# Subfunction of Notation

NotationSafe := proc (drideal)
    local result;
    if type (drideal, ':-list') then
        result := drideal [1][':-_NOTATION_'];
    else
        result := drideal [':-_NOTATION_'];
    end if;
    return result;
end proc:
