##PROCEDURE(help) DifferentialAlgebra0[Parameters]
##HALFLINE        returns the list of *parameters* of its argument
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `Parameters` ('R')
##
##PARAMETERS
##- 'R' : a differential polynomial ring or ideal
##
##DESCRIPTION
##- The function call `Parameters` ('R') returns the list of *parameters*
##  of 'R'.
##
##- This list makes explicit the dependencies of dependent variables
##  with respect to the independent ones. Each *parameter* appears as
##  dependent variable _k_, possibly followed by a sequence of independent
##  variables enclosed by parentheses, i.e., _k_(_x[i[1]]_, `...`, _x[i[q]]_).
##  
##- If _x[j]_ is an independent variable which is not listed among
##  the dependencies of a parameter _k_, then the derivative of _k_
##  with respect to _x[j]_ is zero.
##
##- These implicit equations, which define the *parameters*, are implictly
##  present in every regular differential chain computed from differential
##  polynomial systems of 'R'. They are not listed by the "Equations"
##  or the "RewriteRules" functions, unless the option `fullset`=`true`
##  is provided.
## 
##- If a *parameter* _k_ does not depend on any independent variable, which 
##  is the most common situation, it can be used as a so-called parameter
##  in the design of a parametric differential system.
##
##- When the `diff` or the `Diff` notation is used (see "Notation"), any
##  dependent variable listed in the list of *parameters* is printed
##  using the given sequence of independent variables. 
##  This feature permits, in particular, to customize the order of
##  the independent variables, a given dependent variable depends on.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x, y], blocks = [u,v,a,k], parameters = [v(y,x), a(x), k]);
##
##- _k_ is a parameter in the usual sense of this word ; _a_ only depends
##  on _x_ ; the order of the independent variables, _v_ depends on, is
##  customized.
##
##> Parameters(R);
##< [v(y, x), a(x), k]
##
##> Differentiate([u, v, a, k], x, R);
##< [u[x], v[x], a[x], 0]
##
##> Differentiate([u, v, a, k], x, R, output=diff);
##< [diff(u(x,y),x), diff(v(y,x),x), diff(a(x),x), 0]
##
##> Differentiate([u, v, a, k], y, R, output=diff);
##< [diff(u(x,y),y), diff(v(y,x),y), 0, 0]
##
##> ideal := RosenfeldGroebner([], R):
##> Parameters(ideal);
##< [v(y, x), a(x), k]
##
##> Equations(ideal, fullset=true);
##< [[k[y], k[x], a[y]]]
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing", "Notation", "Equations", "RewriteRules"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Equations" : Help:DifferentialAlgebra0[Equations]
##- "RewriteRules" : Help:DifferentialAlgebra0[RewriteRules]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]


# EXPORTED
#
# Family of Notation
# Returns the parameters of a differential ring, a regular chain
# or a non-empty list of regular chains.

Parameters := proc (drideal)
    local a_table, result;
    if IsIdeal (drideal, ':-possibly_empty'=false) or 
                                    IsDifferentialRing (drideal) then
	if type (drideal, ':-list') then
	    a_table := drideal [1];
	else
	    a_table := drideal;
	end if;
	try
	    result := ParametersSafe (a_table);
	catch:
	    error
	end try
    else
        error _ERRIDRNG_, "first", convert (drideal, ':-string');
    end if;
    return result;
end proc:

# Subfunction of Parameters

ParametersSafe := proc (a_table)
    local arguments, result;
    arguments := "Parameters" (a_table);
    arguments := arguments, ToBMIOptionSequence
				(':-_JET_', ':-_JET_', 0, 0, ':-_SMALL_');
    result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
    return result;
end proc:
