##PROCEDURE(help) DifferentialAlgebra0[Pardi]
##HALFLINE        performs a change of ranking 
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- Pardi ('regchain', 'ranking', 'opts')
##
##PARAMETERS
##
##- 'regchain'        : a regular differential chain
##- 'rnkg'            : a ranking or a differential polynomial ring
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `Pardi` ('regchain', 'rnkg') returns a
##  regular differential chain, with respect to the ranking 'rnkg',
##  defining the same prime differential ideal as 'regchain'.
##  The returned regular differential chain holds the same
##  *attributes* as 'regchain'.
##
##- The ranking 'rnkg' must involve the same independent and dependent
##  variables as the ones defined in the differential polynomial ring
##  of 'regchain'. It can be given using the same syntax as the one used
##  by the "Ranking" function, or, through a  
##  differential polynomial ring.
##
##- The `Pardi` method only applies to prime differential ideals.
##  If the regular differential chain 'regchain' does not hold the
##  `prime` *attribute*, then, providing the option `prime` = `true`
##  permits to force the execution of the function.
##
##OPTIONS
##
##- `prime` = `true`. This option tells the algorithm to assume that
##  the differential ideal defined by 'regchain' is prime. 
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'regchain' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##   notation
##   used for both input and output. This option is overridden by the two
##   options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##SUBSECTION An academic example
##
##> with(DifferentialAlgebra0):
##
##- This example features an academic system of partial differential
##  equations. First, "RosenfeldGroebner" is used to build a regular
##  differential chain.
##
##> R := DifferentialRing(derivations = [x, y], blocks = [[v, u]]);
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]], R);
##
##> Equations(ideal[1]);
##< [u[y]^2-2*u, u[x]^2-4*u, 4*v[y]*u-u[x]*u[y]*u+u[x]*u[y], v[x,x]-u[x]]
##
##- Second, a change of ranking is performed. It is necessary to provide
##  the option `prime` = `true`, since 'ideal' does not hold the `prime`
##  *attribute*. 
##
##> jdeal := Pardi(ideal [1], ranking(derivations = [x,y], blocks = [u,v]), prime = true);
##
##> Equations(jdeal);
##< [v[y,y]^4-2*v[y,y]^2-2*v[y]^2+1, v[x,y]*v[y]-v[y,y]^3+v[y,y], v[x,x]-2*v[y,y], u-v[y,y]^2]
##
##- Last, the inverse change of ranking is performed, and one checks that
##  the returned regular differential chain is identical to that of 'ideal'.
##
##> kdeal := Pardi(jdeal, R);
##
##> Equations(ideal [1]) - Equations(kdeal);
##< [0, 0, 0, 0]
##
##ENDSUBSECTION
##
##SUBSECTION An example featuring a parametric dynamical system
##
##- This example shows the interest of `Pardi` for studying parametric
##  dynamical systems. Indeed, with respect to any *orderly* ranking on
##  the state variables, a dynamical system is a regular differential chain, 
##  which describes a prime differential ideal.
##
##- Here, `Pardi` is used to compute the so-called *input-output relation*
##  of a dynamical system. This relation is used, in control theory, to
##  study the global identifiability, and, simplify the estimation of the
##  unknown parameters from measures.
##
##- First, one defines a parametric dynamical system of two ODE.
##  The two state variables _x[1]_ and _x[2]_ are grouped in one *block*.
##  This ensures that the derivatives of _x[1]_ and _x[2]_ are ranked *orderly*.
##
##> params := [k[e], V[e], k[1,2], k[2,1]];
##< [k[e], V[e], k[1,2], k[2,1]]
##
##> R := DifferentialRing (derivations = [t], blocks = [[x[1], x[2]], params], parameters = params, notation = diff);
##
##> edoA := diff (x[1](t),t) = -k[1,2]*x[1](t) + k[2,1]*x[2](t) - (V[e]*x[1](t))/(k[e]+x[1](t));
##< diff (x[1](t),t) = -k[1,2]*x[1](t) + k[2,1]*x[2](t) - (V[e]*x[1](t))/(k[e]+x[1](t))
##
##> edoB := diff (x[2](t),t) = k[1,2]*x[1](t) - k[2,1]*x[2](t);
##< diff (x[2](t),t) = k[1,2]*x[1](t) - k[2,1]*x[2](t)
##
##- This call to "RosenfeldGroebner" just bundles the two ODE in a 
##  regular differential chain. The parameters are moved to the base
##  field of the equation, to avoid splitting cases by discussing their values.
##
##> ideal := RosenfeldGroebner ([edoA,edoB], basefield = field (generators = params), R);
##
##- The leading ranks of the regular differential chain have degree _1_.
##  The differential ideal is thus prime.
##
##> LeadingRank (ideal[1]);
##< [diff (x[2](t), t), diff (x[1](t), t)]
##
##- Over this example, it is assumed that _x[1]_ is observed.
##  The input-output relation is thus obtained by eliminating _x[2]_. 
##  This is achieved by a change of ranking over 'ideal', using `Pardi`.
##  The input-output relation, is the differential polynomial of 'IOideal'
##  whose leading derivative is a derivative of _x[1]_.
##  The result is given in the `jet` notation.
##
##> IOideal := Pardi (ideal[1], ranking (derivations = [t], blocks = [x[2], x[1], params]));
##
##> IOrel := Equations (IOideal, leader=derivative(x[1](t)), output=jet)[1];
##< x[1][t,t]*x[1]^2+2*x[1][t,t]*x[1]*k[e]+x[1][t,t]*k[e]^2+x[1][t]*x[1]^2*k[1,2]+x[1][t]*x[1]^2*k[2,1]+2*x[1][t]*x[1]*k[e]*k[1,2]+2*x[1][t]*x[1]*k[e]*k[2,1]+x[1][t]*k[e]^2*k[1,2]+x[1][t]*k[e]^2*k[2,1]+x[1][t]*k[e]*V[e]+x[1]^2*V[e]*k[2,1]+x[1]*k[e]*V[e]*k[2,1]
##
##ENDSUBSECTION
##
##SUBSECTION The two steps strategy
##
##- This example features a partial differential system and 
##  illustrates a good strategy for computing a representation of the 
##  radical of the differential ideal generated by a given system _Sigma_,
##  with respect to some target ranking. 
##
##- This strategy usually leads to simpler computations and fewer
##  regular differential chains. It often applies, because, many 
##  differential polynomial system do actually define *prime* differential 
##  ideals. It proceeds in two steps.
##
##-- First, search an easy ranking, different from the target one, 
##   with respect to which, _Sigma_ is almost a regular differential chain 
##   (if possible, a ranking with respect to which the leading ranks 
##   have degree _1_).
##
##-- Second, run `Pardi` over the regular differential chain(s) computed
##   at the first step, with respect to the target ranking.
##
##- The strategy is applied over Euler equations for an incompressible 
##  fluid, in two dimensions. We want to prove that the radical _I_ of 
##  the differential ideal generated by this system is prime, and, has 
##  differential dimension zero.
##
##-- To prove that _I_ is prime, it is sufficient to show that _I_
##   can be presented by a single regular differential chain _C_, whose 
##   leading ranks have degree _1_.
##
##-- To prove that _I_ has differential dimension zero, it is sufficient
##   to compute _C_ with respect to an *orderly* ranking, and, to
##   check that each dependent variable has a derivative among the
##   leading derivatives of _C_ [K73, chapter II, section 12, Theorem 6].
##
##- The target ranking is *orderly*, but the brute force approach, does
##  not allow us to conclude.
##
##> p1 := v1[t]+v1*v1[x]+v2*v1[y]+p[x]:
##> p2 := v2[t]+v1*v2[x]+v2*v2[y]+p[y]:
##> p3 := v1[x]+v2[y]:
##> target_ranking := ranking (derivations=[x,y,t], blocks = [[v1,v2,p]]);
##< ranking (derivations=[x,y,t], blocks = [[v1,v2,p]])
##
##> R := DifferentialRing (op (target_ranking)):
##> ideal := RosenfeldGroebner ([p1,p2,p3], R);
##
##- There exists an easy ranking, with respect to which, 
##  the "RosenfeldGroebner" algorithm does not split cases.
##  This easy ranking is not *orderly*, but, we can use it to
##  apply our strategy. Let us perform the first step.
##
##> easy_ranking := ranking (derivations=[t,x,y], blocks = [lex[p,v1,v2]]);
##< ranking (derivations=[t,x,y], blocks = [lex[p,v1,v2]])
##
##> S := DifferentialRing (op (easy_ranking)):
##> ideal := RosenfeldGroebner ([p1,p2,p3], S);
##
##- Let us, perform the second step.
##  The output proves that _I_ is prime and has differential dimension zero.
##
##> ideal := Pardi (ideal [1], R);
##
##> LeadingRank (ideal);
##< [v2[x], v1[y], v1[x], p[y, y]]
##
##ENDSUBSECTION
##
##SEEALSO
##
##- "DifferentialAlgebra0", "RosenfeldGroebner"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "Ranking" : Help:DifferentialAlgebra0[Ranking]


# EXPORTED
#
# Pardi (regchain, ranking (derivations = [...], blocks = [...]))
# Pardi (regchain, differential ring)

Pardi := proc (regchain::table, ordering_or_dring,
    {prime::boolean:=false},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'},$)

    local result, arguments, ordrng, innot, outnot, freezeseq, unnumberseq;

    if not IsRegularDifferentialChain (regchain) then
        error _ERRREGC_, "first", convert (regchain, ':-string');
    end if;

# If the second parameter is a table then it is a differential ring
# More checking is performed in BMI.

    if type (ordering_or_dring, ':-table') then
        if IsDifferentialRing (ordering_or_dring) then
            ordrng := RankingSafe (ordering_or_dring);
            outnot := FixNotation 
                            (_OUTPUT_, _IONOT_, 
                             NotationSafe (ordering_or_dring),
                             NotationSafe (regchain));
        else
            error _ERRORDR_, "second", convert (ordrng, ':-string');
        end if;
    else
        ordrng := ordering_or_dring;
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (regchain));
    end if;
    freezeseq, unnumberseq := 
	FrozenSymbols (rhs (op (1, ordrng)), rhs (op (2, ordrng)));

    arguments := "Pardi" 
			(regchain, 
			 ConvertString (regchain [':-_FREEZESEQ_'], ordrng),
			 prime);
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (regchain));
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;

# The result is FAIL or a blad_regchain
# FAIL is returned if the ordering is wrong or incompatible.

    if result = FAIL then
        error _ERRORDR_, "second", convert (ordrng, ':-string')
    else
        result := ToRegchain (result, outnot, freezeseq, unnumberseq);
    end if;
    return result;
end proc:
