##PROCEDURE(help) DifferentialAlgebra0[PowerSeriesSolution]
##HALFLINE        computes formal power series solution
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- PowerSeriesSolution ('regchain','n','exppnt','opts')
##
##PARAMETERS
##
##- 'regchain' : a regular differential chain
##- 'n' : a nonnegative integer
##- 'exppnt' (optional) : a list or a set of equations
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `PowerSeriesSolution` ('regchain','n','exppnt')
##  returns an integral (exponents are nonnegative integers),
##  formal (the convergence issue is not addressed),
##  power series solution of the differential ideal defined by 'regchain', 
##  truncated at order 'n'.
##  The 'exppnt' argument provides both the expansion point of the
##  series and the (generalized) initial values.
##  The result is a list of equations _u = series_ (one for each dependent
##  variable) in the `diff` or the `Diff` notation (default is `diff`).
##
##- The 'exppnt' argument is a list or a set of equations of the form
##  _v[i] = alpha[i]_, where the _v[i]_ are *derivatives* or independent
##  variables and the _alpha[i]_ are general expressions. The _v[i]_
##  must be pairwise distinct.
##
##- The equations _v[i] = alpha[i]_ such that the _v[i]_ are independent
##  variables, provide the expansion point of the series. Independent
##  variables which do not occur among the _v[i]_ are taken to be _0_.
##
##- The equations _v[i] = alpha[i]_ such that the _v[i]_ are *derivatives*,
##  provide the arbitrary constants, or functions, the series depends on. 
##  They provide also values for the constants which are algebraic over 
##  the other ones. The algorithm proceeds as follows.
##-- First a generic sum is computed, where
##   each *derivative* is replaced by its normal form.
##-- Second, some substitutions are performed according to the following rules.
##   Consider a *derivative* _u_, occurring in some term of the generic sum.
##--- If _u_ is not the derivative of any _v[i]_, then _u_ is left
##    unevaluated.
##--- If _u_ is the derivative of exactly one _v[i]_, say, _u = theta*v[i]_ 
##    for some derivation operator _theta_, then, _theta*alpha[i]_ is
##    substituted to _v[i]_. The differentiation _theta*alpha[i]_ is 
##    performed using `diff` or `Diff`, depending on the output notation.
##--- If _u_ is the derivative of many different _v[i]_, then, an equation
##    such that _v[i]_ has maximal *order* is selected.
##    The substitution is then performed as explained just above.
##-- Last, the resulting expression is evaluated, at the expansion point.
##
##- The values attributed to the *derivatives* at the expansion point,
##  cannot be chosen freely. They must satisfy the system _F = 0, S <> 0_,
##  returned by "PowerSeriesSystem". These conditions, which are checked
##  by default, can be relinquished. See one of the options below.
##-- If the attributed values are solutions of the system _F = 0, S <> 0_, 
##   then, the returned power series are truncated solutions of the 
##   differential ideal. 
##-- If these values do not satisfy _F = 0_, then, the returned power series
##   are *not* truncated solutions of the differential ideal.
##-- If they satisfy _F = 0_, but do not satisfy _S <> 0_, then, 
##   the returned power series may, or may not, be truncated solutions of 
##   the differential ideal. If they are, they do not need to be unique.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `PowerSeriesSolution(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[PowerSeriesSolution](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `check` = `nothing`, `inequations`, `equations`, `all`.
##  This option permits to customize the consistency test for 'exppnt'
##  (see above).  The default value is `all`. 
##-- `nothing`. No test is performed.
##-- `inequations`. The function only checks that the
##   inequations are not annihilated by 'exppnt'.
##-- `equations`. The function only checks that the
##   equations are annihilated by 'exppnt'.
##-- `all`. The function checks both the equations and the inequations.
## 
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for the left-hand sides of 'exppnt'. 
##  If not specified, the notation of 'regchain' is used (see "Notation").
##
##- `output` = `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. 
##  If not specified, the notation of 'regchain' is used, provided that
##  it is `diff` or `Diff`.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for both input and output. 
##  This option is overridden by the two options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##
##SUBSECTION A simple linear ODE
##
##> R := DifferentialRing(derivations = [t], blocks = [u]);
##
##> exponential := PretendRegularDifferentialChain([u[t]-u], R);
##
##- There are no constraints on the initial conditions for linear systems.
##  Here is the generic series.
##
##> PowerSeriesSolution(exponential, 3);
##< [u(t) = u(0)+u(0)*t+1/2*u(0)*t^2+1/6*u(0)*t^3]
##
##- The generic series, at _t=t[0]_.
##
##> PowerSeriesSolution(exponential, 3, [u=c,t=t[0]]);
##< [u(t) = c+c*(t-t[0])+1/2*c*(t-t[0])^2+1/6*c*(t-t[0])^3]
##
##- A particular series, for _u(0)=1_.
##
##> PowerSeriesSolution(exponential, 3, [u=1]);
##< [u(t) = 1+t+1/2*t^2+1/6*t^3]
##
##ENDSUBSECTION
##
##SUBSECTION A simple nonlinear ODE
##
##- There are constraints on the initial conditions.
##
##> ideal := PretendRegularDifferentialChain ([u[t]^2 - 4*u], R);
##
##>(quote) PowerSeriesSystem (ideal);
##< [eval(diff(u(t),t),{t = 0})^2-4*u(0) = 0], [eval(diff(u(t),t),{t = 0}) <> 0]
##
##- The initial conditions _u(0)=c^2, u[t](0)=2*c_ 
##  below satisfy the constraints.
##  The series, which turns out to be a polynomial, can be computed.
##
##>(quote) PowerSeriesSystem (ideal, [u[t]=2*c, u=c^2]);
##< [0 = 0], [2*c <> 0]
##
##> PowerSeriesSolution (ideal, 3, [u[t]=2*c, u=c^2]);
##< [u(t) = c^2+2*c*t+t^2]
##
##- The condition _2*c <> 0_ is actually pointless.
##  It can be relinquished.
##
##> PowerSeriesSolution (ideal, 3, [u[t]=0, u=0], check=equations);
##< [u(t) = t^2]
##
##- One may wish to compute a generic formula for the solution, 
##  keeping in mind that the initial values must satisfy a relation.
##
##> PowerSeriesSolution (ideal, 3, check=inequations);
##< [u(t) = u(0)+eval(diff(u(t),t),{t = 0})*t+t^2]
##
##ENDSUBSECTION
##
##SUBSECTION The heat equation
##
##- The ranking is *orderly*. Thus,
##  the formal power series solutions of this equation, involves
##  all the derivatives of the form _u[x^n]_ and _u[t,x^n]_ where _n_
##  stands for any nonnegative integer. 
##  Classically, one says that the solutions of the heat equation
##  depend on two arbitrary functions _u(0,x) = f(x)_ and _u[t](0,x) = g(x)_.
##  Such a formal power series can be generated by `FormalPowerSeries`.
##
##> R := DifferentialRing (derivations = [t,x], blocks = [u]);
##
##> heat := PretendRegularDifferentialChain ([u[t,t] - u[x]], R);
##
##> PowerSeriesSolution (heat, 2, [u=f(x), u[t]=g(x)]);
##< [u(t,x) = f(0)+eval(diff(f(x),x),{x = 0})*x+g(0)*t+1/2*eval(diff(diff(f(x),x),x),{x = 0})*x^2+eval(diff(g(x),x),{x = 0})*t*x+1/2*eval(diff(f(x),x),{x = 0})*t^2]
##
##ENDSUBSECTION
##
##SUBSECTION A general PDE
##
##- The regular differential chains computed by the "RosenfeldGroebner"
##  function permit to compute integral formal power series solutions
##  of the input system. One considers, here, a polynomial PDE system.
##
##> R := DifferentialRing (derivations = [x,y], blocks = [u,v]);
##
##> syst := [u[x,x]^2*v+u[x,x]*v+u[x], u[x,y], u[y,y]^2-1];
##< [u[x,x]^2*v+u[x,x]*v+u[x], u[x,y], u[y,y]^2-1]
##
##- The radical of the differential ideal generated by 'syst'
##  is an intersection of two differential ideals presented by
##  regular differential chains. The solution set of 'syst' is thus
##  the union of the solution sets of the two chains.
##
##> ideal := RosenfeldGroebner (syst, [u[x]], R);
##
##- Inspecting the left-hand sides of the equations of the first
##  component, one sees that the solutions depend on one arbitrary
##  function _v(x,0) = f(x)_, three arbitrary constants
##  _u(0,0) = c[0]_, _u[x](0,0) = c[1]_, _u[y](0,0) = c[2]_, and,
##  two constants which are algebraic over the other ones:
##  _u[y,y](0,0)_ and _u[x,x](0,0)_.
##
##> RewriteRules (ideal[1]);
##< [v[y] = 0, u[y,y]^2 = 1, u[x,y] = 0, u[x,x]^2 = -(u[x,x]*v+u[x])/v]
##
##- Let us choose some values. Some of them are symbolic, the other
##  ones are numeric. First one checks that they satisfy the above
##  system. Then, the series is computed.
##
##> iv := [v=1+x*f(x), u[y,y]=1, u=c[0], u[x]=0, u[y]=c[2], u[x,x]=-1];
##< [v=1+x*f(x), u[y,y]=1, u=c[0], u[x]=0, u[y]=c[2], u[x,x]=-1]
##
##> PowerSeriesSystem (ideal[1], iv);
##< [0 = 0, 0 = 0], [1 <> 0, 1 <> 0, -1 <> 0]
##
##> PowerSeriesSolution (ideal[1], 3, iv);
##< [u(x,y) = c[0]+c[2]*y+1/2*y^2-1/2*x^2-1/6*x^3, v(x,y) = 1+f(0)*x+eval(diff(f(x),x),{x = 0})*x^2+1/2*eval(diff(diff(f(x),x),x),{x = 0})*x^3]
##
##- Inspecting the left-hand sides of the equations of the second component,
##  one sees that the solutions depend on three arbitrary constants
##  _u(0,0) = c[0]_, _u[y](0,0) = c[1]_, _v(0,0) = c[2]_, and,
##  one constant _u[y,y](0,0)_ which is equal to +/-1.
##
##> RewriteRules (ideal[2]);
##< [v[y] = 0, v[x] = -2, u[x] = 1/4*v, u[y,y]^2 = 1]
##
##> sols := PowerSeriesSolution (ideal[2], 10, [u=c[0], u[y]=c[1], v=c[2], u[y,y]=1]);
##< [u(x,y) = c[0]+c[1]*y+1/4*c[2]*x+1/2*y^2-1/4*x^2, v(x,y) = c[2]-2*x]
##
##- These solutions are polynomials. One can plug them into the input
##  equations, and check the result.
##
##> syst := Equations (syst, output=diff, R);
##< [diff(diff(u(x,y),x),x)^2*v(x,y)+diff(diff(u(x,y),x),x)*v(x,y)+diff(u(x,y),x), diff(diff(u(x,y),x),y), diff(diff(u(x,y),y),y)^2-1]
##
##> eval (syst, sols);
##< [0, 0, 0]
##
##ENDSUBSECTION
##
##SEEALSO
##- "DifferentialAlgebra0", "PowerSeriesSystem", "NormalForm"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "PowerSeriesSystem" : Help:DifferentialAlgebra0[PowerSeriesSystem]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]


# EXPORTED

PowerSeriesSolution := proc 
    (regchain::table, 
     ordre::nonnegative, 
     exppnt::{list(equation),set(equation)}:=[],
     {check::identical(:-nothing,:-equations,:-inequations,:-all):=':-all'},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local innot, outnot, result;

    if not IsRegularDifferentialChain (regchain) or
		not member (':-differential', Attributes (regchain)) then
	error _ERRALGC_, "first", convert (regchain, ':-string');
    end if;

    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (regchain));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (regchain));
    if outnot = ':-_JET_' or outnot = ':-_TJET_' then
	outnot := ':-_DIFF_'
    end if;

    try
        result := PowerSeriesSolutionSafe 
			(regchain, ordre, exppnt, check, innot, outnot);
    catch:
        error
    end try;
    return result;
end proc:

# Local to PowerSeriesSolution
# regchain  = the regular differential chain to be solved
# ordre     = the order of the series to be computed
# exppnt    = the list/set of equations for EvaluateSafe
# check     = a keyword for customizing the test on initial values
# innot     = the input notation for lhs(exppnt)
# outnot    = the output notation (diff or Diff)

PowerSeriesSolutionSafe := proc 
			(regchain, ordre, exppnt, check, innot, outnot)

    local R, independent_vars, dependent_vars, eqns, ineqns, L, M, 
		lhs_subst, new_subst, expansion_point, 
		i, j, comp, soln;

# The ring of the regchain, in tjet notation

    R := DifferentialRing (regchain, ':-_IONOT_' = ':-_TJET_');
    independent_vars := subs (op (Ranking (R)), ':-derivations');
    dependent_vars := Indets (R, ':-selection' = ':-_DEPVARS_');
    dependent_vars := SortByRank (dependent_vars, ':-_DESC_', R);

# Convert lhs(exppnt) to the tjet notation

    lhs_subst := map (lhs, exppnt);
    lhs_subst := NormalForm (lhs_subst, R, ':-_INPUT_'=innot);
    new_subst := seq (lhs_subst [i] = rhs (exppnt [i]),
                                        i = 1 .. nops (exppnt));

# Complete exppnt with x=0 for each independent variable x

    for i from 1 to nops (independent_vars) do
	if not member (independent_vars [i], lhs_subst) then
	    new_subst := independent_vars [i] = 0, new_subst;
	end if
    end do;

    new_subst := [ new_subst ];

# The exppnt argument is completed. Then check.

    eqns, ineqns := PowerSeriesSystemSafe2 (regchain, new_subst, check, R);
    eqns := remove (x -> evalb (x=0), eqns);
    if eqns <> [] then
	error _ERRPSSE_;
    elif member (0, ineqns) then
	error _ERRPSSN_;
    end if;

# Compute the solutions

    soln := NULL;
    for i from 1 to nops (dependent_vars) do

# Get the derivatives up to order: ordre

	M := PowerSeriesSolutionDerivatives 
		(dependent_vars [i], ordre, R, [], ':-_TJET_', ':-_TJET_');

# Take the normal forms of these derivatives

	L := NormalForm (M, regchain, ':-_IONOT_'=':-_TJET_');

# Run EvaluateSafe over the normal forms

	L, expansion_point := EvaluateSafe (L, new_subst, R, outnot);

# Take the result at the expansion point

	L := map (PowerSeriesSolutionEval, L, expansion_point);

# Multiply by the x^k/k!

	M := map (PowerSeriesSolutionFactor, 
				M, independent_vars, expansion_point);
# Store
	comp := 
	    NormalForm (dependent_vars [i], R, ':-_OUTPUT_'=outnot) =
	    add (L[j]*M[j], j = 1 .. nops (L));
	soln := soln, comp
    end do;
    return [ soln ];
end proc:

# Local to PowerSeriesSolutionDerivatives
# Returns all the derivatives of depv, of total order less than or equal
# to ordre, which are not derivatives of corners.

PowerSeriesSolutionDerivatives := proc (depv, ordre, R, corners, innot, outnot)
    local arguments, result;

    arguments := "AllDerivatives" 
			(ConvertString (R [':-_FREEZESEQ_'], depv), 
			 ConvertString (R [':-_FREEZESEQ_'], ordre), 
			 R, 
			 ConvertString (R [':-_FREEZESEQ_'], corners));
    arguments := arguments,
		    ToBMIOptionSequence (innot, outnot, 0, 0, ':-_SMALL_');
    result := BMIBladEvalSafe (R [':-_UNNUMBERSEQ_'], arguments);
    return result
end proc:

# Work around a annoying eval feature
# Should be: eval (expr, expansion_point)
# expansion_point is the expansion point

PowerSeriesSolutionEval := proc (expr, expansion_point)
    local new_expansion_point;
    if not has (expr, ':-diff') and not has (expr, ':-Diff') then
	return eval (expr, expansion_point);
    else
        new_expansion_point := 
	    remove((t,e) -> op(1, t) = op(2, t) or 
			    type(op(1, t), ':-name') and not has(e, op(1, t)), 
		    expansion_point, expr);
	if indets (map (rhs, new_expansion_point)) = {} then
	    return eval (expr, new_expansion_point);
	else
	    return Eval (expr, new_expansion_point);
	end if;
    end if;
end proc:

# Computes the x^k/k! factor
# v is a derivative in tjet notation
# independent_vars is the list of independent variables
# expansion_point is the expansion point

PowerSeriesSolutionFactor := proc (v, independent_vars, expansion_point)
    local m, i, d;
    m := convert (v, `*`);
    d := [ seq (degree (m, independent_vars[i]), 
				i = 1 .. nops (independent_vars)) ];
    mul ((lhs (expansion_point[i]) - rhs (expansion_point[i]))^d[i]/d[i] !, 
				i = 1 .. nops (independent_vars));
end proc:
