##PROCEDURE(help) DifferentialAlgebra0[PowerSeriesSystem]
##HALFLINE        returns constraints to be satisfied by initial values
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- PowerSeriesSystem ('regchain','exppnt','opts')
##
##PARAMETERS
##
##- 'regchain' : a regular differential chain
##- 'exppnt' (optional) : a list or a set of equations
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- The function call `PowerSeriesSystem` ('regchain') returns
##  a sequence of two lists, a list of equations, and, a list of
##  inequations, that the initial values must satisfy so that
##  an integral formal power series can be computed, by
##  "PowerSeriesSolution".
##
##- The function call `PowerSeriesSystem` ('regchain', 'exppnt')
##  returns the system to be satisfied, evaluated at the expansion
##  point 'exppnt'. 
##
##- The list of equations is the set of the regular differential
##  chain polynomials, which have leading degrees greater than or
##  equal to _2_. See "Equations". The list of inequations is the
##  set of initials and separants of the regular differential
##  chain polynomials. See "Inequations".
##
##- The 'exppnt' argument is a list or a set of equations of the form
##  _v[i] = alpha[i]_, where the _v[i]_ are *derivatives* or independent
##  variables and the _alpha[i]_ are general expressions. The _v[i]_
##  must be pairwise distinct.
##
##- The equations _v[i] = alpha[i]_, such that the _v[i]_ are independent
##  variables, provide the expansion point of the series. Independent
##  variables which do not occur among the _v[i]_ are taken to be _0_.
##
##- The equations _v[i] = alpha[i]_, such that the _v[i]_ are *derivatives*,
##  provide the arbitrary constants (initial values), or functions, the 
##  series depends on.
##  They provide also values for the constants which are algebraic over
##  the other ones. The algorithm proceeds as follows.
##-- First, the generic system of equations and inequations is computed.
##-- Second, some substitutions are performed according to the following rules
##   (see "PowerSeriesSolution").
##   Consider a *derivative* _u_, occurring in some term of the generic system.
##--- If _u_ is not the derivative of any _v[i]_, then _u_ is left
##    unevaluated.
##--- If _u_ is the derivative of exactly one _v[i]_, say, _u = theta*v[i]_
##    for some derivation operator _theta_, then, _theta*alpha[i]_ is
##    substituted to _v[i]_. The differentiation _theta*alpha[i]_ is
##    performed using `diff` or `Diff`, depending on the output notation.
##--- If _u_ is the derivative of many different _v[i]_, then, an equation
##    such that _v[i]_ has maximal *order* is selected.
##    The substitution is then performed as explained just above.
##-- Last, the resulting system is evaluated, at the expansion point.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `PowerSeriesSystem(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[PowerSeriesSystem](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `check` = `nothing`, `inequations`, `equations`, `all`.
##  This option permits to relinquish the conditions that 'exppnt'
##  must satisfy. See "PowerSeriesSolution".
##  The default value is `all`.
##-- `nothing`. The returned system is a sequence of two empty lists.
##-- `inequations`. The list of equations of the returned system is empty.
##-- `equations`. The list of inequations of the returned system is empty.	
##-- `all`. The returned system is complete.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for the left-hand sides of 'exppnt'.
##  If not specified, the notation of 'regchain' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'regchain' is used, provided that
##  it is `diff` or `Diff`.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for both input and output. 
##  This option is overridden by the two options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##
##> R := DifferentialRing(derivations = [t], blocks = [y], notation = diff);
##
##> F := diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2;
##< diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2
##
##> ideal := PretendRegularDifferentialChain([F], R);
##
##- The generic system
##
##>(quote) PowerSeriesSystem(ideal, [t=t0]);
##< [diff(y(t0),t0)^3-4*t0*diff(y(t0),t0)*y(t0)+8*y(t0)^2 = 0], [3*diff(y(t0),t0)^2-4*t0*y(t0) <> 0]
##
##>(quote) PowerSeriesSystem(ideal, [t=t0], check=equations);
##< [diff(y(t0),t0)^3-4*t0*diff(y(t0),t0)*y(t0)+8*y(t0)^2 = 0], []
##
##>(quote) PowerSeriesSystem(ideal, [t=t0], check=inequations);
##< [], [3*diff(y(t0),t0)^2-4*t0*y(t0) <> 0]
##
##>(quote) PowerSeriesSystem(ideal, [t=t0], check=nothing);
##< [], []
##
##- The generic system, at t=0(default value, when not specified)
##
##>(quote) PowerSeriesSystem(ideal);
##< [eval(diff(y(t),t),{t = 0})^3+8*y(0)^2 = 0], [3*eval(diff(y(t),t),{t = 0})^2 <> 0]
##
##- Let us give some names to the values of _y(t)_ and its derivative, at
##  _t = 0_.
##
##>(quote) PowerSeriesSystem(ideal, [y(t)=c[0], diff(y(t),t)=c[1], t=0]);
##< [c[1]^3+8*c[0]^2 = 0], [3*c[1]^2 <> 0]
##
##- The following values are correct.
##
##>(quote) PowerSeriesSystem(ideal, [y(t)=c^3, diff(y(t),t)=-2*c^2, t=0]);
##< [0 = 0], [12*c^4 <> 0]
##
##> PowerSeriesSolution(ideal, 3, [y(t)=c^3, diff(y(t),t)=-2*c^2, t=0]);
##< [y(t) = c^3-2*c^2*t+c*t^2]
##
##SEEALSO
##- "DifferentialAlgebra0", "PowerSeriesSolution", "Equations", "Inequations"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "PowerSeriesSolution" : Help:DifferentialAlgebra0[PowerSeriesSolution]
##- "Equations" : Help:DifferentialAlgebra0[Equations]
##- "Inequations" : Help:DifferentialAlgebra0[Inequations]


PowerSeriesSystem := proc 
    (regchain::table, 
     exppnt::{list(equation),set(equation)}:=[],
     {check::identical(:-nothing,:-equations,:-inequations,:-all):=':-all'},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)

    local innot, outnot, result;

    if not IsRegularDifferentialChain (regchain) or
		not member (':-differential', Attributes (regchain)) then
	error _ERRALGC_, "first", convert (regchain, ':-string')
    end if;

    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (regchain));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (regchain));
    if outnot = ':-_JET_' or outnot = ':-_TJET_' then
	outnot := ':-_DIFF_';
    end if;

    try
        result := PowerSeriesSystemSafe 
			(regchain, exppnt, check, innot, outnot);
    catch:
        error
    end try;
    return result;
end proc:

# Local to PowerSeriesSystem
# regchain  = the regular differential chain to be solved
# exppnt    = the list/set of equations for EvaluateSafe
# check     = a keyword for customizing the test on initial values
# innot     = the input notation for lhs(exppnt)
# outnot    = the output notation (diff or Diff)

PowerSeriesSystemSafe := proc (regchain, exppnt, check, innot, outnot)
    local R, independent_vars, lhs_subst, new_subst, i, eqns, ineqns;

# The ring of the regchain, in tjet notation

    R := DifferentialRing (regchain, ':-_IONOT_' = ':-_TJET_');
    independent_vars := subs (op (Ranking (R)), ':-derivations');

# Convert lhs(exppnt) to the tjet notation

    lhs_subst := map (lhs, exppnt);
    lhs_subst := NormalForm (lhs_subst, R, ':-_INPUT_'=innot);
    new_subst := seq (lhs_subst [i] = rhs (exppnt [i]),
                                        i = 1 .. nops (exppnt));

# Complete exppnt with x=0 for each independent variable x

    for i from 1 to nops (independent_vars) do
        if not member (independent_vars [i], lhs_subst) then
            new_subst := independent_vars [i] = 0, new_subst;
        end if
    end do;

    new_subst := [ new_subst ];

# The exppnt argument is completed. One gets the system.

    eqns, ineqns := PowerSeriesSystemSafe2 (regchain, new_subst, check, R);

    eqns := map (x -> x = 0, eqns);
    ineqns := map (x -> x <> 0, ineqns);

    return eqns, ineqns;
end proc:

# Local to PowerSeriesSystem
# regchain         = the regular differential chain to be solved
# completed_exppnt = the complete list/set of equations for EvaluateSafe
# check            = a keyword for customizing the test on initial values
# R                = the ring of regchain, in tjet notation

PowerSeriesSystemSafe2 := proc (regchain, completed_exppnt, check, R)
    local eqns, ineqns, indepsubs;

    if not member (check, [':-inequations', ':-all']) then
	ineqns := []
    else
	ineqns := Inequations (regchain, ':-_OUTPUT_'=':-_TJET_');
	ineqns, indepsubs := 
		EvaluateSafe (ineqns, completed_exppnt, R, ':-_DIFF_');
	ineqns := map (PowerSeriesSolutionEval, ineqns, indepsubs);
	ineqns := value (ineqns);
	ineqns := map (expand, ineqns);
    end if;

    if not member (check, [':-equations', ':-all']) then
	eqns := [];
    else
	eqns := RewriteRules (regchain, ':-_OUTPUT_'=':-_TJET_');
	eqns := map (proc (eq) 
			if degree (lhs (eq)) >= 2 then 
			    lhs (eq) - rhs (eq) 
			else 
			    NULL 
			end if 
		     end proc, eqns);
	eqns, indepsubs := 
		EvaluateSafe (eqns, completed_exppnt, R, ':-_DIFF_');
	eqns := map (PowerSeriesSolutionEval, eqns, indepsubs);
	eqns := value (eqns);
	eqns := map (expand, eqns);
    end if;
    return eqns, ineqns;
end proc:
