##PROCEDURE(help) DifferentialAlgebra0[PreparationEquation]
##HALFLINE        returns the preparation equation of a differential polynomial
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- PreparationEquation ('f', 'regchain', 'opts')
##
##PARAMETERS
##- 'f' : a differential polynomial
##- 'regchain' : a regular differential chain
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `PreparationEquation` ('f', 'regchain') returns
##  a preparation equation [K73, chapter IV, section 13] for 'f' with respect 
##  to 'regchain'. 
##  The argument 'f' is regarded as a differential polynomial of the
##  embedding ring of 'regchain'.
##
##- Let _I_ denote the differential ideal defined by 'regchain'
##  and denote _A[1]_, `...`, _A[r]_ the differential polynomials
##  which constitute the chain. Introduce _r_ new dependent variables
##  _z[i]_. Each variable _z[i]_ represents the differential polynomial _A[i]_.
##
##- The returned preparation equation is an expression having the
##  form _h*f_ = _c[1]*t[1]_ + `...` + _c[n]*t[n]_.
##  The differential polynomial _h_ is a power product of initials
##  and separants of the _A[k]_.
##  The coefficients _c[i]_ are reduced and regular with respect to _I_.
##  The monomials _t[i]_ are power products of the _z[k]_ variables
##  and their derivatives.
##  They satisfy some further properties, described in 
##  [K73, chapter IV, section 13].
##  If each _z[k]_ is replaced by the corresponding _A[k]_, in all terms 
##  _t[i]_, then the preparation equation becomes a true equality.
##
##- Preparation equations are an important tool in the context of
##  the Low Power Theorem. See "EssentialComponents".
##
##OPTIONS
##
##- The 'opts' arguments may contain one or more of the options below.
##
##- 'congruence' = `true`.
##  In the right hand-side of the returned preparation equation, 
##  denote _q_ the *minimum* total degree of the monomials _t[i]_.
##  All the terms _c[i]*t[i]_ such that
##  the total degree of _t[i]_ is greater than _q_, are removed
##  from the right hand-side of the preparation equation.
##  This stripped preparation equation is called a *preparation
##  congruence* of 'f'.
##
##- 'n' = ::nonnegative:: (default value is _0_).
##  This option is useful in conjunction with the option 
##  'congruence' = `true`. The 'n' first differential polynomials
##  _A[1]_, `...`, _A[n]_ of 'regchain' are considered as equations defining
##  the base field of 'f', and, of the differential polynomials 
##  _A[n+1]_, `...`, _A[r]_. 
##  Reductions by the base field equations are not taken into account 
##  for computing the preparation congruence of 'f': 
##  the terms _t[i]_ involving derivatives of _z[1]_, `...`, _z[n]_
##  are not considered for determining _q_, and, do not appear in
##  the preparation congruence.
##
##- 'zstring' = ::string::.
##  This option permits to customize the identifier used for
##  the new variables _z[k]_. It must be a valid MAPLE identifier
##  (possibly an ::indexed::) involving the substring `"%d"`.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. Specifies
##  the notation used for 'f'.
##  If not specified, the notation of 'regchain' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'regchain' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##
##SUBSECTION Basic illustration
##
##- The following examples illustrate the function, syntactically.
##
##> R := DifferentialRing(derivations = [t], blocks = [u,[s,c]]);
##
##> ideal := PretendRegularDifferentialChain([c^2 + s^2 - 1, c[t] + s, u], R, pretend = false);
##
##> Equations(ideal);
##< [c^2 + s^2 - 1, c[t] + s, u]
##
##> f :=(c[t] + s)*u[t] + u[t]^2;
##< (c[t] + s)*u[t] + u[t]^2
##
##> prepeq := PreparationEquation(f, ideal);
##< u[t]*c[t]+u[t]*s+u[t]^2 = z3[t]^2+z2*z3[t]
##
##- If one substitutes the _A[i]_ to the _z[i]_, the equation becomes
##  an equality.
##
##> expand(eval (prepeq, [z3 = u, z2 = c[t] + s]));
##< u[t]*c[t]+u[t]*s+u[t]^2 = u[t]*c[t]+u[t]*s+u[t]^2
##
##- Changing the identifier for the _z[i]_.
##
##> PreparationEquation(f, ideal, zstring = "A(%d)");
##< u[t]*c[t]+u[t]*s+u[t]^2 = A(3)[t]^2+A(2)*A(3)[t]
##
##- Since all monomials _t[i]_ have degree _q = 2_, the preparation
##  congruence is equal to the preparation equation.
##
##> PreparationEquation(f, ideal, congruence=true);
##< u[t]*c[t]+u[t]*s+u[t]^2 = z3[t]^2+z2*z3[t]
##
##- However, if the two first elements of the regular differential chain
##  are considered as base field defining equations, then, only one
##  monomial _t[i]_ is left in the congruence.
##  
##> PreparationEquation(f, ideal, congruence=true, n=2);
##< u[t]*c[t]+u[t]*s+u[t]^2 = z3[t]^2
##ENDSUBSECTION
##
##SUBSECTION The Low Power Theorem
##- The next example illustrates the Low Power Theorem.
##  See [R50, chapter III] and [K73, chapter IV, section 15].
##
##> R := DifferentialRing(derivations = [t], blocks = [y]);
##
##> f := y[t]^3-4*y[t]*y*t+8*y^2;
##< y[t]^3-4*y[t]*y*t+8*y^2
##
##- First compute a representation of the radical of the differential ideal
##  generated by _f_, by means of "RosenfeldGroebner".
##
##> ideal := RosenfeldGroebner([f], R);
##> Equations(ideal);
##< [[y[t]^3-4*y[t]*y*t+8*y^2], [27*y-4*t^3], [y]]
##
##- Second, remove any regular differential chain which involve two or
##  more differential polynomials, by application of the Component
##  Theorem [K73, chapter IV, section 14]. 
##  In our case, no regular differential chain
##  is removed by this process. 
##  Third, compute a preparation congruence for _f_, with respect
##  to each of the two singular components, i.e., the two last ones.
##
##- In the first case, there is only one monomial _t[1]_, of the
##  form _z[1]^q_. Thus this regular differential chain must be kept
##  in the decomposition.
##
##> rhs(PreparationEquation(f, ideal[2], congruence=true));
##< 167365651248*t^3*z1
##
##- In the second case, the right hand-side of the preparation congruence 
##  involves two monomials. Thus this regular differential chain is redundant.
##
##> rhs(PreparationEquation (f, ideal[3], congruence=true));
##< -4*t*z1*z1[t]+8*z1^2
##
##- Indeed, "EssentialComponents" removes the second singular component 
##  from the decomposition.
##
##> ideal := EssentialComponents(f, R);
##
##> Equations(ideal);
##< [[y[t]^3-4*y[t]*y*t+8*y^2], [27*y-4*t^3]]
##ENDSUBSECTION
##
##SEEALSO
##- "DifferentialAlgebra0", "EssentialComponents", "RosenfeldGroebner"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]

# EXPORTED
#
#

PreparationEquation := proc
     (eqns,
      regchain::table,
      {n::nonnegative:=0},
      {congruence::boolean:=false},
      {zstring::string:="z%d"},
      {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
      {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
      {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
      {_TIMEOUT_::nonnegative:=0},
      {_MEMOUT_::nonnegative:=0},
      {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)
    local arguments, result, leqns, innot, outnot;

    if not IsRegularDifferentialChain (regchain) then
	error _ERRREGC_, "second", convert (regchain, ':-string')
    elif type (eqns, '{:-list,:-set}') then
	error _ERRPOLY_, "first", convert (eqns, ':-string')
    end if;
    try
	leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
    catch:
	error _ERRPOLY_, "first", convert (eqns, ':-string')
    end try;
    arguments := "PreparationEquation" 
		(ConvertString (regchain [':-_FREEZESEQ_'], leqns[1]), 
		 regchain, 
		 ConvertString (regchain [':-_FREEZESEQ_'], n), 
		 congruence,
		 zstring);
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (regchain));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (regchain));
    arguments := arguments,
	ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
	result := BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
    catch:
	error

    end try;
    return result;
end proc:
