##PROCEDURE(help) DifferentialAlgebra0[PretendRegularDifferentialChain]
##HALFLINE        builds a regular differential chain
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- PretendRegularDifferentialChain ('eqns', 'R', 'opts')
##
##PARAMETERS
##- 'eqns' : a list or a set of differential rational fractions
##- 'R'    : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `PretendRegularDifferentialChain` ('eqns', 'R') builds a regular
##  differential chain with the numerators of 'eqns', 
##  regarded as differential rational fractions
##  of 'R', if 'R' is a ring, or its embedding ring if 'R' is an ideal.
##  By default, the built regular differential chain is assumed to hold 
##  the *attributes*: `differential`, `autoreduced`, `primitive`,
##  `squarefree`, `normalized` and `coherent`. For more details on
##  *attributes*, see "DifferentialAlgebra0".
##
##- It is assumed that 'eqns' already forms a regular differential chain
##  with the above *attributes*. The list 'eqns' does not need to be 
##  sorted.
##
##- In principle, the elements of 'eqns' should be differential polynomials
##  with integer numeric coefficients. However, rational differential
##  fractions and expressions involving explicit relational operators,
##  such as _p = q_ and _p <> q_ are accepted. The rational differential
##  fractions are replaced by their numerators. The expressions _p = q_
##  are converted into _p - q_. The expressions _p <> q_ are ignored.
##
##- If 'eqns' involves a *parameter* _p_, the equations stating that some
##  derivatives of _p_ are zero, are automatically inserted in the regular 
##  differential chain, unless _p_, itself, is the leading derivative of some
##  element of 'eqns'. 
##
##- If 'eqns' is empty, the returned regular differential chain
##  represents the zero ideal of 'R'.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `attributes` = 'L' where 'L' is a list involving some of the
##  keywords `differential`, `prime`, `primitive`, `squarefree`, `coherent`,
##  `autoreduced`, `normalized`. This option permits to customize the
##  list of *attributes* of the built regular differential chain.
##  The presence of some of the *attributes* may imply, automatically,
##  the presence of some other ones. The *prime* attribute may be
##  automatically set. For more details, see "DifferentialAlgebra0".
##
##- `pretend` = `false`. With this option, the function checks that 'eqns'
##  holds its *attributes* and performs some further simplifications in
##  order to try to achieve the missing ones. 
##  If these simplifications fail, an error message is raised.
##  This happens, in particular, if the chain is not triangular or if
##  the simplifications lead to split the system into two or more
##  different regular differential chains.
##  The `coherent` *attribute* is not checked.
##  
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'eqns'.
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [t], blocks = [u,[s,c]]);
##
##- The function sorts the differential polynomials by increasing rank.
##
##> ideal := PretendRegularDifferentialChain([s[t] - c, c[t] - s, u[t]^2 - s*u], R);
##
##> Equations(ideal);
##< [c[t]-s, s[t]-c, u[t]^2-s*u]
##
##> Attributes(ideal);
##< [differential, autoreduced, primitive, squarefree, normalized]
##
##- In the next example, the *attribute* `normalized` is omitted.
##
##> ideal := PretendRegularDifferentialChain([c*(s[t] - c), (c[t] - s)^2, u[t]^2 - c[t]*u], attributes = [differential, autoreduced, primitive], pretend = false, R);
##
##> Equations(ideal);
##< [c[t]-s, s[t]-c, u[t]^2-s*u]
##
##> Attributes(ideal);
##< [differential, autoreduced, primitive, squarefree]
##
##SEEALSO
##- "DifferentialAlgebra0", "Attributes", "RosenfeldGroebner"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]

# EXPORTED
#
# PretendRegularDifferentialChain (list(polynomials), attributes = [...], 
#                        pretend = true | false, differential ring)
# The coherence property is never checked
# If pretend = false then all the other ones are checked.

PretendRegularDifferentialChain := proc 
    (equations::{list,set},
     {attributes::list(identical(:-differential,:-prime,:-autoreduced,
                    :-primitive,:-squarefree,:-normalized,:-coherent)):=
            [':-differential',':-autoreduced',':-primitive',':-squarefree',
             ':-normalized',':-coherent']},
     {pretend::boolean:=true},
      drideal,
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'},$)

    local result, arguments, a_table, eqns, ineqs, innot, outnot;

    if not IsIdealDRing (drideal) then
        error _ERRIDRNG_, "second", convert (drideal, ':-string');
    elif type (drideal, ':-list') then
        a_table := drideal[1];
    else
        a_table := drideal;
    end if;
# Process the equations. The inequations are tolerated, but dropped.
# This permits to pass rational fractions.
    eqns := convert (equations, ':-set');
    try
	eqns, ineqs := ToPolynomialEquationsAndInequations (eqns)
    catch:
	error _ERRLSEI_, "first", convert (equations, ':-string')
    end try;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    arguments := "PretendRegularDifferentialChain"
                    (ConvertString (a_table [':-_FREEZESEQ_'], eqns),
                     ConvertString (a_table [':-_FREEZESEQ_'], attributes),
                     ConvertString (a_table [':-_FREEZESEQ_'], pretend),
                     a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    result := ToRegchain (result, outnot, a_table [':-_FREEZESEQ_'], 
						a_table [':-_UNNUMBERSEQ_']);
    return result;
end proc:
