##PROCEDURE(help) DifferentialAlgebra0[Ranking]
##HALFLINE        returns the ranking of its argument
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `Ranking`('R')
##
##PARAMETERS
##- 'R' : a differential polynomial ring or ideal
##
##DESCRIPTION
##- The function call `Ranking`('R') returns a description of the
##  ranking of 'R', or of its embedding ring if 'R' is an ideal.
##  See "DifferentialAlgebra0" for more details.
##
##- The returned description has the form 
##  `ranking(derivations = D, blocks = B)`
##  where 'D', which is the ordered list of independent variables, 
##  provides the derivations, and 'B', which is a list 
##  [_b[1]_, `...`, _b[n]_] of *blocks*, provides the dependent variables.
##
##- Each *block* _b_ has the form: one *block-keyword* among
##  `grlexA`, `grlexB`, `degrevlexA`, `degrevlexB`, `lex`, followed
##  by a list [_u[1]_, `...`, _u[m]_] of dependent variables. 
##
##- The defined ranking is such that _x[1]_ `>` `...` `>` _x[p]_
##  and _b[1]_ `>>` `...` `>>` _b[n]_. Within a block, one has
##  _u[1]_ `>` `...` `>` _u[m]_. For more details, see 
##  "DifferentialAlgebra0".
##
##- The operands of the returned description can be taken by the
##  "DifferentialRing" function as an argument, to build a new
##  differential polynomial ring. 
##
##- When providing a ranking to a function, as an argument, the
##  ranking notation can be simplified:
##
##-- If the *block-keyword* is `grlexA`, then, it can be omitted.
##
##-- If, moreover, the block _[u]_ only involves one dependent variable,
##  then, the square brackets can be omitted.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Ranking(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Ranking](...)`.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [t], blocks = [u, v]);
##
##> Ranking(R);
##< ranking(derivations = [t], blocks = [grlexA[u], grlexA[v]])
##
##> SortByRank([u[t], u[t,t], u], ascending, R, output = diff);
##< [u(t), diff(u(t),t), diff(diff(u(t),t),t)]
##
##> R := DifferentialRing(derivations = [x,y], blocks = [grlexB[d(v),d(u)],lex[v,u]]);
##
##> Ranking(R);
##< ranking(derivations = [x, y], blocks = [grlexB[d(v), d(u)], lex[v, u]])
##
##> SortByRank([d(u), d(u)[x], d(u)[y], d(v), d(v)[x], d(v)[y]], ascending, R);
##< [d(u), d(v), d(u)[y], d(v)[y], d(u)[x], d(v)[x]]
##
##> SortByRank([u, u[x], u[y], u[x,x], v, v[x], v[y]], ascending, R);
##< [u, v, u[y], v[y], u[x], v[x], u[x, x]]
##
##> SortByRank([d(u), d(u)[x], d(u)[y], d(v), d(v)[x], d(v)[y]], ascending, R);
##< [d(u), d(v), d(u)[y], d(v)[y], d(u)[x], d(v)[x]]
##
##> SortByRank([u, u[x], u[y], u[x,x], d(v), d(v)[x], d(v)[y]], ascending, R);
##< [u, u[y], u[x], u[x, x], d(v), d(v)[y], d(v)[x]]
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing", "SortByRank", "FactorDerivative", "Groebner"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]
##- "SortByRank" : Help:DifferentialAlgebra0[SortByRank]
##- "FactorDerivative" : Help:DifferentialAlgebra0[FactorDerivative]
##- "Display" : Help:DifferentialAlgebra0[Display]


# EXPORTED
#
# Family of Notation
# Returns the _ORDERING_ entry of a differential ring, a regular chain
# or a non-empty list of regular chains.

Ranking := proc (drideal)
    local a_table, result;
    if IsIdeal (drideal, ':-possibly_empty'=false) or 
                                    IsDifferentialRing (drideal) then
        if type (drideal, ':-list') then
            a_table := drideal [1];
        else
            a_table := drideal;
        end if;
        try
            result := RankingSafe (a_table);
        catch:
            error
        end try
    else
        error _ERRIDRNG_, "first", convert (drideal, ':-string');
    end if;
    return result;
end proc:

# Subfunction of Ranking

RankingSafe := proc (a_table)
    local arguments, result;
    arguments := "Ranking" (a_table);
    arguments := arguments, ToBMIOptionSequence 
			    (':-_JET_', ':-_JET_', 0, 0, ':-_SMALL_');
    result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
    return result;
end proc:

