##PROCEDURE(help) DifferentialAlgebra0[ReducedForm]
##HALFLINE        computes reduced forms modulo regular differential chains
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- ReducedForm ('p', 'ideal', 'opts')
##- ReducedForm ('L', 'ideal', 'opts')
##
##PARAMETERS
##- 'p' : a differential polynomial
##- 'ideal' : a differential polynomial ideal or ring
##- 'L' : a list or a set of differential polynomials
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- Assume 'ideal' is a regular differential chain. Denote _R_ its
##  embedding differential polynomial ring and _I_ the differential ideal 
##  that it defines.
##
##- The function call `ReducedForm` ('p', 'ideal') returns a differential
##  polynomial, which is a reduced form of 'p' in _R_/_I_, as
##  explained below.
##
##- The reduced form of a differential polynomial _p_, is a
##  differential polynomial _q_, equivalent to _p_ modulo _I_.
##  It is _0_ if _p_ belongs to _I_.
##  If _p_ does not belong to _I_, its reduced form is obtained by 
##  applying *some* simplifications, using the regular differential
##  chain, without introducing any rational fraction in the result.
##  
##- The function call `ReducedForm` ('L', 'ideal') returns the list or the 
##  set of the reduced forms of the elements of 'L' with respect to 'ideal'.
##
##- If 'ideal' is a list of regular differential chains, then the function call
##  `ReducedForm` ('p', 'ideal') returns the list of the reduced forms of 'p'
##  with respect to all the chains, and, the function call 
##  `ReducedForm` ('L', 'ideal') returns the list of the reduced forms of 'L'
##  with respect to all the chains.
##
##- If 'ideal' is a differential polynomial ring, then the function call
##  `ReducedForm` ('p', 'ideal') returns 'p', and, the function call
##  `ReducedForm` ('L', 'ideal') returns 'L'. In both cases, the
##  differential polynomials are expanded.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `ReducedForm(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[ReducedForm](...)`.
##
##OPTIONS
##
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p' and 'L'. 
##  If not specified, the notation of 'ideal' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [w,[v,u],t]);
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+1, v[x,x]-u[x]], R);
##> RewriteRules(ideal);
##< [[u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y])/u, v[x,x] = u[x]]]
##
##- The simplification by the above rewrite rules is not performed since it 
##  would give a rational differential fraction
##> ReducedForm(v[y], ideal);
##< [v[y]]
##
##- The simplification is performed, however, in this case
##> ReducedForm(u*v[y], ideal);
##< [1/4*u[x]*u[y]*u - 1/4*u[x]*u[y]]
##
##- In this case, the reduced form is obtained by simplifying partially
##  the differential polynomial
##> p :=(w+1)*(u*v[y])+ u[x,y]*t;
##< (w+1)*(u*v[y])+ u[x,y]*t
##> rf := ReducedForm(p, ideal[1]);
##< 1/4*w*u[x]*u[y]*u-1/4*w*u[x]*u[y]+u[x,y]*t+1/4*u[x]*u[y]*u-1/4*u[x]*u[y]
##> BelongsTo(p - rf, ideal[1]);
##< true
##
##SEEALSO
##- "DifferentialAlgebra0", "RosenfeldGroebner", "NormalForm", "BelongsTo"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]
##- "BelongsTo" : Help:DifferentialAlgebra0[BelongsTo]

# EXPORTED

ReducedForm := proc (eqns, drideal,
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'},$)

    local arguments, result, leqns, innot, outnot, var, regchain;

    if IsDifferentialRing (drideal) then
# ReducedForm ( ..., ring )
# Just equivalent to Differentiate ( ..., 1, ring )
	try
	    leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "first", convert (eqns, ':-string')
	end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (drideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (drideal));
	try
	    result := DifferentiateSafe (leqns, [], drideal, false,
				innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
	catch:
	    error
	end try;
	result := FromExpandedRatfracList (result, eqns)
    elif not IsIdeal (drideal, ':-possibly_empty' = true) then
        error _ERRIDRNG_, "second", convert (drideal, ':-string');
    elif drideal = [] then
# The unit ideal
# Reduced forms are zero
        if type (eqns, '{:-list,:-set}') then
            return map (x -> 0, eqns);
        else
            return 0;
        end if
    else
# The general case
	try
	    leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
	catch:
	    error _ERRPLPL_, "first", convert (eqns, ':-string')
	end try;
	var := 0;
	innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (drideal));
	outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (drideal));
	if type (drideal, list) then
	    regchain := drideal [1];
            arguments := "ReducedForm" 
			(ConvertString (regchain [':-_FREEZESEQ_'], leqns), 
			 ConvertString (regchain [':-_FREEZESEQ_'], var), 
			 op (drideal))
	else
	    regchain := drideal;
            arguments := "ReducedForm" 
			(ConvertString (regchain [':-_FREEZESEQ_'], leqns), 
		 	 ConvertString (regchain [':-_FREEZESEQ_'], var), 
			drideal)
	end if;
        arguments := arguments, 
            ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        try
            result := BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
        catch:
            error
        end try;
        if type (drideal, ':-table') then
            result := map (op, result)
        end if;
        result := FromExpandedRatfracList (result, eqns);
    end if;
    return result;
end proc:

