##PROCEDURE(help) DifferentialAlgebra0[RewriteRules]
##HALFLINE        selects equations and solve them with respect to their leading derivatives
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `RewriteRules`('ideal','criterion','opts')
##- `RewriteRules`('L','criterion','R','opts')
##
##PARAMETERS
##- 'ideal' : a regular differential chain or a list of regular differential chains
##- 'criterion' (optional) : a selection criterion
##- 'L' : a differential polynomial, a list or a set of differential polynomials
##- 'R' : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- This function is a variant of `Equations`. However, it only applies
##  on differential polynomials which depend on at least one *derivative*
##  or independent variable.
##
##- The function call `RewriteRules`('ideal') returns the list of the
##  differential polynomials which constitute the chain if 'ideal' is
##  a regular differential chain. If 'ideal' is a list of regular
##  differential chains, the function call returns a list of lists
##  of differential polynomials. Differential polynomials are 
##  viewed as equations and are solved with respect to their leading 
##  derivatives.
##
##- The function call `RewriteRules`('ideal','criterion') returns only
##  the differential polynomials which satisfy 'criterion'.
##
##- The function call `RewriteRules`('L','criterion','R') returns the list
##  or the set of the differential polynomials of 'L' which satisfy
##  'criterion'. The elements of 'L' should be differential polynomials
##  of 'R', or of its embedding ring if 'R' is an ideal.
##
##- The possible values of 'criterion' are listed below.
##
##-- `order` 'relop' 'n' 
##  where 'relop' stands for one of the relational operators 
##  `>`, `>=`, `<`, `<=`, `=`, `<>` and 'n' is a nonnegative integer.
##  In this case, the function call only returns differential
##  polynomials whose orders satisfy the criterion. 
##  Differential polynomials which do not depend
##  on any *derivative* are not selected.
##
##-- `rank` 'relop' 'rk'
##  where 'relop' stands for one of the relational operators 
##  `>`, `>=`, `<`, `<=`, `=`, `<>` and 'rk' is a rank.
##  The special ranks _0_ and _1_ are allowed.
##  In this case, the function call only returns differential
##  polynomials whose leading ranks satisfy the criterion.
##
##-- `leader` 'relop' 'v'
##  where 'relop' stands for one of the relational operators 
##  `>`, `>=`, `<`, `<=`, `=`, `<>` and 'v' is a *derivative*. 
##  In this case, the function call
##  only returns differential polynomials whose leading derivatives
##  satisfy the criterion.
##  This form applies also to differential
##  polynomials which depend on independent variables only.
##
##-- `leader` 'relop' `derivative`('v') where 'relop' is either 
##  `=` or `<>` and 'v' is a *derivative*.
##  In this case, the function call only returns differential polynomials
##  whose leading derivatives are equal to (or different from) any derivative
##  of 'v'.
##  This form applies also to differential
##  polynomials which depend on independent variables only.
##
##-- `leader` 'relop' `proper`('v') where 'relop' is either 
##  `=` or `<>` and 'v' is a *derivative*.
##  In this case, the function call only returns differential polynomials
##  whose leading derivatives are equal to (or different from) any proper
##  derivative of 'v'.
##  This form applies also to differential
##  polynomials which depend on independent variables only.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `RewriteRules(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[RewriteRules](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call
##  `RewriteRules`('ideal'), applies the function also over
##  the differential polynomials which state that the derivatives of
##  the parameters are zero. Default value is `false`.
##  This option is incompatible with the `diff` notation.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'L'.
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [[u,v],p], parameters = [p]);
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+p,v[x,x]-u[x]], R);
##
##- The two equations stating that the derivatives of _p_ are zero, are omitted.
##
##> RewriteRules(ideal);
##< [[v[y]^2 = 1/2*u^2-u*p+1/2*p^2, u[y]^2 = 2*u, u[x] = 2*u[y]*v[y]/(u-p), v[x,x]= 2*u[y]*v[y]/(u-p)], [p = 0, u = 0, v[x,x] = 0]]
##
##- The two equations stating that the derivatives of _p_ are zero, are present.
##
##> RewriteRules(ideal[1], fullset=true);
##< [p[y] = 0, p[x] = 0, v[y]^2 = 1/2*u^2-u*p+1/2*p^2, u[y]^2 = 2*u, u[x] = 2*u[y]*v[y]/(u-p), v[x,x] = 2*u[y]*v[y]/(u-p)]
##
##- Selects the rewrite rules or *order* greater than _1_.
##
##> RewriteRules(ideal, order>1);
##< [[v[x,x] = 2*u[y]*v[y]/(u-p)], [v[x,x] = 0]]
##
##- Selects the rewrite rules whose leader is a derivative of _u_.
##
##> RewriteRules(ideal[1], leader=derivative(u));
##< [u[y]^2 = 2*u, u[x] = 2*u[y]*v[y]/(u-p)]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative", "LeadingRank", "Equations"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "Equations" : Help:DifferentialAlgebra0[Equations]


# EXPORTED
#
# Family of Equations.

RewriteRules := proc 
   ({fullset::boolean:=false},
    {solved::boolean:=false},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local result, ideal, a_table, keyword, relop, modifier, derv, 
                                            eqns, innot, outnot;
    if _nrest = 1 then
# RewriteRules (ideal)
# The result is a list
        if not IsIdeal (_rest[1]) then
            error _ERRIDEAL_, "first", convert (_rest[1], ':-string')
        end if;
        ideal := _rest[1];
        if ideal = [] then
            return [];
        elif type (ideal, ':-list') then
            a_table := ideal[1];
        else
            a_table := ideal;
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        if outnot = ':-_DIFF_' and fullset then
            error _ERRFSDF_;
        end if;
        try
            if type (ideal, ':-list') then
                result := map (RewriteRulesSafe, ideal, fullset, outnot,
                             _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := RewriteRulesSafe (ideal, fullset, outnot, 
                             _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if
        catch:
            error
        end try
    elif _nrest = 2 and IsIdeal (_rest[1]) then
# RewriteRules (regchain, criterion)
# The result is a sequence
        ideal := _rest[1];
        if ideal = [] then
            return [];
        elif type (ideal, ':-list') then
            a_table := ideal[1];
        else
            a_table := ideal;
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
	if outnot = ':-_DIFF_' and fullset then
            error _ERRFSDF_;
        end if;
        try
            keyword, relop, modifier, derv := 
                                    EquationsCriterion (_rest[2]);
        catch:
            error _ERRCRIT_, "second", convert (_rest[2], ':-string')
        end try;
        try
            if type (ideal, ':-list') then
                result := map (RewriteRulesWithCriterionSafe, ideal,
                    keyword, relop, modifier, derv, a_table, fullset,
                    innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := RewriteRulesWithCriterionSafe 
                            (ideal, keyword, relop, modifier, derv, 
                             a_table, fullset, innot, outnot, 
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if
        catch:
            error
        end try
    elif _nrest = 3 or _nrest = 2 then
# RewriteRules (equations, leader | order, dring)
# The result is a sequence
        if not IsIdealDRing (_rest[_nrest]) then
            error _ERRIDRNG_, "last", convert (_rest[_nrest], ':-string');
        elif type (_rest[_nrest], ':-list') then
            a_table := _rest[_nrest,1];
        else
            a_table := _rest[_nrest];
        end if;
	try
            eqns := ToExpandedRatfracList (_rest[1], ':-no_denom' = true);
	catch:
	    error _ERRPLPL_, "first", convert (_rest[1], ':-string')
	end try;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        try
            if _nrest = 2 then
                keyword, relop, modifier, derv :=
                            ':-_RANK_', ':-_GE_', ':-_IDENTICAL_', 0
            else
                keyword, relop, modifier, derv :=
                                    EquationsCriterion (_rest[2]);
            end if
        catch:
            error _ERRCRIT_, "second", convert (_rest[2], ':-string')
        end try;
        try
            result := RewriteRulesWithCriterionSafe 
                        (eqns, keyword, relop, modifier, derv, a_table, 
                            fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        catch:
            error
        end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    else
        error _ERRNARGS_, "From one to three", _nrest;
    end if;
    return result;
end proc:

# Subfunction of RewriteRules
# RewriteRules (ideal)
#
# If ideal is a regchain, returns the list of the equations.
# If ideal is an intersection of regchains, returns the list of 
# the lists of equations of the regchains.

RewriteRulesSafe := proc (regchain, fullset, outnot, 
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local result, arguments;

    arguments := "RewriteRules" (regchain, fullset);
    arguments := arguments, 
        ToBMIOptionSequence (':-_JET_', outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    result := BMIBladEvalSafe (regchain [':-_UNNUMBERSEQ_'], arguments);
    return result;
end proc:

# Subfunction of RewriteRules
# RewriteRules (list of equations, leader = derivative(u), ideal)
#
# Returns the sequence of the matching equations

RewriteRulesWithCriterionSafe := 
                    proc (list_or_table, keyword, relop, modifier, derv, 
                            a_table, fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "RewriteRulesWithCriterion"
             (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table),
              keyword, relop, modifier, 
	      ConvertString (a_table [':-_FREEZESEQ_'], derv),
              fullset, a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:
