##PROCEDURE(help) DifferentialAlgebra0[RosenfeldGroebner]
##HALFLINE        the RosenfeldGroebner function
##AUTHOR          F. Boulier
##DATE            September 2009
##
##CALLINGSEQUENCE
##
##- RosenfeldGroebner ('eqns', 'ineqns', 'R', 'opts')
##
##PARAMETERS
##
##- 'eqns'   : a list or a set of differential rational fractions
##- 'ineqns' (optional) : a list or a set of differential polynomials
##- 'R'      : a differential polynomial ring
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##
##- Assume that 'eqns' is a list or a set of differential polynomials.
##
##- The function call `RosenfeldGroebner` ('eqns', 'R') returns
##  a representation of the radical of the differential ideal generated
##  by 'eqns', as an intersection of radical differential ideals,
##  presented by regular differential chains, with respect to the
##  ranking of 'R'. 
##
##- By default, the built regular differential chains are assumed to
##  hold the *attributes*: `differential`, `autoreduced`, `primitive`,
##  `squarefree`, `normalized` and `coherent`. 
##  For more details on *attributes*, see "DifferentialAlgebra0".
##
##- The returned list may be empty. In that case, the radical of the
##  differential ideal generated by 'eqns' is the unit ideal, and,
##  'eqns' has no solutions, by [R50, chapter II, Theorem of zeros].
##
##- If 'eqns' is empty 
##  then, the returned representation involves
##  a single regular differential chain, which describes the zero ideal.
##
##- The function call `RosenfeldGroebner` ('eqns', 'ineqns', 'R') returns
##  a representation of the radical of the differential ideal generated
##  by 'eqns', *saturated* by the multiplicative family generated
##  by 'ineqns'. See "DifferentialAlgebra0" for more details.
##
##- By default, the differential polynomials have coefficients in
##  the field _Q_ (_x[1]_, `...`, _x[p]_) obtained by adjoining
##  the independent variables to the field of the rational numbers.
##  In particular, the computed regular differential chains do
##  not involve any differential polynomial which only depends
##  on the independent variables.
##
##- The two lists, 'eqns' and 'ineqns', may involve rational
##  differential fractions. In that case, the denominators of
##  the rational fractions are automatically inserted in 'ineqns'.
##  The two lists may also involve explicit relational
##  operators, i.e, expressions of the forms _p = q_ and _p <> q_. 
##  If _p = q_ is encountered, then the rational fraction _p - q_
##  is inserted in 'eqns'. If _p <> q_ is encountered, then the 
##  rational fraction _p - q_ is inserted in 'ineqns'.
## 
##OPTIONS
##
##- The 'opts' arguments may contain one or more of the options below.
##
##- `attributes` = 'L' where 'L' is a list involving some of the
##  keywords `differential`, `prime`, `primitive`, `squarefree`, `coherent`,
##  `autoreduced`, `normalized`. This option permits to modify the
##  behavior of the function (e.g. if `differential` is omitted, 
##  the decomposition is purely algebraic ; if `normalized` is
##  omitted, the returned regular differential chains do not need
##  to be normalized).
##  The presence of some of the attributes may imply, automatically,
##  the presence of some other ones. 
##  For more details, see "DifferentialAlgebra0".
##
##- `basefield` = `field` (`generators` = 'G', `relations` = 'regchain').
##  This option permits to perform the computation, regarding the elements
##  of 'eqns' and 'ineqns' as differential rational fractions with
##  coefficients in the differential field _k_ presented by the
##  list of *derivatives* 'G' and the regular differential chain 'regchain'.
##  The field _k_ contains the rational numbers.
##  Its set of generators is made of the independent variables, plus
##  all the dependent variables occuring in 'G' or in the differential 
##  polynomials of 'regchain'.
##  Every polynomial expression belonging to the differential ideal
##  defined by 'regchain', is zero in _k_. Every other polynomial
##  expression between the generators of _k_, is invertible in _k_.
##  Notes:
##-- Any of the arguments `generators` = 'G', and, `relations` = 'regchain'
##   can be omitted.
##-- It is required that the generators of _k_ appear at the bottom of
##   the ranking of 'R', and, that any *block* which involves a generator
##   of _k_, purely consists of generators of _k_.
##
##- `singsol` = `all`, `essential`, `none`.
##  This option controls the splittings performed by `RosenfeldGroebner`.
##-- `all` is the default value.
##-- `essential` ensures that the returned decomposition is *irredundant* 
##   whenever this is possible, i.e., whenever 'eqns' is made of only one 
##   element (see "EssentialComponents"). In the other cases, the option
##   is simply ignored.
##   If some *parameters* appear in 'eqns', it is necessary to put them
##   into the base field so that the option applies. 
##-- `none` makes `RosenfeldGroebner` return at most one regular
##   differential chain. The returned chain is the first one that would
##   have been computed without setting this option. In many cases, this
##   first regular differential chain could be considered as the general
##   component, though this notion is not algebraically well-defined
##   in general.
##
##- `dimlb` = `nocase`, `safecase`, `odecase`, `pdecase`.
##  This option controls the splittings performed by `RosenfeldGroebner`
##  by removing any component whose dimension (differential dimension
##  in the differential case) is lower than the number of input equations.
##-- `nocase` disables this option.
##-- `safecase` is the default value. The option is only activated
##   in the non-differential case and in the case of a single input
##   equation. These two cases are theoretically proven.
##-- `odecase` applies also the option to general ODE systems. 
##   The option implements a conjecture in this case [R50, Questions for 
##   investigation, 10].
##-- `pdecase` applies also the option to general PDE systems.
##   The option implements a conjecture in this case.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`.
##  Specifies the notation used for 'F'.
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##- The main features of `RosenfeldGroebner` are illustrated over 
##  academic examples. Less academic examples are provided in
##  the *Motivating examples* section of the main help page of
##  this package.
##
##SUBSECTION An academic example
##
##- This example features an academic system of partial differential
##  equations. The radical _I_ of the differential ideal generated 
##  by the system is decomposed as an intersection of two radical
##  differential ideals, presented by regular differential chains.
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [p,v,u]);
##
##> syst := [u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]];
##< [u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]]
##
##> ideal := RosenfeldGroebner(syst, R);
##
##- The ranking _p_ `>>` _v_ `>>` _u_ eliminates _p_
##  with respect to _v,u_, and, eliminates _v_ with respect to _u_.
##  Each chain involves three differential polynomials.
##  The first one depends on derivatives of _u_ only.
##  The second one depends on derivatives of _u_ and _v_ only.
##  The third one depends on derivatives of all the dependent variables.
##
##> Equations(ideal);
##< [[u[x]^2-4*u, v[x,x]-u[x], 2*p*u+v[y]*u[x]*u[y]-2*u^2], [u, v[x,x], p]]
##
##- Over this example, the intersection is irredundant, therefore
##  _I_ is not prime. 
##  Indeed, _u[x,x]-2_ belongs only to the differential ideal defined
##  by the first chain(its normal form is _0_), _u[x]_ belongs only
##  to the differential ideal defined by the second chain,
##  while, their product belongs to _I_.
##
##> NormalForm([u[x,x]-2, u[x],(u[x,x]-2)*u[x]], ideal);
##< [[0, -2], [u[x], 0], [0, 0]]
##
##- Processing the same system, with respect to another ranking,
##  produces a redundant intersection. To perform changes of rankings
##  more efficiently, see "Pardi".
##
##> R := DifferentialRing(derivations = [x,y], blocks = [[p,v,u]]);
##
##> ideal := RosenfeldGroebner(syst, R);
##
##- To avoid some regular differential chains in the decomposition, it is
##  possible to use the optional argument 'ineqns'. 
##  If a regular differential chain defines a differential ideal which
##  contains some element of 'ineqns', then, it is discarded from the
##  intersection. Putting _u[x,x]-2_ in 'ineqns', one removes the
##  first regular differential chain.
##
##> ideal := RosenfeldGroebner(syst, [u[x,x]-2], R);
##
##> Equations(ideal);
##< [[u, p, v[x, x]]]
##
##- The `singsol` = `none` option provides another efficient method to avoid
##  splittings.
##
##> ideal := RosenfeldGroebner(syst, R, singsol='none');
##
##> Equations (ideal);
##< [[u[x]^2-4*u, 2*v[y]*u[y]+u[x]*p-u[x]*u, v[x,x]-u[x], p[x,x]*u+u[y]^2-2*u]]
##
##ENDSUBSECTION
##
##SUBSECTION Detecting inconsistencies
##
##- Let us still consider the academic system, defining a differential
##  ideal _I_. The differential polynomial _p[x,x,x]_ belongs to _I_,
##  which means that, the equation _p[x,x,x] = 0_ is a consequence
##  of the system.
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [z,p,v,u]):
##> syst := [u[x]^2-4*u = 0, u[x,y]*v[y]-u+p = 0, v[x,x]-u[x] = 0]:
##> ideal := RosenfeldGroebner(syst, R);
##
##> NormalForm(p[x,x,x], ideal);
##
##- Therefore, if one applies the Rabinowitsch trick, and, enlarges the 
##  system with the equation _z * p[x,x,x] = 1_, one gets an inconsistent system.
##  This is proven by `RosenfeldGroebner`, which returns the unit ideal.
##
##> ideal := RosenfeldGroebner([ op(syst), z * p[x,x,x] = 1 ], R);
##
##- The inconsistency can be proven, also, by putting _p[x,x,x]_
##  in 'ineqns'. The `RosenfeldGroebner` function must return the
##  unit ideal.
##
##> ideal := RosenfeldGroebner(syst, [p[x,x,x]], R);
##
##ENDSUBSECTION
##
##SUBSECTION Attributes
##
##- The `attributes` = 'L' option permits to modify the behavior
##  of the function. Here, one omits the `differential` attribute.
##  Therefore, the PDE system is regarded as a non-differential
##  system. The decomposition is purely algebraic: the *derivative*
##  _u[x,y]_ occurring in the third polynomial, is not simplified
##  by the first polynomial.
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [p,v,u]):
##> syst := [u[x]^2-4*u = 0, u[x,y]*v[y]-u+p = 0, v[x,x]-u[x] = 0]:
##> ideal := RosenfeldGroebner(syst, R, attributes = [normalized]);
##
##> RewriteRules(ideal);
##< [[u[x]^2 = 4*u, v[x,x] = u[x], p = -u[x,y]*v[y]+u]]
##
##> Attributes(ideal);
##< [[autoreduced, primitive, normalized]]
##
##ENDSUBSECTION
##
##SUBSECTION Base fields, parameters
##
##- One introduces base fields, by considering the issue of handling
##  constants, without using the `parameters` option of the
##  "DifferentialRing" function. Then, one shows how to obtain
##  the same result, in a simpler way, by using this option.
##
##- Let us still consider the academic system, with two more equations,
##  _p[x]_ = _p[y]_ = _0_
##  stating that _p_ is a constant. Observe that _p_ is now at the
##  bottom of the ranking. The `RosenfeldGroebner` function actually
##  splits cases, by discussing the value of the constant _p_.
##  The first case is the generic one.
##  The second case corresponds to the case _p = 0_.
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [x,y], blocks = [v,u,p]):
##> syst := [u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x], p[x], p[y]]:
##> ideal := RosenfeldGroebner(syst, R);
##
##> RewriteRules(ideal);
##< [[p[y] = 0, p[x] = 0, u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y]*p)/u, v[x,x] = u[x]], [p = 0, u = 0, v[x,x] = 0]]
##
##- To avoid the special case _p = 0_, one method consists in 
##  telling `RosenfeldGroebner` that the only relations that _p_
##  must satisfy are the two relations _p[x]_ = _p[y]_ = _0_.
##  One can proceed by using the `basefield` argument.
##  The dependent variable _p_ is moved to the base field of the 
##  equations, and, constrained by the relations _p[x]_ = _p[y]_ = _0_.
##  The relations that define the base field are included in
##  the returned regular differential chains.
##
##> fieldrels := PretendRegularDifferentialChain([ p[x], p[y]], R);
##
##> ideal := RosenfeldGroebner(syst, R, basefield = field (relations = fieldrels));
##
##> RewriteRules(ideal);
##< [[p[y] = 0, p[x] = 0, u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y]*p)/u, v[x,x] = u[x]]]
##
##- To achieve the same results, a simpler method consists in declaring
##  _p_ as a *parameter*. The two equations _p[x]_ = _p[y]_ = _0_ which
##  state that _p_ is a constant are now automatically taken into account
##  during the computations. Moreover, they do not appear anymore in
##  the results of the "Equations" and "RewriteRules" functions, unless
##  the option `fullset` = `true` is provided.
##
##- If no base field is provided, the `RosenfeldGroebner` function
##  splits cases, by discussing the value of _p_.
##
##> R := DifferentialRing(derivations = [x,y], blocks = [v,u,p], parameters = [p]):
##> syst := [u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]]:
##> ideal := RosenfeldGroebner(syst, R);
##
##> RewriteRules(ideal);
##< [[u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y]*p)/u, v[x,x] = u[x]], [p = 0, u = 0, v[x,x] = 0]]
##
##- To avoid the special case _p = 0_, it is sufficient to move _p_
##  to the base field of the equations. The relations _p[x]_ = _p[y]_ = _0_
##  which define _p_, are automatically taken into account, in the
##  base field definition.
##
##> ideal := RosenfeldGroebner(syst, R, basefield = field (generators = [p]));
##
##> RewriteRules(ideal);
##< [[u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y]*p)/u, v[x,x] = u[x]]]
##
##> RewriteRules(ideal, fullset=true);
##< [[p[y] = 0, p[x] = 0, u[y]^2 = 2*u, u[x]^2 = 4*u, v[y] = -1/4*(-u[x]*u[y]*u+u[x]*u[y]*p)/u, v[x,x] = u[x]]]
##
##ENDSUBSECTION
##
##SUBSECTION The Low Power Theorem
##
##- If the option `singsol` = `essential` is provided, then, the simplifications
##  performed by the "EssentialComponents" function are applied, whenever 
##  this is possible, i.e., whenever the list 'eqns' involves only one 
##  differential polynomial. 
##
##- Over this example, the option permits to remove a redundant component
##  from the intersection.
##
##> R := DifferentialRing(derivations = [x,y], blocks = [v,u,p]):
##> ideal := RosenfeldGroebner([u[x]^2-4*u^3], R);
##
##> Equations(ideal);
##< [[u[x]^2-4*u^3], [u]]
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u^3], R, singsol='essential');
##
##> Equations(ideal);
##< [[u[x]^2-4*u^3]]
##
##ENDSUBSECTION
##
##SEEALSO
##- "DifferentialAlgebra0", "EssentialComponents", "Pardi"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "Indets" : Help:DifferentialAlgebra0[Indets]
##- "FieldElement" : Help:DifferentialAlgebra0[FieldElement]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]
##- "Pardi" : Help:DifferentialAlgebra0[Pardi]
##- "Equations" : Help:DifferentialAlgebra0[Equations]
##- "RewriteRules" : Help:DifferentialAlgebra0[RewriteRules]


# EXPORTED
#
# The RosenfeldGroebner function
# The list equations is either a list of polynomials or a list of
# polynomial equations (the equal signs being interpreted as substractions).
#
# Returns a list of regular chains.

RosenfeldGroebner := proc 
    (equations::{list,set}, 
     {basefield:=':-_U_'},
     {attributes::list(identical(:-differential,:-prime,:-autoreduced,
                    :-primitive,:-squarefree,:-normalized,:-coherent)):=
            [':-differential',':-autoreduced',':-primitive',':-squarefree',
             ':-normalized',':-coherent']}, 
     {singsol::identical(:-_ALL_,:-_NONE_,:-_ESSENTIAL_):=':-_ALL_'},
     {dimlb::identical(:-_NOCASE_,:-_SAFECASE_,
				:-_ODECASE_,:-_PDECASE_):=':-_SAFECASE_'},
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'})

    local dring, result, arguments, eqns, ineqs, gens, rels, innot, outnot;

    if _nrest < 1 or _nrest > 2 then
	error _ERRNARGS_, "Two or three", _nrest+1;
    end if;
# The last parameter must be a differential ring
    dring := _rest[_nrest];
    if not IsDifferentialRing (dring) then
        error _ERRDRNG_, "last", convert (dring, ':-string');
    elif _nrest = 1 then
# Only equations are provided
	try
	    eqns, ineqs := ToPolynomialEquationsAndInequations (equations)
	catch:
	    error _ERRLSEI_, "first", convert (equations, ':-string')
	end try
    else
# Equations and inequations are provided
        try
	    eqns, ineqs := ToPolynomialEquationsAndInequations 
							(equations, _rest[1])
	catch:
	    error _ERRLSEI_, "first and second", 
				convert ([equations, _rest[1]], ':-string')
	end try
    end if;
# Processing the basefield parameter
    try
        gens, rels := BaseField (basefield);
    catch:
	error
    end try;
# Let us go on
    arguments := "RosenfeldGroebner" 
               (ConvertString (dring [':-_FREEZESEQ_'], eqns),
                ConvertString (dring [':-_FREEZESEQ_'], ineqs),
		ConvertString (dring [':-_FREEZESEQ_'], gens),
		ConvertString (dring [':-_FREEZESEQ_'], rels),
                ConvertString (dring [':-_FREEZESEQ_'], attributes),
                dring,
		convert (singsol, ':-string'),
		convert (dimlb, ':-string'),
		true);     # the call comes from RosenfeldGroebner

    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (dring));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (dring));
    arguments := arguments, 
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (dring [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    result := map (ToRegchain, result, outnot, 
			dring [':-_FREEZESEQ_'], dring [':-_UNNUMBERSEQ_']);
    return result;
end proc:
