##PROCEDURE(help) DifferentialAlgebra0[Separant]
##HALFLINE        returns the separant of a differential polynomial 
##AUTHOR          F. Boulier
##DATE            July 2009
##
##CALLINGSEQUENCE
##- `Separant`('ideal', 'v', 'opts')
##- `Separant`('p', 'v', 'R', 'opts')
##- `Separant`('L', 'v', 'R', 'opts')
##
##PARAMETERS
##- 'ideal'        : a differential ideal
##- 'p'            : a differential rational fraction
##- 'v' (optional) : a variable
##- 'L'            : a list or a set of differential rational fractions
##- 'R'            : a differential polynomial ring
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Separant`('p','v','R') returns the 
##  separant of 'p' regarded as a univariate polynomial in 'v', that
##  is the partial derivative of 'p' with respect to 'v'.
##  The differential polynomial 'p' is assumed to be non-numeric.
##
##- The function call `Separant`('L','v','R') returns the list 
##  or the set of the separants of the elements of 'L' with respect to 'v'.
##
##- If 'ideal' is a regular differential chain, the function call 
##  `Separant`('ideal','v') returns the list of the 
##  separants of the chain elements. 
##  If 'ideal' is a list of regular differential chains, 
##  the function call `Separant`('ideal','v') returns a list 
##  of lists of separants.
##
##- When the parameter 'v' is omitted, it is understood to be the
##  leading derivative of the processed differential polynomial with
##  respect to the ranking of 'R', or the one of its embedding
##  polynomial ring, if 'R' is an ideal.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Separant(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Separant](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call 
##   `Separant`('ideal','v'), applies the function also over 
##   the differential 
##   polynomials which state that the derivatives of the parameters 
##   are zero. Default value is `false`.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p', 'L' and 'v'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[x,y], blocks=[[v,u],p], parameters=[p]);
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]], R);
##
##> Equations(ideal[1]);
##< [u[y]^2-2*u, u[x]^2-4*u, 4*v[y]*u-u[x]*u[y]*u+u[x]*u[y]*p, v[x,x]-u[x]]
##
##- The separants, in the usual sense.
##
##> Separant(ideal[1]);
##< [2*u[y], 2*u[x], 4*u, 1]
##
##- The separant with respect to _u[x,y]_
##
##> Separant(u[x,y]*v[y]-u+p, u[x,y], R);
##< v[y]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]


# EXPORTED
# Returns the separant of a differential polynomial with respect to an ordering
#
# Separant (ideal)
# Separant (ideal, variable)
# Separant (poly, differential ring)
# Separant (list(poly), differential ring)
# Separant (poly, variable, differential ring)
# Separant (list(poly), variable, differential ring)

Separant := proc 
       ({fullset::boolean:=false},
        {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_TIMEOUT_::nonnegative:=0},
        {_MEMOUT_::nonnegative:=0},
        {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})
    local ideal, derv, a_table, eqns, innot, outnot, result;
    if _nrest = 1 or (_nrest = 2 and IsIdeal (_rest[1])) then
# Separant (ideal)
# Separant (ideal, variable)
        if not IsIdeal (_rest[1]) then
            error _ERRIDEAL_, "first", convert (_rest[1], ':-string');
#       elif _nrest = 2 and not type (_rest[2], ':-algebraic') then
#           error _ERRDERV_, "second", convert (_rest[2], ':-string');
        end if;
        ideal := _rest[1];
        if ideal = [] then 
            return [];
        end if;
        if _nrest = 1 then
            derv := 0;
        else
            derv := _rest[2];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (ideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (ideal));
        try
            if type (ideal, ':-list') then
                result := map (SeparantSafe, ideal, derv, 
                            ideal[1], fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := SeparantSafe (ideal, derv, 
                            ideal, fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if
        catch:
            error
        end try
    elif _nrest = 2 or _nrest = 3 then
# Separant (poly, differential ring)
# Separant (list(poly), differential ring)
# Separant (poly, variable, differential ring)
# Separant (list(poly), variable, differential ring)
#       if _nrest = 3 and not type (_rest[2], ':-algebraic') then
#           error _ERRDERV_, "second", convert (_rest[2], ':-string');
#       elif not IsIdealDRing (_rest[_nrest]) then
        if not IsIdealDRing (_rest[_nrest]) then
            error _ERRIDRNG_, "last", convert(_rest[_nrest],':-string');
        elif type (_rest[_nrest], ':-list') then
            a_table := _rest[_nrest] [1];
        else
            a_table := _rest[_nrest];
        end if;
        try
            eqns := ToExpandedRatfracList (_rest[1])
        catch:
            error _ERRPLPL_, "first", convert (_rest[1], ':-string')
        end try;
        if _nrest = 2 then
            derv := 0;
        else
            derv := _rest[2];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        try
            result := SeparantSafe (eqns, derv, a_table, fullset, innot, 
                    outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
        catch:
            error
        end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    else
        error _ERRNARGS_, "From one to three", _nrest;
    end if;
    return result;
end proc:

# SeparantSafe (polynomial, derivative, ideal | differential ring, ...)

SeparantSafe := proc (list_or_table, derv, a_table, fullset, innot, outnot,
					    _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "Separant" 
               (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table), 
                fullset, 
		ConvertString (a_table [':-_FREEZESEQ_'], derv),
		a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:
