##PROCEDURE(help) DifferentialAlgebra0[Simplify]
##HALFLINE        A user-friendly version of RosenfeldGroebner
##AUTHOR          F. Boulier
##DATE            December 2009
##
##CALLINGSEQUENCE
##
##- Simplify ('eqns', 'ineqns', 'ders', 'blks', 'arbtr', 'pars', 'nota', 'opts')
##
##PARAMETERS
##
##- 'eqns'   : a list or a set of differential rational fractions
##- 'ineqns' (optional) : a list or a set of differential polynomials
##- 'ders' : a description of the derivations
##- 'blks' : a list of blocks
##- 'arbtr' (optional) : a list of blocks
##- 'pars' (optional) : a description of the *parameters*
##- 'nota' (optional) : a description of the default notation
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call 
##  `Simplify` ('eqns', 'ineqns', 'ders', 'blks', 'arbtr', 'pars', 'nota') 
##  first calls "DifferentialRing" to build a differential polynomial 
##  ring 'R', then, calls "RosenfeldGroebner" to simplify the system
##  of equations and inequations provided by 'eqns' and 'ineqns'.
##
##- The argument 'ders' has the form `derivations` = 'D' where 'D'
##  is a list of independent variables, and provides the derivations.
##  The argument 'blks' has the form `ranking` = 'B1'.
##  The argument 'arbtr' has the form `arbitrary` = 'B2'.
##  Both 'B1' and 'B2' are disjoint lists of *blocks*.
##  Together, the lists 'D', 'B1' and 'B2' provide the *ranking* of
##  the differential polynomial ring (the *blocks* of 'B2' being 
##  ranked lower than the ones of 'B1').
##  The optional argument 'pars' has the form `parameters` = 'P' where 'P'
##  is a list of *parameters*. All these informations, provided to
##  "DifferentialRing", provide a differential polynomial ring 'R'.
##
##- Then "RosenfeldGroebner" is applied over 'eqns', 'ineqns' and 'R'.
##  The computation is performed, considering that the dependent
##  variables in 'arbtr' belong to the base field of the equations.

Simplify := proc
# RosenfeldGroebner options
    (equations::{list,set},
     {basefield:=':-_U_'},
     {attributes::list(identical(:-differential,:-prime,:-autoreduced,
                    :-primitive,:-squarefree,:-normalized,:-coherent)):=
            [':-differential',':-autoreduced',':-primitive',':-squarefree',
             ':-normalized',':-coherent']},
     {singsol::identical(:-_ALL_,:-_NONE_,:-_ESSENTIAL_):=':-_ALL_'},
     {dimlb::identical(:-_NOCASE_,:-_SAFECASE_,
                                :-_ODECASE_,:-_PDECASE_):=':-_SAFECASE_'},
# DifferentialRing options
     {derivations::{list,:-_U_}:=':-_U_'},
     {blocks::{list,:-_U_}:=':-_U_'},
     {ranking::{list,:-_U_}:=':-_U_'},
     {parameters::{list,set,:-_U_}:=':-_U_'},
# Specific options
     {arbitrary::{list,set,:-_U_}:=[]},
     {extensions::{list,set,:-_U_}:=[]},
# Common options
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_LARGE_'})

    local ordering, dring, arguments, prmtrs, result, 
	freezeseq, unnumberseq, actual_blocks, actual_extensions,
	eqns, ineqs, gens, rels, innot, outnot;

# The DifferentialRing parameters
# Plus the arbitrary
    if derivations = ':-_U_' or (blocks = ':-_U_' and ranking = ':-_U_') then
        error _ERRMDB_
    elif blocks <> ':-_U_' and ranking <> ':-_U_' then
        error _ERRRKBL_
    elif blocks <> ':-_U_' then
        actual_blocks := blocks
    else
        actual_blocks := ranking
    end if;
    actual_blocks := [ op (actual_blocks), op (arbitrary) ];
    if parameters = ':-_U_' then
        prmtrs := []
    else
        prmtrs := convert (parameters, ':-list');
    end if;
    try
        freezeseq, unnumberseq := FrozenSymbols (derivations, actual_blocks);
    catch:
        error
    end try;
# Call DifferentialRing
# The notation of the DifferentialRing is the output notation.
    arguments := "DifferentialRing"
	(ConvertString (freezeseq, derivations),
	 ConvertString (freezeseq, actual_blocks),
	 ConvertString (freezeseq, prmtrs));
    arguments := arguments,
	ToBMIOptionSequence
	    (':-_JET_', ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        ordering := BMIBladEvalSafe (unnumberseq, arguments);
    catch:
        error
    end try;

    innot := FixNotation (_INPUT_, _IONOT_);
    outnot := FixNotation (_OUTPUT_, _IONOT_);

    dring := ToDRing (ordering, outnot, freezeseq, unnumberseq);

# Processing equations and inequations
    if _nrest = 0 then
# Only equations are provided
        try
            eqns, ineqs := ToPolynomialEquationsAndInequations (equations)
        catch:
            error _ERRLSEI_, "first", convert (equations, ':-string')
        end try
    elif _nrest = 1 then
# Equations and inequations are provided
        try
            eqns, ineqs := ToPolynomialEquationsAndInequations
                                                        (equations, _rest[1])
        catch:
            error _ERRLSEI_, "first and second",
                                convert ([equations, _rest[1]], ':-string')
        end try
    else
	error _ERRNARGS_, "One or two", _nrest+1
    end if;

# Processing the basefield 
    try
        actual_extensions := 
		PretendRegularDifferentialChain (extensions, dring)
    catch:
	error _ERRLSEI_, "extensions", convert (extensions, ':-string')
    end try;
    try
	gens, rels := BaseField 
		(':-field' (':-generators' = arbitrary,
			    ':-relations' = actual_extensions))
    catch:
        error
    end try;

# Let us go on
    arguments := "RosenfeldGroebner"
               (ConvertString (dring [':-_FREEZESEQ_'], eqns),
                ConvertString (dring [':-_FREEZESEQ_'], ineqs),
                ConvertString (dring [':-_FREEZESEQ_'], gens),
                ConvertString (dring [':-_FREEZESEQ_'], rels),
                ConvertString (dring [':-_FREEZESEQ_'], attributes),
                dring,
                convert (singsol, ':-string'),
                convert (dimlb, ':-string'),
                true);     # the call comes from RosenfeldGroebner

    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (dring [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    result := map (ToRegchain, result, outnot,
                        dring [':-_FREEZESEQ_'], dring [':-_UNNUMBERSEQ_']);
    return result;
end proc:
