##PROCEDURE(help) DifferentialAlgebra0[SortByRank]
##HALFLINE        sorts a list of differential polynomials
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- `SortByRank`('L','criterion','R','opts')
##
##PARAMETERS
##- 'L' : a list or a set of differential polynomials
##- 'criterion' : one of the keywords `ascending` or `descending`
##- 'R' : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `SortByRank`('L','criterion','R') returns the
##  list of the elements of 'L', sorted according to 'criterion'.
##  The leading ranks of the elements of 'L' are compared.
##  Leading ranks are taken with respect to the ranking of 'R',
##  or its embedding ring, if 'R' is an ideal.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `SortByRank(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[SortByRank](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`.
##  Specifies the notation used for 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation used for the result of the function call. If not specified,
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [t], blocks = [u,v]);
##
##> SortByRank([u[t], 3, v[t,t], 0, t^2 + 1], ascending, R);
##< [0, 3, t^2+1, v[t,t], u[t]]
##
##> SortByRank([u[t], 3, v[t,t], 0, t^2 + 1], descending, R);
##< [u[t], v[t,t], t^2+1, 3, 0]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingRank", "MaxRankElement", "MinRankElement"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingRank" : Help:DifferentialAlgebra0[LeadingRank]
##- "MinRankElement" : Help:DifferentialAlgebra0[MinRankElement]
##- "MaxRankElement" : Help:DifferentialAlgebra0[MaxRankElement]


# EXPORTED
#
# SortByRank (list(polynomials), ascending | descending, differential ring)
#
# Returns a list.

SortByRank := proc (eqns::{list,set},
              ord::identical(:-_ASC_,:-_DESC_), 
              drideal,
        {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
        {_TIMEOUT_::nonnegative:=0},
        {_MEMOUT_::nonnegative:=0},
        {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'},$)
    local result, arguments, a_table, leqns, innot, outnot;
    if not IsIdealDRing (drideal) then
            error _ERRIDRNG_, "third", convert (_rest[3], ':-string');
    elif type (drideal,list) then
        a_table := drideal[1];
    else
        a_table := drideal;
    end if;
    try
        leqns := ToExpandedRatfracList (eqns, ':-no_denom' = true)
    catch:
	error _ERRLPOL_, "first", convert (eqns, ':-string')
    end try;
    arguments := "SortByRank"
                    (ConvertString (a_table [':-_FREEZESEQ_'], leqns), 
                     ConvertString (a_table [':-_FREEZESEQ_'], ord), 
                     a_table);
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    try
        result := BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
    catch:
        error
    end try;
    return result;
end proc:
