##PROCEDURE(help) DifferentialAlgebra0[Tail]
##HALFLINE        returns the tail of a differential polynomial 
##AUTHOR          F. Boulier
##DATE            July 2009
##
##CALLINGSEQUENCE
##- `Tail`('ideal', 'v', 'opts')
##- `Tail`('p', 'v', 'R', 'opts')
##- `Tail`('L', 'v', 'R', 'opts')
##
##PARAMETERS
##- 'ideal'        : a differential ideal
##- 'p'            : a differential rational fraction
##- 'v' (optional) : a variable
##- 'L'            : a list or a set of differential rational fractions
##- 'R'            : a differential polynomial ring or ideal
##- 'opts' (optional) : a sequence of options
##
##DESCRIPTION
##- The function call `Tail`('p','v','R') returns the 
##  tail of 'p' regarded as a univariate polynomial in 'v', that
##  is the differential polynomial 'p', regarded as a univariate
##  polynomial in 'v', minus its leading monomial with respect to 
##  this variable,
##  If 'p' does not depend on 'v' then the function call returns _0_.
##
##- The function call `Tail`('L','v','R') returns the list 
##  or the set of the tails of the elements of 'L' with respect to 'v'.
##
##- If 'ideal' is a regular differential chain, the function call 
##  `Tail`('ideal','v') returns the list of the 
##  tails of the chain elements. 
##  If 'ideal' is a list of regular differential chains, 
##  the function call `Tail`('ideal','v') returns a list 
##  of lists of tails.
##
##- When the parameter 'v' is omitted, it is understood to be the
##  leading derivative of the processed differential polynomial with
##  respect to the ranking of 'R', or the one of its embedding polynomial
##  ring, if 'R' is an ideal. In that case, 'p' must be non-numeric.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `Tail(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[Tail](...)`.
##
##OPTIONS
##- The 'opts' arguments may contain one or more of the options below.
##
##- `fullset` = ::boolean::. In the case of the function call 
##   `Tail`('ideal','v'), applies the function also over 
##   the differential 
##   polynomials which state that the derivatives of the parameters 
##   are zero. Default value is `false`.
##
##- `input` = `jet`, `tjet`, `diff` or `Diff`. 
##  Specifies the notation used for 'p', 'v' and 'L'. 
##  If not specified, the notation of 'R' is used (see "Notation").
##
##- `output` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation used for the result of the function call. If not specified, 
##  the notation of 'R' or of 'ideal' is used.
##
##- `notation` = `jet`, `tjet`, `diff` or `Diff`. Specifies the 
##  notation
##  used for both input and output. This option is overridden by the two
##  options above.
##
##SECTION(nohelp) Other options
##
##- `timeout` = ::nonnegative::. Specifies a time limit, in seconds, for the 
##  computation. Default is zero (no time out).
##
##- `memout` = ::nonnegative::. Specifies a memory limit, in MB, for the 
##  computation. Default is zero (no memory out).
##
##- `cellsize` = `small` or `large`. Specifies the size of the cells that
##  must be allocated by the BLAD libraries.
##
##EXAMPLES
##
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations=[x,y], blocks=[[v,u],p], parameters=[p]);
##
##- The tail, with respect to the leading derivative
##
##> Tail(u[x,y]*v[y]-u+p, R);
##< -u+p
##
##> ideal := RosenfeldGroebner([u[x]^2-4*u, u[x,y]*v[y]-u+p, v[x,x]-u[x]], R);
##
##> Equations(ideal[1]);
##< [u[y]^2-2*u, u[x]^2-4*u, 4*v[y]*u-u[x]*u[y]*u+u[x]*u[y]*p, v[x,x]-u[x]]
##
##- The tails of the equations, with respect to _u[x]_
##
##> Tail(ideal[1], u[x]);
##< [0, -4*u, 4*v[y]*u, v[x,x]]
##
##SEEALSO
##- "DifferentialAlgebra0", "LeadingDerivative", "LeadingCoefficient"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "LeadingCoefficient" : Help:DifferentialAlgebra0[LeadingCoefficient]
##- "Initial" : Help:DifferentialAlgebra0[Initial]
##- "LeadingDerivative" : Help:DifferentialAlgebra0[LeadingDerivative]

# EXPORTED
#
# Family of Separant

Tail := proc 
   ({fullset::boolean:=false},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})

    local ideal, derv, eqns, a_table, innot, outnot, result;

    if _nrest = 1 or (_nrest = 2 and IsIdeal (_rest[1])) then
# Tail (ideal)
# Tail (ideal, variable)
        if not IsIdeal (_rest[1]) then
            error _ERRIDEAL_, "first", convert (_rest[1], ':-string');
        elif _nrest = 2 and not type (_rest[2], algebraic) then
            error _ERRDERV_, "second", convert (_rest[2], ':-string');
        end if;
        ideal := _rest[1];
        if ideal = [] then
            return [];
        end if;
        if _nrest = 1 then
            derv := 0;
        else
            derv := _rest[2];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (ideal));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (ideal));
        try
            if type (ideal, ':-list') then
                result := map (TailSafe, ideal, derv, 
                            ideal [1], fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            else
                result := TailSafe (ideal, derv, 
                            ideal, fullset, innot, outnot,
                            _TIMEOUT_, _MEMOUT_, _CELLSZ_);
            end if;
        catch:
            error
        end try;
    elif _nrest = 2 or _nrest = 3 then
# Tail (poly, differential ring)
# Tail (list(poly), differential ring)
# Tail (poly, variable, differential ring)
# Tail (list(poly), variable, differential ring)
        if _nrest = 3 and not type (_rest[2], ':-algebraic') then
            error _ERRDERV_, "second", convert (_rest[2], ':-string');
        elif not IsIdealDRing (_rest[_nrest]) then
            error _ERRIDRNG_, "last", convert(_rest[_nrest],':-string');
        elif type (_rest[_nrest], ':-list') then
            a_table := _rest[_nrest,1];
        else
            a_table := _rest[_nrest];
        end if;
        try
            eqns := ToExpandedRatfracList (_rest[1])
        catch:
            error _ERRPLPL_, "first", convert (_rest[1], ':-string')
        end try;
        if _nrest = 2 then
            derv := 0;
        else
            derv := _rest[2];
        end if;
        innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
        outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
        try
            result := TailSafe (eqns, derv, a_table, fullset, innot, 
                    outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
        catch:
            error
        end try;
        result := FromExpandedRatfracList (result, _rest[1]);
    else
        error _ERRNARGS_, "From one to three", _nrest;
    end if;
    return result;
end proc:

# Family of SeparantSafe

TailSafe := proc (list_or_table, derv, a_table, fullset, innot, outnot,
					    _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;
    arguments := "Tail" 
              (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], list_or_table), 
               fullset, 
	       ConvertString (a_table [':-_FREEZESEQ_'], derv), 
	       a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:
