# BaseFieldExtension
cdef class BaseFieldExtension:
    r"""
    The class BaseFieldExtension implements base field extensions for differential rings

        These field extensions are defined by generators and relations. 
        They permit to control the splittings performed by RosenfeldGroebner.

        Observe that a base field extension F can be defined without
        specifying the differential ring R = K{U} whose base field K
        is going to be extended. Indeed, the actual extension of K
        is performed by the RosenfeldGroebner algorithm.
 
        The generators of the base field extension of a differential ring
        must belong to the lowest blocks of the ranking (i.e. the block list).
        If the differential ring is not specified at the base field extension
        definition, the test is performed by the RosenfeldGroebner algorithm.

    EXAMPLES:

    First create a differential ring with two parameters $a$, $b$.
    The base field of $R$ is $K = Q(x)$, where $Q$ denotes the field
    of the rational numbers. 

        >>> from sympy import *
        >>> from DifferentialAlgebra import *
        >>> x,a,b = var('x,a,b')
        >>> y = function ('y')
        >>> R = DifferentialRing (derivations = [x], blocks = [y,[a,b]], parameters = [a,b])

    Create now F = Q(a,b) a purely transcendental field extension 
    of the field Q.
    
        >>> from DifferentialAlgebra import DifferentialRing, RegularDifferentialChain, BaseFieldExtension
        >>> F = BaseFieldExtension (generators = [a,b])
        >>> F.generators ()
        [a, b]
        >>> F.relations ()
        regular_differential_chain
        >>> F.relations ().equations ()
        []

    Alternatively, since we know in advance that F is going to be used 
    as a base field extension of the sole ring R, it is safe to specify   
    the ring as an optional parameter, in order to check that a and b
    lie at the bottom of the ranking of R.

        >>> F = BaseFieldExtension (generators = [a,b], ring = R)
        >>> F
        differential_field

    In the next example (F is not yet used), the RosenfeldGroebner 
    algorithm considers three cases.

        >>> L = R.RosenfeldGroebner ([Derivative(y(x),x)**2-a*y(x)])
        >>> [ C.equations () for C in L ]
        [[-a*y(x) + Derivative(y(x), x)**2], [a, Derivative(y(x), x)], [y(x)]]

    In this variant, the RosenfeldGroebner algorithm is applied
    over the base field F(x) = K(a,b) = Q(a,b,x), obtained by
    extending the base field K of R with the two generators a and b.
    The case which considers the vanishing of a is omitted: as a 
    nonzero element of the base field, a cannot vanish.

        >>> L = R.RosenfeldGroebner ([Derivative(y(x),x)**2-a*y(x)], basefield = F)
        >>> [ C.equations () for C in L ]
        [[-a*y(x) + Derivative(y(x), x)**2], [y(x)]]

    The next example illustrates field extensions defined by generators
    and relations. Denote f(x,y) = \tan(x+y**2). Then f(x,y) satisfies 
    the equations of the regular differential chain C. 
    The differential field F is generated (in the differential sense)
    by x, y and f(x,y).  The generator f(x,y) satisfies the 
    relations of C.

        >>> from DifferentialAlgebra import DifferentialRing, RegularDifferentialChain, BaseFieldExtension
        >>> x,y = var ('x,y')
        >>> f,u = function ('f,u')
        >>> R = DifferentialRing (derivations = [x,y], blocks = [u,f], parameters = [u(x)])
        >>> eqns = [Eq(Derivative(f(x,y),x),f(x,y)**2 + 1), Eq(Derivative(f(x,y),y), 2*y*(f(x,y)**2 + 1))]
        >>> C = RegularDifferentialChain (eqns, R, pretend=False)
        >>> F = BaseFieldExtension (relations = C)
        >>> F.generators ()
        [f(x, y)]
        >>> F.relations ().equations ()
        [-2*y*f(x, y)**2 - 2*y + Derivative(f(x, y), y), -f(x, y)**2 + Derivative(f(x, y), x) - 1]

    Again, the base field F permits to avoid some cases in the output of f
    the RosenfeldGroebner algorithm. In the next example (F is not used), 
    two cases are computed.
    
        >>> eqn = Eq(Derivative(u(x),x)**2-(f(x,y)**2+1)*u(x),0)
        >>> L = R.RosenfeldGroebner ([eqns[0], eqns[1], eqn])
        >>> [ C.equations () for C in L ]
        [[f(x, y)**2 + 1, Derivative(u(x), x)], [-2*y*f(x, y)**2 - 2*y + Derivative(f(x, y), y), -f(x, y)**2 + Derivative(f(x, y), x) - 1, u(x)]]

    In the next variant, a single case is computed. The RosenfeldGroebner
    algorithm is run over F. As a nonzero element of F, the polynomial l
    f(x, y)**2 + 1 cannot vanish. The only solution of the
    equations is thus u(x) = 0.

        >>> L = R.RosenfeldGroebner ([eqn], basefield = F)
        >>> [ C.equations () for C in L ]
        [[-2*y*f(x, y)**2 - 2*y + Derivative(f(x, y), y), -f(x, y)**2 + Derivative(f(x, y), x) - 1, u(x)]]

    """
    cdef bool rels_are_provided
    cdef list gens

    cdef notation
#   The default notation

    cdef public object evdict
#   A dictionary used by eval

    cdef RegularDifferentialChain rels
#   rels_are_provided is set to true if relations were provided

# __process_notation
    def __process_notation (self, notation) :
        cdef strnot
        if not isinstance (notation, str) :
            strnot = self.notation
        else :
            strnot = ascii (notation)
            if strnot != BMI_IX_Derivative and strnot != BMI_IX_jet:
                strnot = self.notation
        return strnot

# __init__
    def __init__ (
            self, 
            list generators = [], 
            RegularDifferentialChain relations = None, 
            DifferentialRing ring = None,
            notation = BMI_IX_undefined):
        r"""
        The constructor 

        INPUT:

            ``generators`` -- (default: []) a list of dependent variables
            ``relations``  -- (optional) a regular differential chain
            ``ring``       -- (optional) a differential ring

        OUTPUT:

            An extension F of the base field K of a differential
            ring, defined by ``generators`` and ``relations``. 
            If ``ring`` is not specified, F actually is the
            description of a base field extension, which will be
            performed by a future call to the RosenfeldGroebner 
            algorithm.

            If ``relations`` is not specified, then F is the
            differential field, obtained by extending the field of
            the rational numbers, by the elements of ``generators``.

            If ``relations`` is specified, then F is the differential
            field of fractions of Q{X}/I where X denotes the variables 
            (dependent and independent) present in ``generators``, 
            ``relations``, and I denotes the differential ideal (which 
            needs to be prime) defined by ``relations``.

        NOTES:

            The variables present in ``generators`` and ``relations``
            should belong to the lowest blocks of the ranking of the
            differential ring, whose base field is extended.

        EXAMPLES:

        The field F of the rational numbers

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> F = BaseFieldExtension ()

        The field F = Q(a,b)

            >>> a,b = var ('a,b')
            >>> F = BaseFieldExtension (generators = [a,b])

        The field Q(a,b) again. The differential ring R is provided.
        This permits to check that [a,b] is the lowest block of the
        ranking of R.

            >>> y = function ('y')
            >>> R = DifferentialRing (derivations = [], blocks = [y,[a,b]])
            >>> F = BaseFieldExtension (generators = [a,b], ring = R)

        The field of fractions of Q[a,b]/(a**2 + 1)

            >>> C = RegularDifferentialChain ([a**2 + 1], R)
            >>> F = BaseFieldExtension (generators = [a,b], relations = C)
        """
#
        cdef DifferentialRing R
        cdef bytes strgens, strrels, mesgerr
        cdef bmi_c.ALGEB_string A
        if (relations == None and ring == None):
            self.notation = BMI_IX_Derivative
            self.notation = self.__process_notation (notation)
            R = DifferentialRing (derivations = [], blocks = generators, notation = self.notation)
            self.rels_are_provided = False
            self.rels = RegularDifferentialChain ([], R)
            self.gens = generators 
            self.evdict = {}
        elif relations != None:
            if not isinstance (relations, RegularDifferentialChain):
                raise TypeError ('RegularDifferentialChain expected for relations keyword parameter')
            self.rels_are_provided = True
            self.rels = relations
            self.gens = generators
            self.notation = relations.notation
            self.notation = self.__process_notation (notation)
            self.evdict = relations.evdict
            strrels = ascii (self.__relations ())
            strgens = ascii (generators)
            A = bmi_c.bmi_dapyx_base_field_generators (
                        strgens, strrels, self.rels.regchain, 
                        self.notation, self.notation, 0, 0)
            if bmi_c.bmi_dapyx_is_error (A):
                mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
                bmi_c.bmi_balsa_clear_ALGEB (A)
                raise RuntimeError (str (mesgerr, 'utf-8'))
            self.gens = eval (A.value, globals (), self.evdict)
        elif ring != None:
            if not isinstance (ring, DifferentialRing):
                raise TypeError ('DifferentialRing expected for ring keyword parameter')
            R = ring
            self.rels_are_provided = False
            self.rels = RegularDifferentialChain ([], R)
            self.gens = generators
            self.notation = ring.notation
            self.notation = self.__process_notation (notation)
            self.evdict = ring.evdict
            strrels = ascii (self.__relations ())
            strgens = ascii (generators)
            A = bmi_c.bmi_dapyx_base_field_generators (
                        strgens, strrels, R.dring, 
                        self.notation, self.notation, 0, 0)
            if bmi_c.bmi_dapyx_is_error (A):
                mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
                bmi_c.bmi_balsa_clear_ALGEB (A)
                raise RuntimeError (str (mesgerr, 'utf-8'))
            self.gens = eval (A.value, globals (), self.evdict)

    def __repr__ (self):
#   The external representation
        return 'differential_field'

    def _latex_ (self):
#   The LaTeX representation
        return 'differential\\_field'
# generators
    def generators (self):
        r""" 
        The list of generators of a base field extension

        INPUT:

            Nothing

        OUTPUT:

            The list of generators

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> a,b = var ('a,b')
            >>> R = DifferentialRing (derivations = [], blocks = [a,b])
            >>> C = RegularDifferentialChain ([a**2 + 1], R)
            >>> F = BaseFieldExtension (generators = [a,b], relations = C)
            >>> F.generators ()
            [a, b]
        """
#
        return self.gens
# relations
    def relations (self):
        r""" 
        The regular differential chain defining the relations of 
        a base field extension

        INPUT:

            Nothing

        OUTPUT:

            A regular differential chain. If no relations were provided
            at the field construction, the regular differential chain
            is the zero chain.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> a,b = var ('a,b')
            >>> R = DifferentialRing (derivations = [], blocks = [a,b])
            >>> C = RegularDifferentialChain ([a**2 + 1], R)
            >>> F = BaseFieldExtension (generators = [a,b], relations = C)
            >>> F.relations ()
            regular_differential_chain
            >>> F.relations ().equations ()
            [a**2 + 1]
        """
#
        return self.rels

# __relations
    def __relations (self):
        r""" 
        The relations in a BMI compatible format, as a string 

        INPUT:

            Nothing

        OUTPUT:

            A string

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> a,b = var ('a,b')
            >>> R = DifferentialRing (derivations = [], blocks = [a,b])
            >>> C = RegularDifferentialChain ([a**2 + 1], R)
            >>> F = BaseFieldExtension (generators = [a,b], relations = C)
        """
#
        cdef str strrels, strattr
        if self.rels_are_provided:
            strrels = str (self.rels.equations ())
            strattr = str (self.rels.attributes ())
            strrels = strrels.replace ("'", "")
            strattr = strattr.replace ("'", "")
            return 'regchain (%s, %s)' % (strrels, strattr)
        else:
            return 'regchain ([], [])'

