r"""
The DifferentialAlgebra module

The DifferentialAlgebra module implements elimination methods for systems
of polynomial differential equations. Mathematically, it relies on the
differential algebra theory of Joseph Fels Ritt and Ellis Robert Kolchin.

AUTHORS:

- Nicolas M. Thiery (2011-12): initial version

- François Boulier (2012): the only other developer, so far

REFERENCES:

- [Ritt50] Joseph Fels Ritt. Differential Algebra. Dover Publications Inc. .
  New York, 1950.

- [Kolchin73] Ellis Robert Kolchin. Differential Algebra and Algebraic Groups.
  Academic Press. New York, 1973.

EXAMPLES:

        >>> from sympy import *
        >>> from DifferentialAlgebra import *
        >>> leader,order,rank = var ('leader,order,rank')
        >>> derivative = function ('derivative')

    This example shows how to build the Henri Michaelis Menten formula
    by differential elimination. One considers a chemical reaction
    system describing the enzymatic reaction:

                   k1
        E + S  -----------> ES
                   km1
        ES     -----------> E + S
                   k2
        ES     -----------> E + P

    A substrate S is transformed into a product P, in the presence
    of an enzyme E. An intermediate complex ES is formed. 

        >>> t = var('t')
        >>> km1,k1,k2 = var('km1,k1,k2')
        >>> F1,E,S,ES,P = function('F1,E,S,ES,P')
        >>> params = [km1,k1,k2]
        >>> params
        [km1, k1, k2]

    The main assumption is that k1, km1 >> k2 i.e. that the
    revertible reaction is much faster than the last one. One performs
    a quasi-steady state approximation by considering the following
    differential-algebraic system (it comes from the mass-action law
    kinetics, replacing the contribution of the fast reactions by an
    unknown function $F1(t)$, on the algebraic variety where the fast
    reaction would equilibrate if they were alone).

        >>> syst = [Derivative(E(t),t) + F1(t) - k2*ES(t), Derivative(S(t),t) + F1(t), Derivative(ES(t),t) + k2*ES(t) - F1(t), Derivative(P(t),t) - k2*ES(t), k1*E(t)*S(t) - km1*ES(t) ]
        >>> syst
        [-k2*ES(t) + F1(t) + Derivative(E(t), t), F1(t) + Derivative(S(t), t), k2*ES(t) - F1(t) + Derivative(ES(t), t), -k2*ES(t) + Derivative(P(t), t), k1*E(t)*S(t) - km1*ES(t)]

    Differential elimination permits to simplify this DAE. 
    To avoid discussing the possible vanishing of ``params``, one
    moves them to the base field of the equations.

        >>> Field = BaseFieldExtension (generators = params)
        >>> Field
        differential_field

        >>> R = DifferentialRing (derivations = [t], blocks = [F1, [E,ES,P,S], params], parameters = params)
        >>> R
        differential_ring

    The RosenfeldGroebner considers three cases. The two last ones
    are degenerate cases.

        >>> ideal = R.RosenfeldGroebner (syst, basefield = Field)
        >>> ideal
        [regular_differential_chain, regular_differential_chain, regular_differential_chain]
        >>> [ C.equations (solved = True) for C in ideal ]
        [[Eq(E(t), km1*ES(t)/(k1*S(t))), Eq(Derivative(S(t), t), (-k1*k2*ES(t)*S(t)**2 - k2*km1*ES(t)*S(t))/(k1*S(t)**2 + km1*ES(t) + km1*S(t))), Eq(Derivative(P(t), t), k2*ES(t)), Eq(Derivative(ES(t), t), -k2*km1*ES(t)**2/(k1*S(t)**2 + km1*ES(t) + km1*S(t))), Eq(F1(t), (k1*k2*ES(t)*S(t)**2 + k2*km1*ES(t)*S(t))/(k1*S(t)**2 + km1*ES(t) + km1*S(t)))], [Eq(S(t), -km1/k1), Eq(ES(t), 0), Eq(E(t), 0), Eq(Derivative(P(t), t), 0), Eq(F1(t), 0)], [Eq(S(t), 0), Eq(ES(t), 0), Eq(Derivative(P(t), t), 0), Eq(Derivative(E(t), t), 0), Eq(F1(t), 0)]]

    The sought equation, below, is not yet the Henri-Michaelis-Menten
    formula. This is expected, since some minor hypotheses have not yet
    been taken into account

        >>> ideal [0].equations (solved = True, selection = Eq(leader,derivative (S(t))))
        [Eq(Derivative(S(t), t), (-k1*k2*ES(t)*S(t)**2 - k2*km1*ES(t)*S(t))/(k1*S(t)**2 + km1*ES(t) + km1*S(t)))]

    Let us take them into account. First create two new constants.
    Put them among ``params``, together with initial values.

        >>> E0,ES0,P0,S0,K,V_max = var ('E0,ES0,P0,S0,K,V_max')
        >>> params = [km1,k1,k2,E0,ES0,P0,S0,K,V_max]
        >>> params
        [km1, k1, k2, E0, ES0, P0, S0, K, V_max]

        >>> R = DifferentialRing (blocks = [F1, [ES,E,P,S], params], parameters = params, derivations = [t])
        >>> R
        differential_ring

    There are relations among the parameters: initial values supposed
    to be zero, and equations meant to rename constants.

        >>> relations_among_params = RegularDifferentialChain ([P0, ES0, Eq(K,km1/k1), Eq(V_max,k2*E0)], R)
        >>> relations_among_params
        regular_differential_chain

    Coming computations will be performed over a base field defined 
    by generators and relations

        >>> Field = BaseFieldExtension (generators = params, relations = relations_among_params)
        >>> Field
        differential_field

    Extend the DAE with linear conservation laws. They could have
    been computed from the stoichimetry matrix of the chemical system.

        >>> newsyst = syst + [Eq(E(t) + ES(t), E0 + ES0), Eq(S(t) + ES(t) + P(t), S0 + ES0 + P0)]
        >>> newsyst
        [-k2*ES(t) + F1(t) + Derivative(E(t), t), F1(t) + Derivative(S(t), t), k2*ES(t) - F1(t) + Derivative(ES(t), t), -k2*ES(t) + Derivative(P(t), t), k1*E(t)*S(t) - km1*ES(t), Eq(E(t) + ES(t), E0 + ES0), Eq(ES(t) + P(t) + S(t), ES0 + P0 + S0)]

    Simplify again. Only one case is left.

        >>> ideal = R.RosenfeldGroebner (newsyst, basefield = Field)
        >>> ideal
        [regular_differential_chain]

    To get the traditional Henri-Michaelis-Menten formula, one still needs 
    to neglect the term K*E0

        >>> ideal[0].equations (solved = True, selection = Eq(leader,derivative (S(t)))) )
        [Eq(Derivative(S(t), t), (-K*V_max*S(t) - V_max*S(t)**2)/(E0*K + K**2 + 2*K*S(t) + S(t)**2))]

    One can also get it by computing the right hand side of the equation
    which gives the evolution of the product $P$

        >>> ideal[0].normal_form (Derivative(P(t),t))
        V_max*S(t)/(K + S(t))

REFERENCES:

    [BLLM11] François Boulier, Marc Lefranc, François Lemaire and
    Pierre-Emmanuel Morant. Model Reduction of Chemical Reaction 
    Systems using Elimination. Mathematics in Computer Science vol. 5,
    pp. 289-301. 2011. http://hal.archives-ouvertes.fr/hal-00184558

EXAMPLES:

        >>> from sympy import *
        >>> from DifferentialAlgebra import *
        >>> leader,order,rank = var ('leader,order,rank')
        >>> derivative = function ('derivative')

    This example shows how to compute an input/output relation, having
    a nice form, for a compartmental model. The degradation of the
    studied product, from compartment 1, is supposed to follow a 
    Henri-Michaelis-Menten formula. The exchanges between the two
    compartments are supposed to be linear. 

                          ke,Ve
        compartment 1  --------------> outside the model

                            k12
        compartment 1 ---------------> compartment 2

                            k21
        compartment 2 ---------------> compartment 1

    Functions $x_1(t)$ and $x_2(t)$ are associated to both compartments.

        >>> t = var ('t')
        >>> ke, Ve, k12, k21 = var ('ke, Ve, k12, k21')
        >>> x1,x2 = function ('x1,x2')
        >>> params = [ke,Ve,k12,k21]
        >>> params
        [ke, Ve, k12, k21]

    The dynamical system associated to the above compartmental model
    already is a regular differential chain w.r.t. any orderly ranking,
    such as the one of $R$:

        >>> R = DifferentialRing (derivations = [t], blocks = [[x1,x2], params], parameters = params)
        >>> R
        differential_ring

    Here are the equations which give the dynamics of the system:

        >>> edoA = Eq(Derivative(x1(t),t), -k12*x1(t) + k21*x2(t) - (Ve*x1(t))/(ke + x1(t)))
        >>> edoA
        Eq(Derivative(x1(t), t), -Ve*x1(t)/(ke + x1(t)) - k12*x1(t) + k21*x2(t))

        >>> edoB = Eq (Derivative(x2(t),t), k12*x1(t) - k21*x2(t))
        >>> edoB
        Eq(Derivative(x2(t), t), k12*x1(t) - k21*x2(t))

    The parameters are moved to the base field of the equations, in
    order to avoid discussing their possible vanishing.

        >>> F = BaseFieldExtension (generators = params)
        >>> F
        differential_field

    The following computation does not do anything. The two equations
    are just bundled into a regular differential chain.

        >>> ideal = R.RosenfeldGroebner ([edoA, edoB], basefield = F)
        >>> ideal
        [regular_differential_chain]

        >>> C = ideal[0]
        >>> C.equations (solved = True)
        [Eq(Derivative(x2(t), t), k12*x1(t) - k21*x2(t)), Eq(Derivative(x1(t), t), (-Ve*x1(t) - k12*ke*x1(t) - k12*x1(t)**2 + k21*ke*x2(t) + k21*x1(t)*x2(t))/(ke + x1(t)))]

     Let us assume now that compartment 1 is observed while compartment 2
     is not. In other words, let us assume that the output $y = x_1(t)$.
     The idea is to eliminate the non observed variable $x_2(t)$ and
     compute a relation (the input/output equation), which is a consequence
     of the dynamical system, but does not involve $x_2(t)$ and its
     derivatives. For this purpose, one defines a new differential ring, 
     which is mathematically equivalent to $R$, but with a different ranking.

        >>> IO_R = DifferentialRing (derivations = [t], blocks = [x2,x1,params], parameters = params)
        >>> IO_R
        differential_ring

     One just has to perform a change of ranking over $C$

        >>> IO_C = C.change_ranking (IO_R)
        >>> IO_C
        regular_differential_chain

    Here is the input/output equation. However, it looks quite complicated.

        >>> IO_rel = IO_C.equations (selection = Eq(leader,derivative(x1(t))))[0]
        >>> IO_rel
        Ve*k21*ke*x1(t) + Ve*k21*x1(t)**2 + Ve*ke*Derivative(x1(t), t) + k12*ke**2*Derivative(x1(t), t) + 2*k12*ke*x1(t)*Derivative(x1(t), t) + k12*x1(t)**2*Derivative(x1(t), t) + k21*ke**2*Derivative(x1(t), t) + 2*k21*ke*x1(t)*Derivative(x1(t), t) + k21*x1(t)**2*Derivative(x1(t), t) + ke**2*Derivative(x1(t), (t, 2)) + 2*ke*x1(t)*Derivative(x1(t), (t, 2)) + x1(t)**2*Derivative(x1(t), (t, 2))

    One way to simplify it consists in integrating it. For this purpose,
    one divides it by its initial and then one and then one integrates it. .

        >>> IO_rel = IO_rel / IO_R.initial (IO_rel)
        >>> IO_rel
        (Ve*k21*ke*x1(t) + Ve*k21*x1(t)**2 + Ve*ke*Derivative(x1(t), t) + k12*ke**2*Derivative(x1(t), t) + 2*k12*ke*x1(t)*Derivative(x1(t), t) + k12*x1(t)**2*Derivative(x1(t), t) + k21*ke**2*Derivative(x1(t), t) + 2*k21*ke*x1(t)*Derivative(x1(t), t) + k21*x1(t)**2*Derivative(x1(t), t) + ke**2*Derivative(x1(t), (t, 2)) + 2*ke*x1(t)*Derivative(x1(t), (t, 2)) + x1(t)**2*Derivative(x1(t), (t, 2)))/(ke**2 + 2*ke*x1(t) + x1(t)**2)

    Then, one integrates it. Simpler, isn't it ?

        >>> L = IO_R.integrate (IO_rel, t)
        >>> L
        [Ve*k21*x1(t)/(ke + x1(t)), (-Ve*ke - k12*ke**2 + k12*x1(t)**2 - k21*ke**2 + k21*x1(t)**2)/(ke + x1(t)), x1(t)]

    The precise relationship between the input/output equation and the
    list L is given by the next formula:

        >>> zero = Add (*[IO_R.differentiate (L[i], t**i) for i in range (len (L))]) - IO_rel

    The variable ``zero`` contains 0 but python does not recognize it.
    Let us force the simplification:

        >>> IO_R.normal_form (zero)
        0
"""

# include "sage/ext/python.pxi"
# Used to be useful in sage 5.3 and former releases.
# include "sage/ext/python_ref.pxi"
import sys
import string
import re

from cpython cimport bool

def ascii (s) :
    if isinstance (s, bytes) :
        return s
    else :
        return str (s).encode ('ascii')

import sympy
from sympy import var, Symbol, Derivative, Eq, Integer, Rational, oo, Add, Mul, Pow, degree

def function (s) :
    L = re.split (r'[, ]+', s)
    if len(L) == 1 :
        return sympy.Function(L[0])
    else :
        return tuple ([sympy.Function(z) for z in L])

def indexedbase (s) :
    L = re.split (r'[, ]+', s)
    if len(L) == 1 :
        return sympy.IndexedBase(L[0])
    else :
        return tuple ([sympy.IndexedBase(z) for z in L])

def listof_symbols (expr, subrankings) : 
    if expr is None :
        return []
    elif isinstance (expr, sympy.core.symbol.Symbol) :
        return [ expr ]
    elif isinstance (expr, sympy.core.function.UndefinedFunction) :
        return [ expr ]
    elif isinstance (expr, sympy.IndexedBase) :
        return [ expr ]
    elif isinstance (expr, sympy.tensor.indexed.Indexed) :
        if expr.base in subrankings :
            L = [ listof_symbols (e, subrankings) for e in expr.indices ]
            return [ e for f in L for e in f ]
        else :
            return [expr.base, expr]
    elif hasattr (expr, '__iter__') :
        L = [ listof_symbols (e, subrankings) for e in expr ]
        return [ e for f in L for e in f ]
    else :
        return [expr]

# bmi_c.pxd is automatically built from bmi_balsa.h
# bmi_strings.pyx is automatically built from bmi_indices-h.in

cimport bmi_c
include "bmi_strings.pyx"
include "DifferentialRing.pyx"
include "RegularDifferentialChain.pyx"
include "BaseFieldExtension.pyx"

