# DifferentialRing
cdef class DifferentialRing:
    r"""
    The class DifferentialRing implements differential polynomial rings
    
        Differential rings can be endowed with zero, one or many different
        derivations. 

        Each derivation is denoted using a symbol and is associated to 
        an independent variable (for each derivation x, it is possible 
        to differentiate w.r.t. x, and to manipulate differential polynomials
        depending on x).

        Differential rings are built over an alphabet of dependent variables
        (called differential indeterminates by Ritt and Kolchin), which
        are presented in a list of blocks. 

        Differential polynomials are plain polynomials, built over the
        infinite alphabet of the derivatives of the dependent variables,
        often called derivatives in this package, plus the finite set
        of the independent variables.

        While building differential rings, it is possible to customize
        the dependencies of the dependent variables w.r.t. the independent
        variables. By default, each dependent variable depends on all
        the independent variables, following the order of the list of
        derivations. It is possible to customize the order of the
        independent variables. It is also possible to specify the list
        of the independent variables the dependent variables depend on.
        This feature is especially useful to define parameters, i.e. 
        dependent variables which do not depend on any independent variables.

        The list of blocks, together with the list of derivations, define 
        a so-called ranking. Observe that rankings apply to parameters also
        and that parameters do not need to lie at the bottom of the rankings.

        A ranking is any total ordering over the set of the derivatives (in 
        this package, rankings are extended to the independent variables), 
        which satisfies the two axioms of rankings:
        
        -- Every derivative u is less than any of its proper derivatives.
        
        -- If u and v are derivatives such that u is less than v and x is
           any independent variable, then the derivative of u with respect 
           to x is less than the derivative of v with respect to x.
        
        In this package, rankings are defined by the list x[1] > ... > x[p] 
        of the independent variables plus a list b[1] >> ... >> b[n] of blocks. 
        Each block b is defined by a list u[1] > ... > u[m] of dependent 
        variables. Any dependent variable must appear in exactly one block.
        
        The >> operator between blocks indicates a block elimination
        ranking: if b[i] >> b[j] are two blocks, v[i] is any derivative of 
        any dependent variable occuring in b[i], and, v[j] is any derivative 
        of any dependent variable occuring in b[j], then v[i] > v[j].
        
        Within a given block b = u[1] > ... > u[m], in the ordinary 
        differential case (only one derivation), the derivatives are ordered 
        by the unique orderly ranking such that u[1] > ... > u[m]:
        the derivatives of higher order are ranked above the derivatives 
        if lower order ; two derivatives having the same order are
        ranked using the ordering defined by the list.
        
    NOTES:

        Rankings for partial differential equations are not yet fully 
        supported, due to limitations of Sage functions. 

    EXAMPLES:
    
    The examples focus on the concept of rankings and parameters.
    See the help page of DifferentialAlgebra for relevant examples.

        >>> from sympy import *
        >>> from DifferentialAlgebra import *
        >>> x,a,b = var ('x,a,b')
        >>> y,z = function ('y,z')

    Mathematically, R is a differential ring, with a single derivation.
    Among the four dependent variables listed in the block list, two
    of them do not depend on any independent variable. They thus should
    be viewed as parameters, in the usual sense. Following Ritt and Kolchin
    notation, R = Q[x,a,b]{z,y} with derivation d/dx.

        >>> R = DifferentialRing (derivations = [x], blocks = [z,y,a,b], parameters = [a,b])
        >>> R
        differential_ring

    Here is a differential polynomial which belongs to R

        >>> poly = a*Derivative(y(x),x) + b + x*y(x) + Derivative(z(x),x)**2 + z(x)

    The ranking is a pure elimination ranking:

        >>> R.sort (R.indets (poly, selection = 'all'), 'descending')  
        [Derivative(z(x), x), z(x), Derivative(y(x), x), y(x), a, b, x]

    Mathematically, S and R define the same ring. The ranking of S is
    different from that of R. Observe that parameters do not lie at the
    bottom of the ranking.

        >>> S = DifferentialRing (derivations = [x], blocks = [b,a,y,z], parameters = [a,b])
        >>> S
        differential_ring
        >>> S.sort (R.indets (poly, selection = 'all'), 'descending')
        [b, a, Derivative(y(x), x), y(x), Derivative(z(x), x), z(x), x]

    Mathematically, T, S and R define the same ring. The derivatives of
    z and y are ranked orderly.

        >>> T = DifferentialRing (derivations = [x], blocks = [[z,y],a,b], parameters = [a,b])
        >>> T
        differential_ring
        >>> T.sort (R.indets (poly, selection = 'all'), 'descending')
        [Derivative(z(x), x), Derivative(y(x), x), z(x), y(x), a, b, x]

    Last, here is an example of a PDE ring. The ranking is orderly. 
    The dependent variable v does not depend on x.

        >>> x,y = var('x,y')
        >>> u,v = function ('u,v')
        >>> R = DifferentialRing (derivations = [x,y], blocks = [[u,v]], parameters = [v(y)])
        >>> poly = u(x,y) + v(y)
        >>> R.sort (R.indets (poly, selection = 'all'), 'descending')
        [u(x, y), v(y)]
    """

    cdef bmi_c.ALGEB dring
#    Pointer to the BALSA internal data structure

    cdef bytes notation
#   The default notation

    cdef public object evdict
#   A dictionary used by eval

# __process_notation
    def __process_notation (self, notation) :
        cdef strnot
        if not isinstance (notation, str) :
            strnot = self.notation
        else :
            strnot = ascii (notation)
            if strnot != BMI_IX_Derivative and \
               not strnot.startswith (BMI_IX_jet) :
                strnot = self.notation
        return strnot

# __init__
    def __init__ (
                self, 
                list derivations = [], 
                list blocks = [], 
                list parameters = [], 
                notation = BMI_IX_undefined):
        r"""
        Constructor of DifferentialRing

        INPUT:

            ``derivations`` -- a list of variables
            ``blocks``      -- a list of blocks, each block being either
                               a dependent variable or, a list of 
                               dependent variables
            ``parameters``  -- a list of dependent variables, with
                               special dependencies
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            Create a new differential ring with derivations (independent
            variables) ``derivations``, differential indeterminates (dependent 
            variables) ``blocks``, parameters ``parameters`` and default
            notation ``notation``. 

        RANKINGS:

            The ranking is mostly defined by the order of the blocks 
            in ``blocks``. The leftmost blocks are ranked higher than
            the rightmost ones. Dependent variables which belong to a
            same block are ranked orderly.

            Parameters are dependent variables which do not need to
            depend on all independent variables, or have special dependencies. 
            This definition generalizes parameters, in the usual sense: 
            dependent variables which do not depend on any independent 
            variables. Parameters do not need to lie at the bottom of 
            the ranking.

            See the help page of DifferentialRing for more details on rankings.

        NOTATIONS (EXAMPLES):

            >>> from sympy import *
            >>> from DifferentialAlgebra import *

            >>> x,y = var ('x,y')
            >>> f = function ('f')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [f])
            >>> poly = Derivative (f(x,y), x, y)**2 + f(x,y) + x + 3

        A few notations are available: ``Derivative``, ``jet`` and ``jet(arg)``
        All notations (with jet(arg) = jet('_')) are accepted on input. 
        The ``notation`` argument is applied for the output only.

        The ``Derivative`` notation is the default one.
        Differential indeterminates must be declared using ``function``

            >>> R.normal_form (poly)
            x + f(x, y) + Derivative(f(x, y), x, y)**2 + 3

        The ``jet`` notation
        Differential indeterminates must be declared using ``indexedbase``
        Here is the same derivative as above, in the jet notation

            >>> f = indexedbase ('f')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [f], notation = 'jet')
            >>> R.normal_form (poly)
            x + f[x, y]**2 + f + 3

        The ``jet(arg)`` notation
        It is a variant of the jet notation where the order zero derivatives
        are indexed using arg 

            >>> R = DifferentialRing (derivations = [x,y], blocks = [f], notation = 'jet(_)')
            >>> R.normal_form (poly)
            x + f[_] + f[x, y]**2 + 3

        RANKINGS AND PARAMETERS (EXAMPLES):

            >>> x,a,b = var ('x,a,b')
            >>> y,z = function ('y,z')

        Mathematically, R is a differential ring, with a single derivation.
        Among the four dependent variables listed in the block list, two
        of them do not depend on any independent variable. They thus should
        be viewed as parameters, in the usual sense. Following Ritt and Kolchin
        notation, R = Q[x,a,b]{z,y} with derivation d/dx.

            >>> R = DifferentialRing (derivations = [x], blocks = [z,y,a,b], parameters = [a,b])
            >>> R
            differential_ring

        Here is a differential polynomial which belongs to R

            >>> poly = a*Derivative(y(x),x) + b + x*y(x) + Derivative(z(x),x)**2 + z(x)

        The ranking is a pure elimination ranking:

            >>> R.sort (R.indets (poly, selection = 'all'), 'descending')  
            [Derivative(z(x), x), z(x), Derivative(y(x), x), y(x), a, b, x]

        Mathematically, S and R define the same ring. The ranking of S is
        different from that of R. Observe that parameters do not lie at the
        bottom of the ranking.

            >>> S = DifferentialRing (derivations = [x], blocks = [b,a,y,z], parameters = [a,b])
            >>> S
            differential_ring
            >>> S.sort (R.indets (poly, selection = 'all'), 'descending')
            [b, a, Derivative(y(x), x), y(x), Derivative(z(x), x), z(x), x]

        Mathematically, T, S and R define the same ring. The derivatives of
        z and y are ranked orderly.

            >>> T = DifferentialRing (derivations = [x], blocks = [[z,y],a,b], parameters = [a,b])
            >>> T
            differential_ring
            >>> T.sort (R.indets (poly, selection = 'all'), 'descending')
            [Derivative(z(x), x), Derivative(y(x), x), z(x), y(x), a, b, x]

        Last, here is an example of a PDE ring. The ranking is orderly. 
        The dependent variable v does not depend on x.

            >>> x,y = var('x,y')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[u,v]], parameters = [v(y)])
            >>> poly = u(x,y) + v(y)
            >>> R.sort (R.indets (poly, selection = 'all'), 'descending')
            [u(x, y), v(y)]
        """
#
        cdef bmi_c.ALGEB A
        cdef bytes mesgerr, strders, strblks, strpars
        cdef list L
        strders = ascii (derivations)
        strblks = ascii (blocks)
        strpars = ascii (parameters)
# The default notation
        self.notation = BMI_IX_Derivative
        self.notation = self.__process_notation (notation)
# First build the BMI differential ring and exit on error
        A = bmi_c.bmi_dapyx_differential_ring (
                strders, strblks, strpars, self.notation)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
# The dring field with its reference counter set to 1
        self.dring = A
        bmi_c.bmi_balsa_increment_nbref (self.dring)
# The dictionary to be used by eval
        L = listof_symbols ([derivations, blocks], \
                indexedbase ('lex,grlexA,grlexB,degrevlexA,degrevlexB'))
        self.evdict = { str(e):e for e in L }

    def __dealloc__ (self):
#   The destructor
        bmi_c.bmi_balsa_decrement_nbref (self.dring)
        bmi_c.bmi_balsa_clear_ALGEB (self.dring)

    def __repr__ (self):
#   The external representation
        return 'differential_ring'

    def _latex (self):
#   The LaTeX representation
        return 'differential\\_ring'

# coeffs
    def coeffs (
            self,
            equation,
            variable = None,
            BaseFieldExtension basefield = None,
            notation = BMI_IX_undefined):
        r"""
        The coefficients and monomials of a rational differential fraction

        INPUT:

            ``equation``  -- a rational fraction
            ``variable``  -- (default: None) a derivative
            ``basefield`` -- (default: None) a base field
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            A sequence of two lists: the list C` of the coefficients and 
            the list M of the monomials. Both lists have the same length 
            and the following relation holds:

            ``equation`` == add (C[i]*M[i] for i in range (len (C)))

            The monomials M[i] are possibly rational fractions.

            If ``variable`` and ``basefield`` are both None, then
            the coefficients C[i] are numerical.

            If ``variable`` is not None, then ``basefield`` must be None,
            the C[i] only depend on variables strictly less than
            ``variable``, and the M[i] only depend on variables greater
            than or equal to ``variable``, w.r.t. the ranking.

            If ``basefield`` is not None, then ``variable`` must be None,
            the C[i] only depend on variables which belong to the field,
            and the M[i] only depend on variables which do not belong
            to the field.

            Restriction: the denominator of ``equation`` must depend on 
            variables which are all, either strictly less than, or greater 
            than ``variable``.
            
        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> f,u = function ('f,u')

        The differential ring is equipped with an elimination ranking such h
        that u >> f.

            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,f], parameters = [u(x)])

        Let us define a differential field F

            >>> eqns = [Eq (Derivative(f(x,y),x), f(x,y)**2 + 1), Eq(Derivative(f(x,y),y), 2*y*(f(x,y)**2 + 1))]
            >>> C = RegularDifferentialChain (eqns, R, pretend = False)
            >>> F = BaseFieldExtension (relations = C)

        Let us start with a differential polynomial

            >>> eqn = x*Derivative(u(x),x)**2-(f(x,y)**2+1)*u(x)

        The numerical coefficients and the corresponding monomials

            >>> R.coeffs (eqn)
            ([1, -1, -1], [x*Derivative(u(x), x)**2, f(x, y)**2*u(x), u(x)])

        The coefficients of eqn w.r.t. Derivative(f(x,y),x). The variables x
        and f(x,y), both less than Derivative(f(x,y),x), w.r.t. the ranking,
        arise in the coefficients.

            >>> R.coeffs (eqn, Derivative(f(x,y),x))
            ([x, -f(x, y)**2 - 1], [Derivative(u(x), x)**2, u(x)])

        The coefficients of eqn, viewed as a differential polynomials
        with coefficients in F.

            >>> R.coeffs (eqn, basefield = F)
            ([x, -f(x, y)**2 - 1], [Derivative(u(x), x)**2, u(x)])

        The coefficients of eqn, viewed as a differential polynomials
        with coefficients in Q(x,y).

            >>> R.coeffs (eqn, basefield = BaseFieldExtension ())
            ([x, -1, -1], [Derivative(u(x), x)**2, f(x, y)**2*u(x), u(x)])

        The coeffs function applies to rational fractions too.

            >>> eqn = x*Derivative(u(x),x)**2-u(x)/Derivative(f(x,y),x)

        The numerical coefficients and the corresponding monomials,
        in the extended sense.

            >>> R.coeffs (eqn)
            ([1, -1], [x*Derivative(u(x), x)**2, u(x)/Derivative(f(x, y), x)])

        The coefficients of eqn, viewed as a differential polynomials
        with coefficients in F.

            >>> R.coeffs (eqn, basefield = F)
            ([x, -1/Derivative(f(x, y), x)], [Derivative(u(x), x)**2, u(x)])

        The coefficients of eqn, viewed as a differential polynomials
        with coefficients in Q(x,y).

            >>> R.coeffs (eqn, basefield = BaseFieldExtension ())
            ([x, -1], [Derivative(u(x), x)**2, u(x)/Derivative(f(x, y), x)])

        """
#
        cdef result
        cdef bytes streqns, strvar, strgens, strrels, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if variable != None and basefield != None:
            raise RuntimeError ('the variable and the base field must not be both specified')
        streqns = ascii (equation)
        if variable == None:
            if basefield == None:
                strvar = ascii (0)
                strgens = ascii ("")
                strrels = ascii ("")
            else:
                strvar = ascii ("")
                strgens = ascii (basefield.generators ())
                strrels = ascii (basefield._BaseFieldExtension__relations ())
        else:
            strvar = ascii (variable)
            if basefield == None:
                strgens = ascii ("")
                strrels = ascii ("")
            else:
                raise RuntimeError ('the variable and the base field must not be both specified')
        A = bmi_c.bmi_dapyx_coeffs (
                    streqns, strvar, strgens, strrels, self.dring, 
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# equations
    def equations (
            self,
            equations,
            bool solved = False,
            selection = None,
            notation = BMI_IX_undefined):
        r"""
        A selection in a list of polynomials

        INPUT:

            ``equations`` -- a polynomial or a list of polynomials 
            ``solved``    -- (default: False) a boolean
            ``selection`` -- (default: None) a relational expression
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            The same list, but processed according to the other parameters.

            If ``solved`` is True, the differential polynomials are
            displayed as equations, solved with respect to their
            leading rank.

            The optional parameter ``selection`` is a relational
            expression combining one of the keywords 'order', 'leader',
            'rank' on the one side, with a value on the other side,
            using one of the relational operators. Only the equations
            which satisfy the condition are selected.
            
            - If the keyword is 'order', then the value must be a 
              nonnegative integer.

            - If the keyword is 'rank', then the value must be a
              rank, including the special ranks 0 and 1.

            - If the keyword is 'leader', and the relational
              operator is '>=', '>', '<=' or '<', then the value
              must be a derivative.

            - If the keyword is 'leader', and the relational
              operator is '==' or '!=', then the value may also have
              the form: 'modifier' (derivative), where 'modifier' is
              one of the keywords 'derivative' or 'proper'.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var('x,y')
            >>> w,z,a = function ('w,z,a') 
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[w,a],z], parameters = [a(y)]) 
            >>> L = [0, 18, a(y)**2, Derivative(a(y),y), w(x,y) - z(x,y)]
            >>> leader,order,rank = var('leader,order,rank')
            >>> derivative,proper = function('derivative,proper')

        Just check that the elements of L are polynomials of R

            >>> R.equations (L) 
            [0, 18, a(y)**2, Derivative(a(y), y), w(x, y) - z(x, y)]

        The only polynomial with rank 0 is 0

            >>> R.equations (L, selection = Eq(rank,0))
            [0]

        The polynomials with rank 1 are the nonzero numbers

            >>> R.equations (L, selection = Eq(rank,1))
            [18]

        The polynomials which are not numbers, solved with respect
        to their leaders

            >>> R.equations (L, solved = True, selection = rank > 1)
            [Eq(a(y)**2, 0), Eq(Derivative(a(y), y), 0), Eq(w(x, y), z(x, y))]

        The polynomials whose leaders are derivatives of a(y)

            >>> R.equations (L, selection = Eq(leader,derivative (a(y))))
            [a(y)**2, Derivative(a(y), y)]

        The polynomials whose leaders are derivatives of w(x,y), solved
        with respect to their leaders

            >>> R.equations (L, solved = True, selection = Eq(leader,derivative (w(x,y))))
            [Eq(w(x, y), z(x, y))]

        Solving a single equation w.r.t. its leading derivative

            >>> R.equations (w(x,y) - a(y), solved=True)
            Eq(w(x, y), a(y))

        """
#
        cdef result
        cdef bytes streqns, strsel, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        if selection == None:
            strsel = ascii ('rank >= 0')
        elif isinstance (selection, sympy.core.relational.Equality):
            strsel = ascii (str (selection.lhs) + str (selection.rel_op) + str(selection.rhs))
        else:
            strsel = ascii (selection)
        A = bmi_c.bmi_dapyx_equations_with_criterion_DR (
                    streqns, self.dring, int (False), int (solved), strsel,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# __process_evdict
    def __process_evdict (
            self,
            evdict0) :
        r"""
        Normalize the notation of keys in the evaluate dictionary

        INPUT:

            ``evdict0`` -- a dictionary as used by evaluate

        OUTPUT:

            The normalized dictionary. The only transform is this:
            in the case of the Derivative notation, UndefinedFunction
            (say w) are transformed into functions (say w(x)).
            
        """
        evdict = dict ()
        for key in evdict0 :
            if self.notation == BMI_IX_Derivative:
                if isinstance (key, sympy.core.function.UndefinedFunction) :
                    A = bmi_c.bmi_dapyx_leading_derivative (
                            ascii ([key]), self.dring, int (False),
                            BMI_IX_jet, self.notation, 0, 0)
                    if bmi_c.bmi_dapyx_is_error (A): 
                        mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
                        bmi_c.bmi_balsa_clear_ALGEB (A)
                        raise RuntimeError (str (mesgerr, 'utf-8'))
                    result = eval (A.value, globals (), self.evdict)
                    bmi_c.bmi_balsa_clear_ALGEB (A)
                    newkey = result[0]
                    evdict [newkey] = evdict0 [key]
                else :
                    evdict [key] = evdict0 [key]
            else :
                theta, y = self.factor_derivative (key)
                if theta == 1 :
                    evdict [y] = evdict0 [key]
                else :
                    evdict [key] = evdict0 [key]
        return evdict

# __evaluate
    def __evaluate (
            self, 
            P, 
            evdict) :
        r"""
        Subfunction of evaluate
        """
        cdef newargs, theta, y, result, ders, x, d
        if isinstance (P, Add) or isinstance (P, Mul) or isinstance (P, Pow):
            newargs = [ self.evaluate (arg, evdict) for arg in P.args ]
            return P.func (*newargs)
        elif isinstance (P, Integer) or isinstance (P, Rational) :
            return P
        elif isinstance (P, list) :
            return [ self.evaluate (arg, evdict) for arg in P ]
        elif isinstance (P, set) :
            return { self.evaluate (arg, evdict) for arg in P }
        else :
            try :
                result = evdict[P]
            except :
                try :
                    theta, y = self.factor_derivative (P)
                    result = evdict[y]
                    ders = self.indets (selection = 'derivations')
                    if self.notation in [BMI_IX_jet, BMI_IX_jet0] and \
                            self.is_differential_fraction (result) :
                        result = self.differentiate (result, theta)
                    else :
                        for x in ders :
                            d = degree (theta, x)
                            if d > 0 :
                                result = Derivative (result, (x, d))
                except :
                    result = P
            return result

# evaluate
    def evaluate (
            self, 
            P, 
            evdict) :
        r"""
        Evaluates P at evdict

        INPUT:

            ``P``      -- a differential fraction or a list of
                            differential fractions
            ``evdict`` -- a dictionary 

        OUTPUT:

            The expression P evaluated at evdict

        Note:

            In the classical derivative notation, evaluate behaves
            almost as subs

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x = var('x')
            >>> y = indexedbase('y')
            >>> R = DifferentialRing(derivations = [x], blocks = [y], notation = 'jet')

        An order two differential polynomial in jet notation

            >>> edo = y*y[x,x] + y[x]**2 - 6*y

        A formal power series

            >>> ybar = Add(*[y[i]*x**i for i in range (5)])
            >>> ybar
            x**4*y[4] + x**3*y[3] + x**2*y[2] + x*y[1] + y[0]

        The differential polynomial evaluated at the series

            >>> serie = R.evaluate (edo, {y:ybar})
            >>> serie
            -6*x**4*y[4] - 6*x**3*y[3] - 6*x**2*y[2] - 6*x*y[1] + (x**4*y[4] + x**3*y[3] + x**2*y[2] + x*y[1] + y[0])*Derivative(x**4*y[4] + x**3*y[3] + x**2*y[2] + x*y[1] + y[0], (x, 2)) + Derivative(x**4*y[4] + x**3*y[3] + x**2*y[2] + x*y[1] + y[0], x)**2 - 6*y[0]

        The series in collected form

            >>> collect (expand (serie.doit ()), x)
            28*x**6*y[4]**2 + 42*x**5*y[3]*y[4] + x**4*(30*y[2]*y[4] + 15*y[3]**2 - 6*y[4]) + x**3*(20*y[1]*y[4] + 20*y[2]*y[3] - 6*y[3]) + x**2*(12*y[0]*y[4] + 12*y[1]*y[3] + 6*y[2]**2 - 6*y[2]) + x*(6*y[0]*y[3] + 6*y[1]*y[2] - 6*y[1]) + 2*y[0]*y[2] - 6*y[0] + y[1]**2

        Observe c is not a differential inderminate

            >>> c = var('c')
            >>> ybar = Rational(1/4)*(c + 2*x)**2

        The series ybar is a zero of the differential polynomial

            >>> serie = R.evaluate (edo, {y:ybar})
            >>> serie
            (c + 2*x)**2*Derivative((c + 2*x)**2/4, (x, 2))/4 - 3*(c + 2*x)**2/2 + Derivative((c + 2*x)**2/4, x)**2
            >>> collect (expand (serie.doit ()), x)
            0
        """ 
        cdef evdict1
        evdict1 = self.__process_evdict (evdict)
        return self.__evaluate (P, evdict1)

# factor_derivative
    def factor_derivative (
            self,
            derivative,
            notation = BMI_IX_undefined):
        r"""
        Splits a derivative into a derivation operator and a dependent variable

        INPUT:

            ``derivative`` -- a derivative
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            A pair (``theta``, ``symb``) such that ``derivative`` is
            the derivative of ``symb`` w.r.t. ``theta``. The derivation
            operator ``theta`` is expressed as a monomial on the 
            independent variables.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> a,x,y = var ('a,x,y')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v,a], parameters = [a,v(y)])

        A basic example

            >>> R.factor_derivative (Derivative(u(x,y),x))
            (x, u(x, y))

        A parameter is considered as a dependent variable which does not
        depend on any independent variable. The monomial ``theta`` is 1, ,
        meaning that a is not differentiated.

            >>> R.factor_derivative (a)
            (1, a)

        The following examples illustrates the relationship between
        factor_derivative and differentiate.

            >>> derv = Derivative(u(x,y),x,x,y)
            >>> derv
            Derivative(u(x, y), (x, 2), y)
            >>> theta, symb = R.factor_derivative (derv)
            >>> theta
            x**2*y
            >>> symb
            u(x, y)
            >>> R.differentiate (symb, theta)
            Derivative(u(x, y), (x, 2), y)
        """        
        cdef result
        cdef bytes strder, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        strder = ascii (derivative)
        A = bmi_c.bmi_dapyx_factor_derivative (
                    strder, self.dring,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# indets
    def indets (
                self, 
                equations = None,
                selection = BMI_IX_derivs,
                derivation = None,
                notation = BMI_IX_undefined):
        r"""
        The indeterminates of a DifferentialRing or a rational fraction
        
        INPUT:

            ``equations``  -- (default: None) a rational fraction
                              or a list of rational fractions
            ``selection``  -- (default: 'derivatives') a string
            ``derivation`` -- (default: None) a variable
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            A list of variable and function, selected according to 
            ``selection``. See the examples below.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y,z,t = var ('x,y,z,t')
            >>> u,v,w = function ('u,v,w')
            >>> R = DifferentialRing (derivations = [x,y,t], blocks = [u,[v,w],z], parameters = [u(t,y,x),z,w(y)])

        With no selection, or with 'derivatives' or 'dependent', 
        return the list of the dependent variables (or differential 
        indeterminates) of R
        
            >>> R.indets ()
            [u(t, y, x), v(x, y, t), w(y), z]
            >>> R.indets (selection = 'dependent')
            [u(t, y, x), v(x, y, t), w(y), z]
            >>> R.indets (selection = 'derivatives')
            [u(t, y, x), v(x, y, t), w(y), z]
        
        With 'derivations' or 'independent', return the list of 
        the derivations (or independent variables) of R
        
            >>> from DifferentialAlgebra import DifferentialRing
            >>> R.indets (selection = 'derivations')
            [x, y, t]
            >>> R.indets (selection = 'independent')
            [x, y, t]
        
        With 'parameters', return the list of the parameters of R
        
            >>> R.indets (selection = 'parameters')
            [u(t, y, x), w(y), z]
        
        With 'all', return the list of the independent and the dependent 
        variables of R
        
            >>> R.indets (selection = 'all')
            [x, y, t, u(t, y, x), v(x, y, t), w(y), z]
        
        With 'constants', possibly followed by a derivation, return 
        the list of the dependent variables the derivatives (w.r.t.
        the derivation, if any) of which, are zero.
        
            >>> R.indets (selection = 'constants')
            [z]
            >>> R.indets (selection = 'constants', derivation = y)
            [z]
            >>> R.indets (selection = 'constants', derivation = t)
            [w(y), z]

        The indets of a rational fraction

            >>> eq = u(t,y,x) + 1/w(y)
            >>> R.indets (eq)
            [u(t, y, x), w(y)]
            >>> R.indets (eq, selection = 'constants', derivation = x)
            [w(y)]

        """
#
        cdef list result
        cdef bytes streqns, strder, strsel, mesgerr
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        if equations == None:
            streqns = ascii ("")
        elif isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        if derivation == None:
            strder = ascii (0)
        else:
            strder = ascii (derivation)
        strsel = ascii (selection)
        L = bmi_c.bmi_dapyx_indets (
                streqns, self.dring, strsel, strder, 
                self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# is_differential_fraction
    def is_differential_fraction (
            self,
            equations,
            notation = BMI_IX_undefined):
        r"""
        Return True if an expression is a rational differential fraction

        INPUT:

            ``equations`` -- an expression or a list of expressions
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            True or False

        NOTES:

            The function actually tests if its argument can be parsed
            by the BLAD parser

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x = var('x')
            >>> w = function('w')
            >>> R = DifferentialRing (derivations = [x], blocks = [w])
            >>> F = w(x)*Derivative(w(x),x) + x + 1
            >>> R.is_differential_fraction([F, F**(-3)])
            True
            >>> R.is_differential_fraction(F**Rational(3,4))
            False
        """
#
        cdef result
        cdef bytes streqns, strders
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        strders = ascii ([])
        A = bmi_c.bmi_dapyx_differentiate (
                    self.dring, streqns, int (False), strders,
                    self.notation, strnot, 0, 0)
        result = not bmi_c.bmi_dapyx_is_error (A)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# is_constant
    def is_constant (
            self,
            ratfrac,
            derivation = None,
            notation = BMI_IX_undefined):
        r"""
        Test if a rational fraction is constant

        INPUT:

            ``ratfrac``    -- a rational fraction or a list of rational
                              fractions
            ``derivation`` -- (default: None) an independent variable

        OUTPUT:

            True, if the rational fraction is a constant w.r.t. 
            ``derivation``, else False. If ``derivation`` is omitted, 
            True if the rational fraction is a constant w.r.t.
            all the derivations of the ring.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> a,x,y = var ('a,x,y')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v,a], parameters = [a,v(y)])

            >>> R.is_constant (3)
            True
            >>> R.is_constant ([1/a,v(y),Derivative(u(x,y),x)/(a+1)])   
            [True, False, False]
            >>> R.is_constant ([1/a,v(y),Derivative(u(x,y),x)/(a+1)], x) 
            [True, True, False]
        """
        cdef result
        cdef bytes streqns, strder, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (ratfrac, list):
            streqns = ascii (ratfrac)
        else:
            streqns = ascii ([ratfrac])
        if derivation == None:
            strder = ascii (0)
        else:
            strder = ascii (derivation)
        A = bmi_c.bmi_dapyx_is_constant (
                    streqns, strder, self.dring,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (ratfrac, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# leading_derivative
    def leading_derivative (
            self,
            equations,
            notation = BMI_IX_undefined):
        r"""
        The leading derivative of a rational fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of the leading derivatives of the equations.

            The leading derivative of a rational fraction F is 
            defined as the highest derivative u such that the 
            partial derivative of F w.r.t. u is nonzero. 

            This definition is extended to rational fractions 
            which only depend on independent variables. 

            The leading derivative of constant rational fractions
            is not defined.
 
        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> L = [t, u(t)**2 - u(t), Derivative(v(t),t), 1/v(t)]
            >>> R.leading_derivative (L)
            [t, u(t), Derivative(v(t), t), v(t)]
            >>> R.leading_derivative (1/u(t)**2)
            u(t)
            >>> R.leading_derivative (u(t)**2)
            u(t)
        """
#
        cdef result
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        A = bmi_c.bmi_dapyx_leading_derivative (
                    streqns, self.dring, int (False),
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# leading_rank
    def leading_rank (
            self,
            equations,
            bool listform = False,
            notation = BMI_IX_undefined):
        r"""
        The leading rank of a rational fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``listform``  -- (default: False) a bool
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of the leading ranks of the equations.

            If the optional parameter ``listform`` is True, then
            the ranks are returned as lists rather than monomials.
            This is feature is relevant when the rank is a derivative
            raised at the 0th power.

            The leading rank of a rational fraction F = P/Q is defined 
            as the leading derivative u of F, raised at the power
            deg (P, u) - deg (Q, u).

            Nonzero constant rational fractions have leading rank 1.

            The zero rational fraction has leading rank 0.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t') 
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> L = [t, u(t)**2 - u(t), Derivative(v(t),t), 1/v(t)]

        The list of leading ranks of the elements of L, in monomial form m

            >>> R.leading_rank (L)
            [t, u(t)**2, Derivative(v(t), t), 1/v(t)]

        The same list, in list form

            >>> R.leading_rank (L, listform = True)
            [[t, 1], [u(t), 2], [Derivative(v(t), t), 1], [v(t), -1]]

        The case of a single rational fraction

            >>> R.leading_rank (1/u(t)**2)
            u(t)**(-2)
            >>> R.leading_rank (1/u(t)**2, listform = True)
            [u(t), -2]
            >>> R.leading_rank (u(t)**2)
            u(t)**2
            >>> R.leading_rank (u(t)**2, listform = True)
            [u(t), 2]

        The case of a rank, being a derivative raised at the 0th power

            >>> R.leading_rank ((u(t)-1)/(u(t)+2))
            1
            >>> R.leading_rank ((u(t)-1)/(u(t)+2), listform = True)
            [u(t), 0]

        """
#
        cdef result
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        A = bmi_c.bmi_dapyx_leading_rank (
                    streqns, self.dring, int (False), int (listform),
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# initial
    def initial (
            self,
            equations,
            notation = BMI_IX_undefined):
        r"""
        The initial of a rational fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            The list of the initials of the equations, i.e. the 
            leading coefficients of the equations w.r.t. their
            leading derivatives.
        
            The initial of a rational fraction F = P/Q, with 
            leading derivative u, is itself a rational fraction:
            the leading coefficient of P w.r.t. u, divided by
            the leading coefficient of Q w.r.t. u, recalling
            that the leading coefficient w.r.t. u, of a polynomial
            R which does not depend on u, is defined as R itself.

        NOTES:

            The above definition is not yet fully implemented.

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> L = [t, v(t)*u(t)**2 - u(t) + Derivative(v(t),t) + 1, t*u(t)/Derivative(v(t),t)]
            >>> R.initial (L)
            [1, v(t), t/Derivative(v(t), t)]
            >>> R.initial (u(t)**2 + 1/v(t))
            1
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        strvar = ascii (0)
        A = bmi_c.bmi_dapyx_leading_coefficient (
                    streqns, self.dring, int (False), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# leading_coefficient
    def leading_coefficient (
            self,
            equations,
            variable = None,
            notation = BMI_IX_undefined):
        r"""
        The leading coefficient of a rational fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``variable``  -- (default: None) a variable
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
 
            The list of the leading coefficients of the equations,
            w.r.t. ``variable``. If ``variable`` is None, the
            list of the initials of the equations is returned.

            The leading coefficient of a rational fraction F = P/Q,
            w.r.t. some derivative u, is a rational fraction:
            it is the leading coefficient of P w.r.t. u, divided
            by the leading coefficient of Q w.r.t. u,  recalling
            that the leading coefficient w.r.t. u, of a polynomial
            R which does not depend on u, is defined as R itself.

        NOTES:

            The above definition is not yet fully implemented.

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> L = [t, v(t)*u(t)**2 - u(t) + Derivative(v(t),t) + 1, t*u(t)/Derivative(v(t),t)]

        The initials of the elements of L

            >>> R.leading_coefficient (L)
            [1, v(t), t/Derivative(v(t), t)]

        The initial of a single rational fraction

            >>> R.leading_coefficient (u(t)**2 + Derivative(v(t),t)/v(t))
            1

        The leading coefficient of the elements of L, w.r.t. t.

            >>> R.leading_coefficient (L, t)
            [1, u(t)**2*v(t) - u(t) + Derivative(v(t), t) + 1, u(t)/Derivative(v(t), t)]

        The leading coefficient of a single rational fraction,
        w.r.t. a derivative.

            >>> R.leading_coefficient (u(t)**2 + Derivative(v(t),t)/v(t), Derivative(v(t),t))
            1/v(t)
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        if variable == None:
            strvar = ascii (0)
        else:
            strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_leading_coefficient (
                    streqns, self.dring, int (False), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# tail
    def tail (
            self,
            equations,
            variable = None,
            notation = BMI_IX_undefined):
        r"""
        The tail of a rational fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``variable``  -- (default: None) a variable
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of tails of the equations, w.r.t. ``variable``.
            If ``variable`` is omitted, the list of tails w.r.t. the
            leading derivatives of the equations is returned.

            The tail of a rational fraction F is defined as
            the rational fraction F - i(F) * rg(F) where i(F)
            and rg(F) denote the initial and the leading rank of F.

            More generally, the tail, w.r.t. some derivative u, of 
            a rational fraction F = P/Q is defined as the rational
            fraction F - lc(F,u) * rg(F,u), where lc(F,u) is defined 
            as the leading coefficient of F w.r.t. u and rg(F,u) is 
            the monomial obtained by raising u at the degree 
            deg (P,u) - deg (Q,u)

        NOTES:

            The above definition is not yet fully implemented.

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> L = [t + 3, v(t)*u(t)**2 - u(t) + Derivative(v(t),t) + 1, t*u(t)/Derivative(v(t),t)]

        The list of tails of the elements of L, w.r.t. their leading
        derivatives

            >>> R.tail (L)
            [3, -u(t) + Derivative(v(t), t) + 1, 0]

        The tail of a single rational fraction, w.r.t. its leading
        derivative

            >>> eq = u(t)**2 + Derivative(v(t),t)/v(t)
            >>> R.tail (eq)
            Derivative(v(t), t)/v(t)

        Check the definition of tails

            >>> R.initial (eq) * R.leading_rank (eq) + R.tail (eq) - eq
            0

        The list of tails of the elements of L, w.r.t. t

            >>> R.tail (L, t)
            [3, 0, 0]

        The tail of a single rational fraction, w.r.t. a derivative

            >>> eq = u(t)**2 + Derivative(v(t),t)/v(t)
            >>> derv = Derivative(v(t), t)
            >>> R.tail (eq, derv)
            u(t)**2

        Check the definition of tails w.r.t. some derivative

            >>> R.leading_coefficient (eq, derv) * derv + R.tail (eq, derv) - eq
            0
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        if variable == None:
            strvar = ascii (0)
        else:
            strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_tail (
                    streqns, self.dring, int (False), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# separant
    def separant (
            self,
            equations,
            variable = None,
            notation = BMI_IX_undefined):
        r"""
        The separant of a rational fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``variable``  -- (default: None) a variable
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
 
            The list of separants of the equations, w.r.t. ``variable``.
            If ``variable`` is omitted, the list of separants of the
            equations, w.r.t. their leading derivatives, is returned.

            The separant of a rational fraction F w.r.t. some derivative
            $u$, is defined as the partial derivative of F w.r.t. u.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> L = [t + 3, v(t)*u(t)**2 - u(t) + Derivative(v(t),t) + 1, t*u(t)/Derivative(v(t),t)]

        The list of separants of the elements of L, w.r.t. their
        leading derivatives.

            >>> R.separant (L)
            [1, 2*u(t)*v(t) - 1, t/Derivative(v(t), t)]

        The separant of a single rational fraction, w.r.t. its 
        leading derivative.

            >>> R.separant (u(t)**2 + Derivative(v(t),t)/v(t))
            2*u(t)

        The list of separants of the elements of L, w.r.t. t

            >>> R.separant (L, t)
            [1, 0, u(t)/Derivative(v(t), t)]

        The separant of a single rational fraction, w.r.t. a derivative

            >>> R.separant (u(t)**2 + Derivative(v(t),t)/v(t), Derivative(v(t),t))
            1/v(t)
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        if variable == None:
            strvar = ascii (0)
        else:
            strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_separant (
                    streqns, self.dring, int (False), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# integrate
    def integrate (
            self,
            equations,
            variable,
            bool iterated = True,
            notation = BMI_IX_undefined):
        r"""
        Integrate a rational differential fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``variable``  -- a variable
            ``iterated``  -- (default: True) a bool
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            A list of rational fractions, representing a decomposition
            of the equation or, a list of lists of rational fractions
            if ``equations`` is a list.

            The parameter ``variable`` must be an independent variable.

            Assume ``equations`` is a single rational fraction.
            Denote R the differential ring, F the rational
            fraction, t the independent variable and L the 
            returned list. Then 

            F == add (R.differentiate (L[i], t**i) for i in range (len (L)))

            Moreover, if F is the derivative of some other
            rational fraction, w.r.t. t, then L[0] is zero.

            The returned list is ranking dependent. Orderly rankings
            usually give more interesting results.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [[u,v]])

        Observe that the ranking is orderly.

            >>> eq = Derivative(u(t),t)**2 + Derivative(v(t)**2,t,t) - Derivative(u(t)*v(t) + t*v(t), t)
            >>> L = R.integrate (eq.doit(), t)
            >>> L
            [Derivative(u(t), t)**2, -t*v(t) - u(t)*v(t), v(t)**2]

        The two expressions are equal, for their difference is 0

            >>> eq - Add (*[R.differentiate (L[i], t**i) for i in range (len (L))])
            t*Derivative(v(t), t) + u(t)*Derivative(v(t), t) + v(t)*Derivative(u(t), t) - 2*v(t)*Derivative(v(t), (t, 2)) + v(t) - Derivative(t*v(t) + u(t)*v(t), t) - 2*Derivative(v(t), t)**2 + Derivative(v(t)**2, (t, 2))

        The case of a rational fraction which is the derivative of
        some other fraction

            >>> eq = (v(t)**2 + t)/(3*Derivative(v(t),t))              
            >>> L = R.integrate (Derivative(eq, t).doit(), t)
            >>> L
            [0, (t + v(t)**2)/(3*Derivative(v(t), t))]
            >>> L[1] - eq
            0
        """
#
        cdef list result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_integrate (
                    self.dring, streqns, strvar, int (iterated),
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# differentiate
    def differentiate (
            self,
            equations,
            *args,
            notation = BMI_IX_undefined):
        r"""
        Differentiate a rational differential fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``args``      -- a possibly empty sequence of monomials 
                             involving independent variables only
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of derivatives of the equations, w.r.t. the
            product of the monomials given in ``args``. The monomials
            must depend on independent variables only.

        NOTES:

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])
            >>> eq = u(t)**2

        Do not differentiate at all

            >>> R.differentiate (eq)
            u(t)**2
            >>> R.differentiate (eq, 1)
            u(t)**2

        Similar behaviour as 'diff'

            >>> R.differentiate (eq, t)
            2*u(t)*Derivative(u(t), t)
            >>> R.differentiate (eq, t, t)
            2*u(t)*Derivative(u(t), (t, 2)) + 2*Derivative(u(t), t)**2

        Another way to differentiate twice w.r.t. t

            >>> R.differentiate (eq, t**2) 
            2*u(t)*Derivative(u(t), (t, 2)) + 2*Derivative(u(t), t)**2

        Differentiate three times w.r.t. t

            >>> R.differentiate (eq, t**2, t)
            2*u(t)*Derivative(u(t), (t, 3)) + 6*Derivative(u(t), t)*Derivative(u(t), (t, 2))

        """
#
        cdef result
        cdef bytes streqns, strders, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        strders = ascii ([arg for arg in args])
        A = bmi_c.bmi_dapyx_differentiate (
                    self.dring, streqns, int (False), strders,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# differential_prem
    def differential_prem (
             self,
             polynomial,
             list redset,
             mode = BMI_IX_fully,
             notation = BMI_IX_undefined):
        r"""
        Ritt's reduction algorithm by a set of polynomials

        INPUT:

            ``polynomial`` -- a polynomial
            ``redset``     -- a list of polynomials with integer coefficients
            ``mode``       -- (default: 'full') a string
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            A pair (h, r) such that h*p = r modulo the differential
            ideal generated by ``redset`` (denoting p for ``polynomial``).

            The optional argument ``mode`` = 'full', 'partial' or 'algebraic'.
            If 'full', the remainder r is fully reduced w.r.t. ``redset``
            If 'partial', the remainder r is partially reduced w.r.t. 
            ``redset``. If 'algebraic', the remainder is algebraically 
            reduced w.r.t. ``redset``.

            In all cases, h is a product of powers of the initials and
            separants involved in the reduction process.

        NOTES:
           
            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]])
            >>> eq1 = Derivative(u(x,y),x)**2 - 4*u(x,y)
            >>> eq2 = Derivative(u(x,y),x,y)*Derivative(v(x,y),y) - u(x,y) + 1
            >>> eq3 = Derivative(v(x,y),x,x) - Derivative(u(x,y),x)
            >>> redset = [eq1, eq2, eq3]

        Here is the reduction set

            >>> R.equations (redset, solved = True)
            [Eq(Derivative(u(x, y), x)**2, 4*u(x, y)), Eq(Derivative(u(x, y), x, y), (u(x, y) - 1)/Derivative(v(x, y), y)), Eq(Derivative(v(x, y), (x, 2)), Derivative(u(x, y), x))]
            >>> ideal = R.RosenfeldGroebner (redset)
            >>> ideal
            [regular_differential_chain]

        Here is a description of the radical differential ideal 
        generated by the reduction set

            >>> C = ideal [0]

        The first polynomial is fully reduced w.r.t. redset

            >>> poly = (1/Integer(7))*u(x,y)
            >>> h, r = R.differential_prem (poly, redset)
            >>> h, r
            (1, u(x, y)/7)
            >>> C.normal_form (h * poly - r)
            0

        The next polynomial is not fully reduced w.r.t redset

            >>> poly = (1/Integer(7))*Derivative(v(x,y),x,y) + Derivative(u(x,y),x,x)
            >>> h, r = R.differential_prem (poly, redset)
            >>> h, r
            (2*Derivative(u(x, y), x), 2*(Derivative(v(x, y), x, y) + 14)*Derivative(u(x, y), x)/7)
            >>> C.normal_form (h * poly - r)
            0

        It is not partially reduced either.

            >>> h, r = R.differential_prem (poly, redset, mode = 'partial')
            >>> h, r
            (2*Derivative(u(x, y), x), 2*(Derivative(v(x, y), x, y) + 14)*Derivative(u(x, y), x)/7)
            >>> C.normal_form (h * poly - r)
            0

        However, it is algebraically reduced w.r.t redset

            >>> h, r = R.differential_prem (poly, redset, mode = 'algebraic')
            >>> h, r
            (1, Derivative(u(x, y), (x, 2)) + Derivative(v(x, y), x, y)/7)
            >>> C.normal_form (h * poly - r)
            0
        """
#
        cdef result
        cdef bytes streqns, strredset, mesgerr, strmode
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii (polynomial)
        strredset = ascii (redset)
        strmode = ascii (mode)
        L = bmi_c.bmi_dapyx_differential_prem2 (
            streqns, strredset, strmode, self.dring, 
            self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# normal_form
    def normal_form (
            self,
            equations,
            notation = BMI_IX_undefined):
        r"""
        The normal form of a rational differential fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of 
                             rational fractions 
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            The list of normal forms of the equations.

        NOTES:

            Mathematically, this function does not do anything.
            Were there more than one notation, it could perform
            changes of notations. At least, it is useful to 
            recognize rational fractions of differential polynomials
            equal to 0.

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var ('t')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,v])

        A complicated rational fraction of differential polynomials

            >>> eq = Derivative(u(t) + (v(t)**2 + t)/(3*u(t)+Derivative(v(t),t)), t)
            >>> L = R.integrate (eq.doit(), t)

        This should be zero

            >>> zero = (eq - Derivative(L[1], t)).doit()
            >>> zero                      
            (t + v(t)**2)*(-3*Derivative(u(t), t) - Derivative(v(t), (t, 2)))/(3*u(t) + Derivative(v(t), t))**2 + (2*v(t)*Derivative(v(t), t) + 1)/(3*u(t) + Derivative(v(t), t)) + Derivative(u(t), t) - (6*u(t)*Derivative(u(t), t) + u(t)*Derivative(v(t), (t, 2)) + 2*v(t)*Derivative(v(t), t) + Derivative(u(t), t)*Derivative(v(t), t) + 1)/(3*u(t) + Derivative(v(t), t)) - (-3*Derivative(u(t), t) - Derivative(v(t), (t, 2)))*(t + 3*u(t)**2 + u(t)*Derivative(v(t), t) + v(t)**2)/(3*u(t) + Derivative(v(t), t))**2

            >>> R.normal_form (zero)      
            0
        """
#
        cdef result
        cdef arg
        cdef bytes streqns, strders, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        strders = ascii ([1])
        A = bmi_c.bmi_dapyx_differentiate (
                    self.dring, streqns, int (False), strders,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# positive_integer_roots
    def positive_integer_roots (
            self,
            ratfrac,
            variable,
            notation = BMI_IX_undefined):
        r"""
        The nonnegative integer roots of a rational fraction

        INPUT:

            ``ratfrac``  -- a fraction or a list of fractions
            ``variable`` -- the variable for which roots are sought
            ``notation`` -- (default: 'undefined') a string

        OUTPUT:

            The list of the nonnegative integer roots of the rational fraction

            The roots of a fraction are the roots of its denominator.
            The list is sorted by increasing order.

        NOTE:

            This method is used for analyzing the existence and uniqueness
            of formal power series solutions of systems of polynomial ODE.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,q = var ('x,q')
            >>> y = indexedbase ('y')
            >>> R = DifferentialRing (derivations = [x], blocks = [y,q], parameters = [q], notation = 'jet')

            >>> F = (y[x]*y[x,x]*(q-3)*(q-13)**2 + 5*y[x]**2*(q-3)*(q-13)*(q**2-2))/y[x,x]

        u[x,x] - 2 is a zero modulo the differential ideal defined by C

            >>> R.positive_integer_roots (F, q)
            [3, 13]

        """
        cdef result
        cdef bytes mesgerr, str_notation, str_ratfrac, str_variable
        cdef bmi_c.ALGEB_string L
        str_notation = self.__process_notation (notation)
        str_ratfrac = ascii (ratfrac)
        str_variable = ascii (variable)
        L = bmi_c.bmi_dapyx_positive_integer_roots (
            str_ratfrac, str_variable, self.dring, self.notation,
            str_notation, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# sort
    def sort (
            self,
            list polynomials,
            mode = BMI_IX_ascending,
            notation = BMI_IX_undefined):
        r"""
        Sort a list of polynomials according to their leading monomials

        INPUT:

            ``polynomials`` -- a list of polynomials
            ``mode``        -- (default: 'increasing') a string
            ``notation``    -- (default: 'undefined') a string

        OUTPUT:

            The list polynomials, sorted by increasing leading monomial,
            if ``mode`` is 'increasing', by decreasing leading monomial,
            if ``mode`` is 'decreasing'.

            The leading monomial of a nonzero numeric polynomial is 1.
            The leading monomial of a nonzero, non numeric polynomial,
            is the product of the leading rank of the polynomial, by the
            leading monomial of its initial. Leading monomials are
            compared lexicographically, w.r.t. the ranking. 
            This ordering is extended to 0, which is considered lower 
            than any other polynomial.

        NOTES:

            In the future, this function will be extended to rational 
            fractions.

            So far, there is a single notation available, because of
            limitations of function.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> a,x,y = var ('a,x,y')
            >>> u,v = function ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v,a], parameters = [a,v(y)])
            >>> L = [Derivative(v(y),y,y), Derivative(v(y),y,y)**3, 0, a*u(x,y)**2, x]

        The list L is sorted by increasing leading monomial

            >>> R.sort (L)
            [0, x, Derivative(v(y), (y, 2)), Derivative(v(y), (y, 2))**3, a*u(x, y)**2]
            >>> R.sort (L, 'ascending')
            [0, x, Derivative(v(y), (y, 2)), Derivative(v(y), (y, 2))**3, a*u(x, y)**2]

        By decreasing leading monomial

            >>> R.sort (L, 'descending')
            [a*u(x, y)**2, Derivative(v(y), (y, 2))**3, Derivative(v(y), (y, 2)), x, 0]

        Another example

            >>> L = [ u(x,y), u(x,y)*v(y), u(x,y)*v(y)**2, u(x,y)*v(y)*a ]
            >>> R.sort (L, 'ascending')
            [u(x, y), u(x, y)*v(y), a*u(x, y)*v(y), u(x, y)*v(y)**2]
            >>> R.sort (L, 'descending')
            [u(x, y)*v(y)**2, a*u(x, y)*v(y), u(x, y)*v(y), u(x, y)]
        """
#
        cdef list result
        cdef bytes streqns, mesgerr, strmode
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii (polynomials)
        strmode = ascii (mode)
        L = bmi_c.bmi_dapyx_sort (
                streqns, strmode, self.dring, self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result
 
# RosenfeldGroebner
    def RosenfeldGroebner (
            self,
            equations, 
            BaseFieldExtension basefield = None,
            list attributes = [
                    'differential', 'autoreduced', 'primitive', 
                    'squarefree', 'normalized', 'coherent'], 
            singsol = BMI_IX_all, 
            dimlb = BMI_IX_safecase,
            notation = BMI_IX_undefined,
            int timeout = 0, 
            int memout = 0):
        r"""
        The Rosenfeld-Groebner simplifier
        
        INPUT:
        
            ``equations``  -- a list of differential polynomials, differential 
                              polynomial equations or inequations
            ``basefield``  -- (default: None) a BaseFieldExtension
            ``attributes`` -- (default: ['differential', 'autoreduced', 
                              'primitive', 'squarefree', 'normalized', 
                              'coherent']) a list of strings
            ``singsol``    -- (default: 'all') a string
            ``dimlb``      -- (default: 'safecase') a string
            ``notation``   -- (default: 'undefined') a string
            ``timeout``    -- (default: 0) a nonnegative integer
            ``memout``     -- (default: 0) a nonnegative integer
        
        OUTPUT:
        
            A representation of the radical of the differential ideal generated 
            by ``equations`` (assuming, for simplicity, that ``equations`` 
            only involves differential polynomials), as an intersection of 
            radical differential ideals, presented by regular differential 
            chains, with respect to the ranking of the DifferentialRing.  
    
            The list ``equations`` may involve differential polynomials,
            differential rational fractions, and, even relational expressions
            between differential rational fractions, built using the '==' and 
            '!=' operators. In the presence of inequations (denominators of
            rational fractions, or relational expression built using '!='),
            the defined ideal is the radical of the differential ideal 
            generated by the equations of ``equations``, saturated by 
            the multiplicative family generated by the inequations of 
            ``equations``.
    
            If ``basefield`` is not specified, then any case implying a
            relations between independent variables is discarded. If 
            ``basefield`` is specified, then, moreover, any case implying 
            the vanishing of some nonzero element of ``basefield`` is 
            discarded also. 
    
            The returned list may be empty, meaning the ideal is the unit one 
            and that ``equations`` has no solution.
                                      
            If ``equations`` is empty then the output involves a single 
            regular differential chain, which describes the zero ideal.
     
            The optional parameter ``attributes`` permit to specify the 
            attributes of the regular differential chains to be computed.
    
            The optional parameter ``singsol`` = 'all', 'essential', 'none'. 
            It controls the splittings performed by RosenfeldGroebner. 
    
            - 'all' is the default value. 
        
            - 'essential' ensures that the returned decomposition is 
              irredundant. It only applies in the case of a single 
              equation, without any parameter.
        
            - 'none' makes RosenfeldGroebner return at most one regular 
              differential chain. This chain is the first one that would have 
              been computed without this option. In many cases, this 
              first chain could be considered as the general component, 
              though this notion is not always mathematically well-defined.
    
            The optional parameter ``dimlb`` = 'nocase', 'safecase', 'odecase', 
            'pdecase' controls the splittings performed by RosenfeldGroebner 
            by discarding any chain whose dimension (differential dimension 
            in the differential case) is lower than the number of input 
            equations.
    
            - 'nocase' disables this option.
    
            - 'safecase' is the default value. The option is only activated 
              in the non-differential case and in the case of a single input 
              equation. These two cases are theoretically proven.
    
            - 'odecase' applies also the option to general ODE systems. The
              option implements a conjecture in this case (see Ritt's book, 
              Questions for investigation, 10).
    
            - 'pdecase' applies also the option to general PDE systems. The 
              option implements a conjecture in this case. It is still
              experimental.
    
            If not defined using ``notation``, the default notation of the
            returned regular differential chains is the one of the
            DifferentialRing.
     
            ``timeout`` and ``memout`` permit to limit in seconds and
            megabytes, the resources allowed for the computation.
    
        NOTES:

            The first version of the RosenfeldGroebner algorithm
            appeared in 1994. It benefited from many improvements,
            coming from many different people. The current version 
            is close to the one described in [Boulier06]. It does
            not rely on any Gröbner basis computation.

            The RosenfeldGroebner method contains also an implementation
            of the Low Power Theorem, which is a major result of Ritt
            and Kolchin. See also [Hubert99]. The implemented version
            applies over differential base fields defined by generators
            and relations.

        REFERENCES:

            [Boulier06] François Boulier. Réécriture algébrique
            dans les systèmes d'équations différentielles polynomiales
            en vue d'application dans les Sciences du Vivant. Mémoire
            d'Habilitation à Diriger les Recherches. Université Lille I.
            2006. http://tel.archives-ouvertes.fr/tel-00137153

            [Hubert99] Évelyne Hubert. Essential Components of an 
            Algebraic Differential Equation. Journal of Symbolic
            Computation 28(4-5), pp. 657-680. 1999.

        EXAMPLES:
    
        The first example features one nonlinear ODE. According to
        the Low Power Theorem, its solution $y(x) = 0$ is a singular
        solution. 
        
            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x = var('x')
            >>> y = function ('y')
            >>> R = DifferentialRing (derivations = [x], blocks = [y])
            >>> L = R.RosenfeldGroebner ([Derivative(y(x),x)**2-4*y(x)], singsol = 'essential')
            >>> L
            [regular_differential_chain, regular_differential_chain]
            >>> L[0].equations ()
            [-4*y(x) + Derivative(y(x), x)**2]
            >>> L[1].equations ()
            [y(x)]
        
        The second example features a system on nonlinear PDE. This
        academic example illustrates what a system of differential
        polynomials may look like, in general. Some integrability
        conditions are computed.
        
            >>> from DifferentialAlgebra import DifferentialRing
            >>> x,y = var('x,y')
            >>> u,v = function('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]])
            >>> eq1 = Derivative(u(x,y),x)**2 - 4*u(x,y)
            >>> eq2 = Derivative(Derivative(u(x,y),x),y) * Derivative(v(x,y),y) - u(x,y) + 1
            >>> eq3 = Derivative(Derivative(v(x,y),x),x) - Derivative(u(x,y),x)
            >>> L = R.RosenfeldGroebner ([eq1, eq2, eq3])
            >>> L
            [regular_differential_chain]
            >>> L[0].equations (solved=True)
            [Eq(Derivative(u(x, y), y)**2, 2*u(x, y)), Eq(Derivative(u(x, y), x)**2, 4*u(x, y)), Eq(Derivative(v(x, y), y), (u(x, y)*Derivative(u(x, y), x)*Derivative(u(x, y), y) - Derivative(u(x, y), x)*Derivative(u(x, y), y))/(4*u(x, y))), Eq(Derivative(v(x, y), (x, 2)), Derivative(u(x, y), x))]

        """
#
        cdef list result, eqns, ineq
        cdef BaseFieldExtension F
        cdef RegularDifferentialChain C
        cdef bmi_c.ALGEB_list L
        cdef bytes mesgerr, streqns, strineq, strgens, strrels, strattr
        cdef bytes strsingsol, strdimlb
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            eqns, ineq = self.__process_equations (equations, strnot)
        else:
            eqns, ineq = self.__process_equations ([equations], strnot)
        streqns = ascii (eqns)
        strineq = ascii (ineq)
        strattr = ascii (attributes)
        strattr = strattr.replace (b"'", b"")
        strsingsol = ascii (singsol)
        strdimlb = ascii (dimlb)
        if basefield == None:
            F = BaseFieldExtension ()
        else:
            F = basefield
        strgens = ascii (F.generators ())
        strrels = ascii (F._BaseFieldExtension__relations ())
        L = bmi_c.bmi_dapyx_RosenfeldGroebner (
                    streqns, strineq, strgens, strrels,
                    strattr, self.dring, strsingsol, strdimlb, 1, 
                    self.notation, strnot, timeout, memout)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = []
        for i in range (1, L.value.size) :
            C = RegularDifferentialChain.__new__ (RegularDifferentialChain)
            C.dring = self
            C.notation = strnot
            C.evdict = self.evdict
            C.regchain = L.value.tab [i]
            bmi_c.bmi_balsa_increment_nbref (C.regchain)
            result.append (C)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# __process_equations
    def __process_equations (self, list L, strnot):
        r"""
        Split a list into equations and inequations

        INPUT:
    
            ``L`` -- a list
    
        OUTPUT:
    
            A pair of two lists: a first list of polynomials that should 
            be considered as equations and a second one that should be 
            considered as inequations.
    
            The elements of ``L`` can be rational differential fractions. Their 
            denominators are then considered as inequations. The elements of
            ``L`` may also be relational expressions, built using '==' and '!='.

        NOTES:

            This function is used by RosenfeldGroebner, mostly.

        EXAMPLES:
    
        """
#
        cdef bmi_c.ALGEB_listof_string P
        cdef bytes L_as_string, mesgerr
        cdef list eqns, ineq
        L_as_string = ascii (L)
        P = bmi_c.bmi_dapyx_process_equations (
                    L_as_string, self.dring,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (P):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (P)
            bmi_c.bmi_balsa_clear_ALGEB (P)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        eqns = eval (P.value.tab [1].value, globals (), self.evdict)
        ineq = eval (P.value.tab [2].value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (P)
        return (eqns, ineq)

# __ranking
    def __ranking (self):
        r"""
        The ranking of a DifferentialRing

        INPUT:

            Nothing

        OUTPUT:

            A string describing the ranking of the differential ring.

        NOTES:

            This function is used by the change_ranking method of
            the RegularDifferentialChain class.
        """
#
        cdef bytes result
        cdef bytes mesgerr
        cdef bmi_c.ALGEB_string L
        L = bmi_c.bmi_dapyx_ranking (self.dring)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = L.value
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

