# RegularDifferentialChain
cdef class RegularDifferentialChain:
    r"""
        The class RegularDifferentialChain implements differential regular chains.
        
            A regular differential chain is a set of polynomial differential
            equations in some simplified form. Regular differential chains
            belong to (polynomial) differential rings. They are usually
            produced, from raw systems of polynomial differential equations,
            by the RosenfeldGroebner algorithm.
    
            Regular differential chains are slight generalizations of Ritt
            characteristic sets. A self-contained definition of regular 
            differential chains and a normal form algorithm are provided 
            in [BL10]. An equivalence theorem is provided in [BLPM19].
    
        REFERENCES:
    
            [BL10] François Boulier and François Lemaire. A Normal 
            Form Algorithm for Regular Differential Chains. Mathematics
            in Computer Science 4(2), pp. 185-201. 2010.
            https://doi.org/10.1007/s11786-010-0060-3
    
            [BLPM19] François Boulier and François Lemaire and Adrien
            Poteaux and Marc Moreno Maza. An Equivalence Theorem for 
            Regular Differential Chains. Journal of Symbolic Computation
            vol 93, pp. 34-5. 2019
            https://doi.org/10.1016/j.jsc.2018.04.011
            hal.archives-ouvertes.fr/hal-01391768
    
        EXAMPLES:
        
        Create a DifferentialRing, then a RegularDifferentialChain:
        
            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,a,b = var ('x,a,b')
    
        In the Derivative notation:
    
            >>> y = function ('y')
            >>> R = DifferentialRing (derivations = [x], blocks = [y,a,b], parameters = [a,b])
            >>> C = RegularDifferentialChain ([Derivative(y(x),x)**2 - a*y(x)], R)
            >>> C
            regular_differential_chain
    
        In the jet notation:
    
            >>> y = indexedbase ('y')
            >>> R = DifferentialRing (derivations = [x], blocks = [y,a,b], parameters = [a,b], notation = 'jet')
            >>> C = RegularDifferentialChain ([y[x]**2 - a*y], R)
            >>> C
            regular_differential_chain
    """

    cdef DifferentialRing dring
#   The differential ring the chain belongs to

    cdef notation
#   The default notation

    cdef public object evdict
#   A dictionary used by eval

    cdef bmi_c.ALGEB regchain
#    Pointer to the BALSA internal data structure

# __process_notation
    def __process_notation (self, notation) :
        cdef strnot
        if not isinstance (notation, str) :
            strnot = self.notation
        else :
            strnot = ascii (notation)
            if strnot != BMI_IX_Derivative and strnot != BMI_IX_jet:
                strnot = self.notation
        return strnot

# __init__
    def __init__ (
            self, 
            list equations,
            DifferentialRing DRing,
            list attributes = [
                'differential', 'autoreduced', 'primitive', 
                'squarefree', 'normalized', 'coherent'],
            bool pretend = True,
            notation = BMI_IX_undefined, 
            int timeout = 0,
            int memout = 0):
        r"""
        Constructor of RegularDifferentialChain
    
        INPUT:
    
            ``equations``  -- a list of differential polynomial or
                              differential polynomial equations
            ``DRing``      -- a DifferentialRing
            ``attributes`` -- (default: ['differential', 'autoreduced', 
                              'primitive', 'squarefree', 'normalized', 
                              'coherent']) a list of strings
            ``pretend``    -- (default: True) a bool,
            ``notation``   -- (default: 'undefined') a string
            ``timeout``    -- (default: 0) a nonnegative integer
            ``memout``     -- (default: 0) a nonnegative integer
    
        OUTPUT:
    
            Create a regular differential chain of the differential
            polynomial ring ``DRing``, from ``equations``. The attributes
            of the chain depend on ``attributes``. By default, almost no test 
            is performed, to check that the equations do constitute a chain, 
            unless ``pretend`` is set to False.
    
            If not defined using ``notation``, the default notation is the
            one of ``DRing``.
    
            ``timeout`` and ``memout`` permit to limit in seconds and
            megabytes, the resources allowed for the computation.

        EXAMPLES:

        Create a regular differential chain with three PDE. 

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v], notation = 'jet')
            >>> eq1 = Eq(u[x],v)
            >>> eq2 = u[y] 
            >>> eq3 = v[y]
            >>> C = RegularDifferentialChain ([eq1,eq2,eq3], R)
            >>> C.equations (solved=True)
            [Eq(v[y], 0), Eq(u[y], 0), Eq(u[x], v)]
        """
#
        cdef list eqns, ineq
        cdef bmi_c.ALGEB A
        cdef bytes streqns, strattr, mesgerr
        cdef strnot
        self.notation = DRing.notation
        self.notation = self.__process_notation (notation)
        strnot = self.notation
        eqns, ineq = DRing._DifferentialRing__process_equations (equations, strnot)
        streqns = ascii (eqns)
        strattr = ascii (attributes)
        strattr = strattr.replace (b"'", b"")
        A = bmi_c.bmi_dapyx_pretend_regular_differential_chain (
                streqns, DRing.dring, strattr, int (pretend),
                self.notation, strnot, timeout, memout)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        self.dring = DRing
        self.notation = strnot
        self.evdict = DRing.evdict
        self.regchain = A
        bmi_c.bmi_balsa_increment_nbref (self.regchain)

    def __dealloc__ (self):
#   The destructor
        bmi_c.bmi_balsa_decrement_nbref (self.regchain)
        bmi_c.bmi_balsa_clear_ALGEB (self.regchain)

    def __repr__ (self):
#   The external representation
        return 'regular_differential_chain'

    def _latex_ (self):
#   The LaTeX representation
        return 'regular\\_differential\\_chain'

# differential_ring
    def differential_ring (self):
        r""" 
        The differential ring the chain belongs to

        INPUT:

            Nothing

        OUTPUT:

            The differential ring the chain belongs to

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v], notation = 'jet')
            >>> eq1 = Eq (u[x], v)
            >>> eq2 = u[y]
            >>> eq3 = v[y]
            >>> C = RegularDifferentialChain ([eq1,eq2,eq3], R)

        Here, one recovers the ring from the chain

            >>> S = C.differential_ring ()
            >>> S
            differential_ring
            >>> S.indets ()
            [u, v]
        """
#
        return self.dring

# equations
    def equations (
            self, 
            bool solved = False,
            selection = None,
            notation = BMI_IX_undefined):
        r"""
        The list of equations of a regular differential chain

        INPUT:

            ``solved``    -- (default: False) a boolean
            ``selection`` -- (default: None) a relational expression
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            A list of the differential polynomials which constitute the
            chain, sorted by increasing rank. 

            If ``solved`` is True, the differential polynomials are
            displayed as equations, solved with respect to their
            leading rank.

            The optional parameter ``selection`` is a relational
            expression combining one of the keywords 'order', 'leader',
            'rank' on the one side, with a value on the other side,
            using one of the relational operators. Only the equations
            which satisfy the condition are selected.
            
            - If the keyword is 'order', then the value must be a 
              nonnegative integer.

            - If the keyword is 'rank', then the value must be a
              rank, including the special ranks 0 and 1.

            - If the keyword is 'leader', and the relational
              operator is '>=', '>', '<=' or '<', then the value
              must be a derivative.

            - If the keyword is 'leader', and the relational
              operator is '==' or '!=', then the value may also have
              the form: 'modifier' (derivative), where 'modifier' is
              one of the keywords 'derivative' or 'proper'.

            If not defined using ``notation``, the notation is the
            chain default notation.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,a,b = var('x,a,b')
            >>> y = indexedbase ('y')
            >>> R = DifferentialRing (derivations = [x], blocks = [y,[a,b]], parameters = [a,b], notation = 'jet')
            >>> C = RegularDifferentialChain ([y[x]**2-a*y], R)
            >>> C.equations ()
            [-a*y + y[x]**2]
            >>> C.equations (solved = True)
            [Eq(y[x]**2, a*y)]

        The next example illustrates the use of ``selection``

            >>> from DifferentialAlgebra import DifferentialRing, RegularDifferentialChain
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v], notation = 'jet')
            >>> eq1 = Eq(u[x], v)
            >>> eq2 = Eq(u[y], 0)
            >>> eq3 = Eq(v[y], 0)
            >>> C = RegularDifferentialChain ([eq1,eq2,eq3], R)
            >>> C.equations (solved=True)
            [Eq(v[y], 0), Eq(u[y], 0), Eq(u[x], v)]
            >>> leader,rank,order = var ('leader,rank,order')
            >>> derivative,proper = function ('derivative,proper')

        The equations of order greater than 0. Observe that the ranking
        is not necessarily orderly.

            >>> C.equations (selection = order > 0)
            [v[y], u[y], -v + u[x]]

        The equation whose leader is u[x]
        
            >>> C.equations (selection = Eq(leader,u[x]))
            [-v + u[x]]

        The equations, whose leaders are derivatives of u[x]

            >>> C.equations (selection = Eq(leader,derivative (u[x])))
            [-v + u[x]]
        """
#
        cdef list result
        cdef bytes strsel, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if selection == None:
            A = bmi_c.bmi_dapyx_equations (
                    self.regchain, int (False), int (solved),
                    self.notation, strnot, 0, 0)
        else:
            if isinstance (selection, sympy.core.relational.Equality):
                strsel = ascii (str (selection.lhs) + str (selection.rel_op) + str(selection.rhs))
            else:
                strsel = ascii (selection)
            A = bmi_c.bmi_dapyx_equations_with_criterion_RDC (
                    self.regchain, int (False), int (solved), strsel,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# attributes
    def attributes (self):
        r""" 
        The list of attributes of a RegularDifferentialChain

        INPUT:

            Nothing

        OUTPUT:

            A list of strings, providing the attributes of
            a regular differential chain

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [u,v], notation = 'jet')
            >>> eq1 = Eq(u[x], v)
            >>> eq2 = u[y] 
            >>> eq3 = v[y]
            >>> C = RegularDifferentialChain ([eq1,eq2,eq3], R)
            >>> C.attributes ()
            ['differential', 'prime', 'autoreduced', 'primitive', 'squarefree', 'coherent', 'normalized']

        """
#
        cdef list result
        cdef bytes mesgerr
        cdef dict dico
        cdef bmi_c.ALGEB_string A
        A = bmi_c.bmi_dapyx_attributes (self.regchain)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        dico = dict (
                [
                        ('differential', 'differential'), 
                        ('prime','prime'), 
                        ('autoreduced','autoreduced'), 
                        ('squarefree','squarefree'), 
                        ('coherent','coherent'), 
                        ('primitive','primitive'), 
                        ('normalized','normalized')
                ])
        result = eval (A.value, dico)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# leading_derivative
    def leading_derivative (
            self,
            bool fullset = False,
            notation = BMI_IX_undefined):
        r"""
        The leading derivatives of a regular differential chain

        INPUT:

            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
 
            The list of the leading derivatives of the elements
            of the chain. See the documentation of differential
            rings.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  
            >>> C.leading_derivative()
            [u[y], u[x], v[y], v[x,x]]
        """
#
        cdef result
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        A = bmi_c.bmi_dapyx_leading_derivative (
                    streqns, self.regchain, int (fullset),
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# leading_rank
    def leading_rank (
            self,
            bool fullset = False,
            bool listform = False,
            notation = BMI_IX_undefined):
        r"""
        The leading ranks of a regular differential chain

        INPUT:

            ``listform``  -- (default: False) a bool
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            The list of the leading ranks of the elements
            of the chain. See the documentation of differential rings.

            If the optional parameter ``listform`` is True, then
            the ranks are returned as lists rather than monomials.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  
            >>> C.leading_rank()      
            [u[y]**2, u[x]**2, v[y], v[x,x]]
        """
#
        cdef result
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        A = bmi_c.bmi_dapyx_leading_rank (
                    streqns, self.regchain, int (fullset), int (listform),
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# initial
    def initial (
            self,
            bool fullset = False,
            notation = BMI_IX_undefined):
        r"""
        The initials of a regular differential chain

        INPUT:

            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of initials of the regular differential chain.
            See the documentation of differential rings.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  
            >>> C.initial ()
            [1, 1, 4*u, 1]
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        strvar = ascii (0)
        A = bmi_c.bmi_dapyx_leading_coefficient (
                    streqns, self.regchain, int (fullset), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# leading_coefficient
    def leading_coefficient (
            self,
            variable = None,
            bool fullset = False,
            notation = BMI_IX_undefined):
        r"""
        The leading coefficients of a regular differential chain

        INPUT:

            ``variable``  -- (default: None) a variable
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            The list of leading coefficients w.r.t. ``variable``
            of the regular differential chain. If ``variable``
            is omitted, the leading coefficients are taken w.r.t.
            the leading derivatives. See the documentation of
            differential rings.
        
        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  

        If ``variable`` is omitted, the initials are returned

            >>> C.leading_coefficient ()
            [1, 1, 4*u, 1]

        The leading coefficients w.r.t. u

            >>> C.leading_coefficient (u)
            [-2, -4, -u[x]*u[y] + 4*v[y], -u[x] + v[x,x]]
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        if variable == None:
            strvar = ascii (0)
        else:
            strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_leading_coefficient (
                    streqns, self.regchain, int (fullset), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# tail
    def tail (
            self,
            variable = None,
            bool fullset = False,
            notation = BMI_IX_undefined):
        r"""
        The tails of a regular differential chain

        INPUT:

            ``variable``  -- (default: None) a variable
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of the tails, w.r.t. ``variable``, of the
            regular differential chain. If ``variable`` is omitted,
            the tails are taken w.r.t. the leading derivatives.
            See the documentation of differential rings.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  

        If ``variable`` is omitted, the tails are taken w.r.t. the
        leading derivatives

            >>> C.tail ()
            [-2*u, -4*u, -u*u[x]*u[y] + u[x]*u[y], -u[x]]

        The tails w.r.t. u

            >>> C.tail (u)
            [u[y]**2, u[x]**2, u[x]*u[y], 0]
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        if variable == None:
            strvar = ascii (0)
        else:
            strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_tail (
                    streqns, self.regchain, int (fullset), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# separant
    def separant (
            self,
            variable = None,
            bool fullset = False,
            notation = BMI_IX_undefined):
        r"""
        The separants of a regular differential chain

        INPUT:

            ``variable``  -- (default: None) a variable
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:
        
            The list of the separants, w.r.t. ``variable``, of
            the regular differential chain. If ``variable`` is
            omitted, the separants are taken w.r.t. the leading
            derivatives. See the documentation of differential
            rings.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  

        If ``variable`` is omitted, the separants are taken w.r.t. the
        leading derivatives

            >>> C.separant ()
            [2*u[y], 2*u[x], 4*u, 1]

        The separants w.r.t. u

            >>> C.separant (u)
            [-2, -4, -u[x]*u[y] + 4*v[y], 0]
        """
#
        cdef result
        cdef bytes streqns, strvar, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        if variable == None:
            strvar = ascii (0)
        else:
            strvar = ascii (variable)
        A = bmi_c.bmi_dapyx_separant (
                    streqns, self.regchain, int (fullset), strvar,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# indets
    def indets (
                self, 
                selection = BMI_IX_derivs,
                bool fullset = False,
                notation = BMI_IX_undefined):
        r"""
        The indeterminates of a RegularDifferentialChain
        
        INPUT:

            ``selection``  -- (default: 'derivatives') a string
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            A list of variable and indexedbase, selected according to 
            ``selection``. See the examples below.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  

        With no selection, or with 'derivatives', return the list of 
        the derivatives occuring in the elements of C 

            >>> C.indets ()
            [u, u[x], v[y], v[x,x], u[y]]
            >>> C.indets (selection = 'derivatives')
            [u, u[x], v[y], v[x,x], u[y]]

        With 'dependent', return the list of the differential
        indeterminates, or dependent variables, occuring in the elements of C

            >>> C.indets (selection = 'dependent')  
            [u, v]

        With 'parameters', return the list of the parameters 
        occuring in the elements of C

            >>> C.indets (selection = 'parameters')
            []

        With 'all', return the list of the independent and the dependent 
        variables occuring in the elements of C

            >>> C.indets (selection = 'all')       
            [u, u[x], v[y], v[x,x], u[y]]
        """
#
        cdef list result
        cdef bytes streqns, strfull, mesgerr, strsel
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii ("")
        if fullset:
            strfull = ascii ("true")
        else:
            strfull = ascii ("false")
        strsel = ascii (selection)
        L = bmi_c.bmi_dapyx_indets (
                        streqns, self.regchain, strsel, strfull, 
                        self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# differential_prem
    def differential_prem (
             self,
             polynomial,
             mode = BMI_IX_fully,
             notation = BMI_IX_undefined):
        r"""
        Ritt's reduction algorithm by a regular differential chain

        INPUT:

            ``polynomial`` -- a polynomial
            ``mode``       -- (default: 'full') a string
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            A pair (h, r) such that h*p = r modulo the differential
            ideal generated by the equations of the regular differential
            chain (denoting p for ``polynomial``).

            The optional argument ``mode`` = 'full', 'partial' or 'algebraic'.
            If 'full', the remainder r is fully reduced w.r.t. the
            chain equations. If 'partial', the remainder r is partially
            reduced w.r.t. the chain equations. If 'algebraic', the
            remainder is algebraically reduced w.r.t. the chain equations.

            In all cases, h is a product of powers of the initials and
            separants involved in the reduction process.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]

        Here are the chain equations

            >>> C.equations (solved = True)
            [Eq(u[y]**2, 2*u), Eq(u[x]**2, 4*u), Eq(v[y], (u*u[x]*u[y] - u[x]*u[y])/(4*u)), Eq(v[x,x], u[x])]

        The first polynomial is already fully reduced w.r.t. the chain 
        equations.

            >>> poly = (Integer(1)/Integer(7))*u
            >>> h, r = C.differential_prem (poly)
            >>> h, r
            (1, u/7)
            >>> C.normal_form (h * poly - r)
            0

        The next polynomial is not reduced w.r.t. the chain equations

            >>> poly = (Integer(1)/Integer(7))*v[x,y] + u[x,x]
            >>> h, r = C.differential_prem (poly)
            >>> h, r
            (16*u**2*u[x]*u[y], 32*(u + 7*u[y])*u**2*u[x]/7)
            >>> C.normal_form (h * poly - r)
            0

        Let us try a partial reduction, instead of a full one.

            >>> h, r = C.differential_prem (poly, mode = 'partial')
            >>> h, r
            (16*u*u[x]*u[y], -4*(-u*u[x]**2 - 2*u*u[y]**2 - 56*u*u[y] - u[x]**2*u[y]**2 + u[x]**2 + 4*u[x]*u[y]*v[y] + 2*u[y]**2)*u[x]/7)
            >>> C.normal_form (h * poly - r)
            0

        The polynomial actually is algebraically reduced w.r.t the chain.

            >>> h, r = C.differential_prem (poly, mode = 'algebraic')
            >>> h, r
            (1, u[x,x] + v[x,y]/7)
            >>> C.normal_form (h * poly - r)
            0

        """
#
        cdef result
        cdef bytes streqns, mesgerr, strmode
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        streqns = ascii (polynomial)
        strmode = ascii (mode)
        L = bmi_c.bmi_dapyx_differential_prem (
            streqns, strmode, self.regchain, self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# prem
    def prem (
            self,
            ratfrac,
            notation = BMI_IX_undefined,
            int timeout = 0,
            int memout = 0):
        r"""
        The pseudo-remainder of a differential fraction by a regular differential chain

        INPUT:

            ``ratfrac``  -- a fraction or a list of fractions
            ``notation`` -- (default: 'undefined') a string

        OUTPUT:

            The pseudo-remainder of ratfrac modulo the regular differential 
            chain.

            If ratfrac is a polynomial, the result is a polynomial
            If it is a rational fraction, then the result is defined 
                as the ratio of the pseudo-remainder of the numerator over 
                the pseudo-remainder of the denominator of the fraction.

            The pseudo-remainder of a polynomial is zero if and only if
            the polynomial is zero modulo the ideal defined by the regular 
            differential chain.

        NOTE:

            See also differential_prem

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x = var ('x')
            >>> u,a = indexedbase ('u,a')
            >>> R = DifferentialRing (derivations = [x], blocks = [u,a], parameters = [a], notation = 'jet')
            >>> C = RegularDifferentialChain([u[x]**2-4*u], R, ['differential'])

        u[x,x] - 2 is a zero modulo the differential ideal defined by C

            >>> C.prem (u[x,x]-2)
            0
            
        u[x,x] is not

            >>> C.prem (u[x,x])
            4*u[x]
            
            >>> C.prem (1/u[x,x])
            1/(4*u[x])
            
            >>> C.prem ([u[x,x]-2, 1/u[x,x]])
            [0, 1/(4*u[x])]
            
        """
#
        cdef result
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (ratfrac, list) :
            streqns = ascii (ratfrac)
        else :
            streqns = ascii ([ratfrac])
        L = bmi_c.bmi_dapyx_prem (
            streqns, self.regchain, self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        if not isinstance (ratfrac, list) :
            result = result[0]
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# resultant
    def resultant (
            self,
            ratfrac,
            notation = BMI_IX_undefined,
            int timeout = 0,
            int memout = 0):
        r"""
        The resultant of a differential fraction modulo a regular differential chain

        INPUT:

            ``ratfrac``  -- a fraction or a list of fractions
            ``notation`` -- (default: 'undefined') a string

        OUTPUT:

            The resultant of ratfrac modulo the regular differential chain.

            If ratfrac is a polynomial, the result is a polynomial
            If it is a rational fraction, then the result is defined 
                as the ratio of the resultant of the numerator over the
                resultant of the denominator of the fraction.

            The resultant of a polynomial is zero if and only if
            the polynomial is a zerodivisor modulo the ideal defined
            by the regular differential chain [BLPM19].

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x = var ('x')
            >>> u,a = indexedbase ('u,a')
            >>> R = DifferentialRing (derivations = [x], blocks = [u,a], parameters = [a], notation = 'jet')
            >>> C = RegularDifferentialChain([(u[x]**2-4*u)*(u[x]**2-4*u**3)], R, ['differential'])

        u[x,x] - 2 is a zerodivisor modulo the differential ideal defined by C

            >>> C.resultant(u[x,x]-2)
            0

        u[x] is not

            >>> C.resultant(u[x])
            -16*u**4

            >>> C.resultant(1/u[x])
            -1/(16*u**4)

            >>> C.resultant([u[x,x] - 2, u[x]])
            [0, -16*u**4]
        """
#
        cdef result
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string L
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (ratfrac, list) :
            streqns = ascii (ratfrac)
        else :
            streqns = ascii ([ratfrac])
        L = bmi_c.bmi_dapyx_resultant (
            streqns, self.regchain, self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        if not isinstance (ratfrac, list) :
            result = result[0]
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# change_ranking
# pardi
    def change_ranking (
            self,
            DifferentialRing R = None,
            bool prime = False,
            notation = BMI_IX_undefined,
            int timeout = 0,
            int memout = 0):
        r"""
        Change of ranking for regular differential chains defining prime ideals

        INPUT:

            ``R``          -- (default: None) a differential ring
            ``prime``      -- (default: False) a bool
            ``notation``   -- (default: 'undefined') a string
            ``timeout``    -- (default: 0) a nonnegative integer
            ``memout``     -- (default: 0) a nonnegative integer

        OUTPUT:

            A regular differential chain defining the same ideal
            as ``self``, w.r.t. the ranking of ``R``.

            The differential ring ``R`` must define the same
            mathematical differential ring as the underlying 
            ring of ``self`` (only the ranking should be different). 

            The optional parameter ``prime`` permit to force the
            application of the algorithm over ``self``, even if
            ``self`` does not define a prime ideal.

            ``timeout`` and ``memout`` permit to limit in seconds and
            megabytes, the resources allowed for the computation.

        NOTES:

            The implemented algorithm is the PARDI algorithm. See
            references below.

        REFERENCES:

            [BLM10] François Boulier, François Lemaire and Marc Moreno
            Maza. Computing differential characteristic sets by change
            of ordering. Journal of Symbolic Comp. 45(1), pp. 124-149. 2010

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]

        The regular differential chain C defines a prime ideal.
        Its ranking is orederly.
        
            >>> C = ideal [0]  
            >>> C.equations (solved = True)
            [Eq(u[y]**2, 2*u), Eq(u[x]**2, 4*u), Eq(v[y], (u*u[x]*u[y] - u[x]*u[y])/(4*u)), Eq(v[x,x], u[x])]

        The ring Rbar is mathematically equivalent to R, but w.r.t.
        an elimination ranking.

            >>> Rbar = DifferentialRing (derivations = [x,y], blocks = [u,v], notation = 'jet')

        Perform the change of ranking, claiming that the ideal is prime.

            >>> Cbar = C.change_ranking (Rbar, prime=True)
            >>> Cbar.equations (solved = True)
            [Eq(v[y,y]**4, 2*v[y]**2 + 2*v[y,y]**2 - 1), Eq(v[x,y], (v[y,y]**3 - v[y,y])/v[y]), Eq(v[x,x], 2*v[y,y]), Eq(u, v[y,y]**2)]
        """
#
        cdef bmi_c.ALGEB L
        cdef bytes mesgerr, ranking
        cdef RegularDifferentialChain C
        cdef strnot
        strnot = self.__process_notation (notation)
        ranking = R._DifferentialRing__ranking ()
        L = bmi_c.bmi_dapyx_pardi (
                self.regchain, ranking, int (prime),
                self.notation, strnot, timeout, memout)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        C = RegularDifferentialChain.__new__ (RegularDifferentialChain)
        C.dring = R
        C.notation = strnot
        C.evdict = self.evdict
        C.regchain = L
        bmi_c.bmi_balsa_increment_nbref (C.regchain)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return C

# normal_form
    def normal_form (
            self,
            equations,
            bool casesplit = False,
            notation = BMI_IX_undefined):
        r"""
        The normal form of a rational differential fraction

        INPUT:

            ``equations`` -- a rational fraction or a list of rational
                             fractions
            ``casesplit`` -- (default: False) a bool
            ``notation``  -- (default: 'undefined') a string

        OUTPUT:

            The normal form of the equation w.r.t. the regular
            differential chain. If ``equations`` is a list, a
            list of normal form is returned. 

            Given a regular differential chain, not all rational
            fractions have normal forms: the denominator should not
            be zero, or even a zerodivisor, modulo the ideal defined
            by the regular differential chain. 

            The optional argument ``casesplit`` permit to handle
            the case of rational fractions which do not have normal
            forms, by splitting cases.

        NOTES:

            The argument ``casesplit`` = True is not yet implemented.

            The current implementation of normal forms may fail, even
            if the normal form does exist (very rare but possible situation).
            A complete algorithm is available but not yet implemented.
            See references below.

        REFERENCES:

            [BLS11] François Boulier, François Lemaire and Alexandre
            Sedoglavic. On the Regularity Property of Differential 
            Polynomials Modulo Regular Differential Chains. In proc.
            of CASC 2011. LNCS 6885, pp. 61-72. 2011.
            http://hal.archives-ouvertes.fr/hal-00599440

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,y = var ('x,y')
            >>> u,v = indexedbase ('u,v')
            >>> R = DifferentialRing (derivations = [x,y], blocks = [[v,u]], notation = 'jet')
            >>> eq1 = u[x]**2 - 4*u
            >>> eq2 = u[x,y]*v[y] - u + 1
            >>> eq3 = v[x,x] - u[x]
            >>> ideal = R.RosenfeldGroebner ([eq1,eq2,eq3])
            >>> ideal
            [regular_differential_chain]
            >>> C = ideal [0]  

        A few basic examples

            >>> p = u[y,y] / u[x]
            >>> C.normal_form (p)
            u[x]/(4*u)
            >>> C.normal_form (1/p)
            u[x]

        An expected behaviour of normal forms

            >>> C.normal_form (C.normal_form (p) * C.normal_form (1/p))
            1
        """
#
        cdef result, elt
        cdef bytes streqns, mesgerr
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        if casesplit:
            raise NotImplementedError ('casesplit = True')
        if isinstance (equations, list):
            streqns = ascii (equations)
        else:
            streqns = ascii ([equations])
        A = bmi_c.bmi_dapyx_normal_form (
                    self.regchain, streqns,
                    self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        result = [elt[0] for elt in result]
        if not isinstance (equations, list):
            result = result [0]
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# prolongation_prerequisites (
    def prolongation_prerequisites (
            self,
            q,
            point,
            edo = None,
            x = None,
            bool normal_form = False,
            notation = BMI_IX_undefined):
        r"""
        Prerequisite data for investigating a prolongation process

            These data, plus the positive integer roots of the polynomial A(q) 
            modulo the ideal defined by ``self`` permit to compute prolongation 
            bounds for investigating the existence and uniqueness of formal 
            power series solutions of ODE.

            The returned data as defined in [DL84, Lemma 2.3]

        INPUT:

            ``q``           -- a variable
            ``point``       -- a dictionary 
            ``edo``         -- (default: None) a differential polynomial
            ``x``           -- (default: None) an independent variable
            ``normal_form`` -- (default: False) a boolean
            ``notation``    -- (default: 'undefined') a string

        OUTPUT:

            Return a list of lists [u, n, k, r, c0, A(q)]. 
            There is one list per differential polynomial in ``self``. 
            The elements of the tuples are
            
            -- ``u``    -- the leader of the concerned differential 
                           polynomial f
            -- ``n``    -- the order of this leader
            -- ``k``    -- the valuation of the separant of f evaluated 
                           at point (Lemma 2.3, formula (1))
            -- ``r``    -- an integer less than or equal to k (Lemma 2.3 
                           formula (2))
            -- ``c0``   -- the coefficient of degree k in the separant 
                           evaluated at point (Lemma 2.3, formula (1))
            -- ``A(q)`` -- a polynomial (Lemma 2.3 formula (2))

            If ``edo`` is specified, it should be one the differential 
            polynomials of ``self``. 

            If ``x`` is specified it should provide the derivation to be 
            taken into account.

        REFERENCES:

            [DL84] Jan Denef and Leonard Lipshitz. Power Series Solutions of
            Algebraic Differential Equations. Mathematische Annalen vol. 267, 
            pp. 213–238. 1984

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x,q = var('x,q')
            >>> y = indexedbase('y')

        The variable q and the coefficients of the initial value
        encoding formal power series ybar

            >>> params = [q] + [y[i] for i in range (9, -1, -1)]
            >>> params
            [q, y[9], y[8], y[7], y[6], y[5], y[4], y[3], y[2], y[1], y[0]]

            >>> ybar = Add (*[y[i]*x**i/factorial(i) for i in range (0,5)])
            >>> ybar
            x**4*y[4]/24 + x**3*y[3]/6 + x**2*y[2]/2 + x*y[1] + y[0]

        The variable point provides the expansion point of the sought series
        and the relationship between the differential indeterminates and
        the initial value encoding series

            >>> point = { x:0, y:ybar }
            >>> point
            {x: 0, y: x**4*y[4]/24 + x**3*y[3]/6 + x**2*y[2]/2 + x*y[1] + y[0]}

            >>> R = DifferentialRing (derivations = [x], blocks = [y, params], parameters = params, notation = 'jet')

        The ODE under study

            >>> f = x*y*y[x,x] + y[x] + y**2 + 1

        The regular differential chain C contains both the ODE system
        and polynomial constraints on the coefficients of the sought
        formal power series

            >>> C = RegularDifferentialChain ([f, 5*y[0] + 1], R)

        The list L contains a single tuple since C contains a single ODE
        Its leader is y[x,x] which has order 2.
        Observe that the polynomial A(q) is not under normal form

            >>> L = C.prolongation_prerequisites (q, point)
            >>> L
            [[y[x, x], 2, 1, 1, y[0], q*y[0] + 4*y[0] + 1]]

        A variant of the former call.
        The polynomial A(q) is now under normal form
            >>> u, n, k, r, c0, A = C.prolongation_prerequisites (q, point, edo=f, normal_form=True)
            >>> u, n, k, r, c0, A
            (y[x, x], 2, 1, 1, -1/5, 1/5 - q/5)

        This polynomial vanishes at q=1
            >>> R.positive_integer_roots (A, q)
            [1]
        """
        cdef result
        cdef bytes mesgerr, str_notation, str_edo, str_q
        cdef bytes str_x, str_point
        cdef bmi_c.ALGEB_string L
        str_q = ascii (q)
        str_point = ascii (point)
        if edo == None:
            str_edo = ascii ("")
        else :
            str_edo = ascii (edo)
        if x == None :
            str_x = ascii ("")
        else :
            str_x = ascii (x)
        str_notation = self.__process_notation (notation)
        L = bmi_c.bmi_dapyx_DL_prolongation_prerequisites (
            str_edo, str_x, str_q, self.regchain,
            str_point, int (normal_form), self.notation, str_notation, 0, 0)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (L.value, globals (), self.evdict)
        if edo != None :
            result = result[0]
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result

# preparation_equation
    def preparation_equation (
            self,
            polynomial,
            BaseFieldExtension basefield = None,
            bool congruence = False,
            zstring = 'z%d',
            notation = BMI_IX_undefined):
        r"""
        The preparation equation of a polynomial 

        INPUT:

            ``polynomial`` -- a polynomial
            ``basefield``  -- (default: None) a BaseFieldExtension
            ``congruence`` -- (default: False) a bool
            ``zstring``    -- (default: 'z%d') a string
            ``notation``   -- (default: 'undefined') a string

        OUTPUT:

            Return a preparation equation [Kolchin73, chapter IV, section 13]
            of ``polynomial`` w.r.t. the regular differential chain.

            Preparation equations are an important tool of the Low Power
            Theorem [Ritt50, chapter III, Kolchin73, chapter IV, section 15].
            The Low Power Theorem is applied by the RosenfeldGroebner
            method, when there is a single equation to process and
            the optional parameter ``singsol`` = True.

            Let I the differential ideal defined by the chain, denote
            A[1], ..., A[r] the differential polynomials which constitute
            the chain. Introduce r new dependent variables z[i]. Each
            variable z[i] representing the differential polynomial A[i].

            The returned preparation equation is an expression having the
            form h*f = c[1]*t[1] + c[2]*t[2] + ... + c[n]*t[n], where f
            stands for ``polynomial``. The differential polynomial h is
            a power product of initials and separants of the A[i]. The
            coefficients c[i] are reduced w.r.t. the chain and regular 
            with respect to I. The monomials t[i] are power products of 
            the z[k] variables and their derivatives. They satisfy some
            further properties, described in [Kolchin73, chapter IV, 
            section 13]. If each z[k] is replaced by the corresponding
            polynomial A[k], then the preparation equation becomes an
            equality.
            
            The optional parameter ``congruence``. In the right hand-side 
            of the preparation equation, denote q the minimum total
            degree (the low power, actually) of the monomials t[i]. 
            If ``congruence`` is true, then all the terms c[i]*t[i] 
            such that the total degree of t[i] is strictly greater 
            than q are removed. The remaining sum is call a preparation 
            congruence of f.

            The optional parameter ``basefield`` is useful in conjunction
            with ``congruence`` = True. Reductions by the A[i] which
            belong to the base field are not taken into account for 
            computing the preparation congruence of f. The terms t[k]
            which depend on such z[i] and their derivatives are not
            considered for computing the degree q and do not appear
            in the preparation congruence.

            The optional parameter ``zstring`` permits to denote the
            polynomial A[i] by symbols different from z[i]. This
            may be useful since the method creates one indexedbase (in
            the Sage sense) for each z[i]. The ``zstring`` should
            correspond to a valid indexedbase identifier and involve 
            the substring '%d'.

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var('t')
            >>> u = indexedbase('u')
            >>> R = DifferentialRing (derivations = [t], blocks = [u], notation = 'jet')

        A famous example of Ritt. Two cases are considered.

            >>> poly = u[t]**2 - 4*u
            >>> ideal = R.RosenfeldGroebner ([poly])
            >>> [ C.equations (solved = True) for C in ideal ]
            [[Eq(u[t]**2, 4*u)], [Eq(u, 0)]]

        Is the solution u(t) = 0 an essential component ?
        
            >>> C = ideal [1]

        Here is a preparation equation for poly, w.r.t. the chain C. 
        Replacing z1 by u, it is easy to see check the result.

            >>> C.preparation_equation (poly)
            Eq(-4*u + u[t]**2, -4*z1 + z1[t]**2)

        Here is the preparation congruence (q = 1).

            >>> C.preparation_equation (poly, congruence = True)
            Eq(-4*u + u[t]**2, -4*z1)

        By [Kolchin73, chapter IV, section 15, theorem 6], the
        solution is, indeed, essential. This result can be interpreted
        as follows: the solution u(t) = 0 is not a particular case
        of the general solution u(t) = (t + c)**2, where c is an arbitrary y
        constant.

            >>> ideal = R.RosenfeldGroebner ([poly], singsol = 'essential')
            >>> [ C.equations (solved = True) for C in ideal ]
            [[Eq(u[t]**2, 4*u)], [Eq(u, 0)]]

        Another famous example. Two similar cases are generated.

            >>> poly = u[t]**2 - 4*u**3
            >>> ideal = R.RosenfeldGroebner ([poly])
            >>> ideal
            [regular_differential_chain, regular_differential_chain]
            >>> [ C.equations (solved = True) for C in ideal ]
            [[Eq(u[t]**2, 4*u**3)], [Eq(u, 0)]]

        Is the solution u(t) = 0 an essential component ?

            >>> C = ideal [1]

        Here is a preparation equation for poly, w.r.t. the chain C. 
        Replacing z1 by u, it is easy to see check the result.

            >>> C.preparation_equation (poly)
            Eq(-4*u**3 + u[t]**2, -4*z1**3 + z1[t]**2)

        Here is the preparation congruence (q = 2).

            >>> C.preparation_equation (poly, congruence = True)
            Eq(-4*u**3 + u[t]**2, z1[t]**2)

        By Ritt's theorem, the solution is particular. This result 
        can be interpreted as follows: the solution u(t) = 0 is
        the limit, when the arbitrary constant c tends towards infinity,
        of the general solution u(t) = 1/(t + c)**2.

            >>> ideal = R.RosenfeldGroebner ([poly], singsol = 'essential')
            >>> [ C.equations (solved = True) for C in ideal ]
            [[Eq(u[t]**2, 4*u**3)]]

        The following example just illustrates the use of base fields.
        The F is a differential field involving two indexedbases s(t), c(t),
        which can be viewed as sin(t) and cos(t).

            >>> t = var('t')
            >>> u,s,c = indexedbase ('u,s,c')
            >>> R = DifferentialRing (derivations = [t], blocks = [u,[s,c]], notation = 'jet')
            >>> C = RegularDifferentialChain ([c[t] + s, s**2 + c**2 - 1], R)
            >>> F = BaseFieldExtension (relations = C)

        Observe that, over F, poly and poly2 are the same polynomial.
        They thus should have the same preparation congruence.

            >>> poly = u[t]**2 - 4*u**3
            >>> poly2 = u[t]**2 - 4*u**3 + s**2 + c**2 - 1

        Two cases are generated

            >>> ideal = R.RosenfeldGroebner ([poly], basefield = F)
            >>> [ C.equations (solved = True) for C in ideal ]
            [[Eq(s**2, 1 - c**2), Eq(c[t], -s), Eq(u[t]**2, 4*u**3)], [Eq(s**2, 1 - c**2), Eq(c[t], -s), Eq(u, 0)]]

        Is the solution u(t) = 0 essential ?

            >>> C = ideal[1]
            >>> C.equations (solved = True)
            [Eq(s**2, 1 - c**2), Eq(c[t], -s), Eq(u, 0)]

        The preparation congruence of poly (playing with zstring).

            >>> C.preparation_equation (poly, zstring = 'C%d')
            Eq(-4*u**3 + u[t]**2, -4*C3**3 + C3[t]**2)

        The preparation congruence of poly shows that the solution
        is not essential (see the previous example).

            >>> C.preparation_equation (poly, congruence=True)
            Eq(-4*u**3 + u[t]**2, z3[t]**2)

        Let us now consider poly2, which is equal to poly, over F

            >>> C.preparation_equation (poly2)
            Eq(c**2 + s**2 - 4*u**3 + u[t]**2 - 1, z1 - 4*z3**3 + z3[t]**2)

        The term t[i] of minimal degree correspond to a reduction
        by a field equation.

            >>> C.preparation_equation (poly2, congruence=True)
            Eq(c**2 + s**2 - 4*u**3 + u[t]**2 - 1, z1)

        Specifying that we should consider poly2 over F, the preparation
        congruence changes and we conclude similarly as for poly.

            >>> C.preparation_equation (poly2, congruence=True, basefield=F)
            Eq(c**2 + s**2 - 4*u**3 + u[t]**2 - 1, z3[t]**2)

        We check below that RosenfeldGroebner recognizes the solution
        as not essential.

            >>> ideal = R.RosenfeldGroebner ([poly], singsol = 'essential', basefield = F)
            >>> [ C.equations (solved = True) for C in ideal ]
            [[Eq(s**2, 1 - c**2), Eq(c[t], -s), Eq(u[t]**2, 4*u**3)]]
        """
#
        cdef result, nbz, i, zevdict, zsymb, zval
        cdef bytes mesgerr, strpoly, strgens, strrels, newval, strzstring
        cdef bmi_c.ALGEB_string A
        cdef strnot
        strnot = self.__process_notation (notation)
        strpoly = ascii (polynomial)
        if basefield == None:
            F = BaseFieldExtension ()
        else:
            F = basefield
        strgens = ascii (F.generators ())
        strrels = ascii (F._BaseFieldExtension__relations ())
        strzstring = ascii (zstring)
        A = bmi_c.bmi_dapyx_preparation_equation (
                strpoly, self.regchain, strgens, strrels, int (congruence),
                strzstring, self.notation, strnot, 0, 0)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        nbz = self._RegularDifferentialChain__number_of_equations ()
        zevdict = self.evdict.copy ()
        for i in range (nbz):
            zsymb = zstring % (i+1)
            if self.notation == BMI_IX_Derivative :
                zval = sympy.Function (zsymb)
            else :
                zval = sympy.IndexedBase (zsymb)
            zevdict[zsymb] = zval
        newval = ascii (A.value)
        newval = b'Eq(' + newval.replace (b'=', b',') + b')'
        result = eval (newval, globals (), zevdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# __number_of_equations
    def __number_of_equations (self):
        r"""
        The number of equations of a regular differential chain

        INPUT:

            Nothing

        OUTPUT:

            The number of equations of a regular differential chain

        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> t = var('t')
            >>> u = indexedbase('u')
            >>> R = DifferentialRing (derivations = [t], blocks = [u], notation = 'jet')
            >>> poly = u[t]**2 - 4*u
            >>> ideal = R.RosenfeldGroebner ([poly])
            >>> ideal
            [regular_differential_chain, regular_differential_chain]
            >>> C = ideal [0]
            >>> C.equations ()
            [-4*u + u[t]**2]
        """
#
        cdef bmi_c.ALGEB_string A
        cdef bytes mesgerr
        cdef result
        A = bmi_c.bmi_dapyx_number_of_equations (self.regchain)
        if bmi_c.bmi_dapyx_is_error (A):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (A)
            bmi_c.bmi_balsa_clear_ALGEB (A)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = eval (A.value, globals (), self.evdict)
        bmi_c.bmi_balsa_clear_ALGEB (A)
        return result

# RosenfeldGroebner
    def RosenfeldGroebner (
            self,
            equations,
            BaseFieldExtension basefield = None,
            list attributes = [
                    'differential', 'autoreduced', 'primitive', 
                    'squarefree', 'normalized', 'coherent'],
            singsol = BMI_IX_all, 
            dimlb = BMI_IX_safecase,
            notation = BMI_IX_undefined,
            int timeout = 0, 
            int memout = 0):
        r"""
        A variant of the Rosenfeld-Groebner simplifier
            
        INPUT:
            
            ``equations``  -- a list of differential polynomials, differential 
                              polynomial equations or inequations
            ``basefield``  -- (default: None) a BaseFieldExtension
            ``attributes`` -- (default: ['differential', 'autoreduced', 
                              'primitive', 'squarefree', 'normalized', 
                              'coherent']) a list of strings
            ``singsol``    -- (default: 'all') a string
            ``dimlb``      -- (default: 'safecase') a string
            ``notation``   -- (default: 'undefined') a string
            ``timeout``    -- (default: 0) a nonnegative integer
            ``memout``     -- (default: 0) a nonnegative integer
            
        OUTPUT:
            
            Let A be the set of differential polynomials of ``self`` 
            and I the differential ideal that it defines. 
            Let S be the set of the initials of A, plus its separants
            if A is differential or squarefree.
            
            This variant enlarges ``equations`` with A = 0, S != 0
            before executing the classical RosenfeldGroebner method.
            The output ideals all contain I.
            
            See DifferentialRing.RosenfeldGroebner
            
        EXAMPLES:

            >>> from sympy import *
            >>> from DifferentialAlgebra import *
            >>> x = var('x')
            >>> y,z = indexedbase ('y,z')
            >>> R = DifferentialRing (derivations = [x], blocks = [z,y], notation = 'jet')

            >>> f = y[x]**2 - 4*y
            >>> g = z[x]**2 - y*z

            Let us start with a single regular differential chain made of f
            and extend it using g

            >>> C = RegularDifferentialChain ([f], R)
            >>> ideal = C.RosenfeldGroebner ([g])
            >>> ideal
            [regular_differential_chain, regular_differential_chain]
            >>> [ C.equations () for C in ideal ]
            [[y[x]**2 - 4*y, z[x]**2 - y*z], [y[x]**2 - 4*y, z]]

        The above output contains one less regular differential chain
        than the one of the following call to the classical method

            >>> ideal = R.RosenfeldGroebner ([f,g])
            >>> ideal
            [regular_differential_chain, regular_differential_chain, regular_differential_chain]
            >>> [ C.equations () for C in ideal ]
            [[y[x]**2 - 4*y, z[x]**2 - y*z], [y[x]**2 - 4*y, z], [y, z[x]]]
        """
        cdef list result, eqns, ineq
        cdef BaseFieldExtension F
        cdef RegularDifferentialChain C
        cdef bmi_c.ALGEB_list L
        cdef bytes mesgerr, streqns, strineq, strgens, strrels, strattr
        cdef bytes strsingsol, strdimlb
        cdef strnot
        strnot = self.__process_notation (notation)
        if isinstance (equations, list):
            eqns, ineq = self.dring._DifferentialRing__process_equations (equations, strnot)
        else:
            eqns, ineq = self.dring._DifferentialRing__process_equations ([equations], strnot)
        streqns = ascii (eqns)
        strineq = ascii (ineq)
        strattr = ascii (attributes)
        strattr = strattr.replace (b"'", b"")
        strsingsol = ascii (singsol)
        strdimlb = ascii (dimlb)
        if basefield == None:
            F = BaseFieldExtension ()
        else:
            F = basefield
        strgens = ascii (F.generators ())
        strrels = ascii (F._BaseFieldExtension__relations ())
        L = bmi_c.bmi_dapyx_RosenfeldGroebner (
                    streqns, strineq, strgens, strrels,
                    strattr, self.regchain, strsingsol, strdimlb, 1,
                    self.notation, strnot, timeout, memout)
        if bmi_c.bmi_dapyx_is_error (L):
            mesgerr = bmi_c.bmi_dapyx_mesgerr (L)
            bmi_c.bmi_balsa_clear_ALGEB (L)
            raise RuntimeError (str (mesgerr, 'utf-8'))
        result = []
        for i in range (1, L.value.size) :
            C = RegularDifferentialChain.__new__ (RegularDifferentialChain)
            C.dring = self.dring
            C.notation = strnot
            C.evdict = self.evdict
            C.regchain = L.value.tab [i]
            bmi_c.bmi_balsa_increment_nbref (C.regchain)
            result.append (C)
        bmi_c.bmi_balsa_clear_ALGEB (L)
        return result
