from sympy import *
from DifferentialAlgebra import *

class Hurwitz :
    r"""
    The class Hurwitz implements [Lemma 2.2, Denef and Lipshitz, 1984]
        which is a restatement of a result due to A. Hurwitz (1889)
        and [Lemma 2.3, Denef and Lipshitz, 1984]
    """
    def __init__ (self, R, P) :
        if isinstance (R, DifferentialRing) :
            self.R = R
            self.P = P
        else :
            self.R = P
            self.P = R
        self.x = self.R.indets (selection = 'derivations')[0]
        v = self.R.leading_derivative (self.P)
        theta, self.y = self.R.factor_derivative (v)
        self.n = degree (theta)
    def f (self, k) :
        dp = self.R.differentiate (self.P, self.x**(2*k+2))
        result = []
        for j in range (self.n, self.n + k + 1) :
            m = 2*self.n + 2*k + 2 - j
            v = self.R.differentiate (self.y, self.x**m)
            koeff = dp.coeff (v, 1)
            dp = expand (dp - koeff*v)
            eqn = Eq (indexedbase ('f') [ var('n') + j - self.n ], koeff)
            result.append (eqn)
        eqn = Eq (indexedbase ('f') [ var('n') + k + 1 ], dp)
        result.append (eqn)
        return result
    def k (self, evdict) :
        r"""
        Return the valuation k of [Lemma 2.2, Denef and Lipshitz, 1984]
        """
        fn = self.R.separant (self.P)
        evaluated_fn = collect (expand (self.R.evaluate (fn, evdict).doit ()), self.x)
        d = degree (evaluated_fn, self.x)
        for e in range (d-1, -1, -1) :
            if evaluated_fn.coeff (self.x, e) != 0 :
                d = e
        return d
    def binomial_coefficients (self, k, q) :
        r"""
        Return the list of the coefficients enclosed between square brackets
        of the Formula (3) of [Lemma 2.2, Lemma 2.2, Denef and Lipshitz, 1984]

        These coefficients are polynomials in q and provide a generic
        formula for P^(2*k + 2 + q)

        With the notations of the paper, return

        [ f[n], 
            f[n+1] + q*f[n]', 
            ...,
            f[n+k] + q*f[n+k-1]' + ... + binomial(q,k)*f[n]^(k)]

        The coefficients are given in symbolic binomial form. they can
        be expanded using the sympy/combsimp method
        """
        L = self.f (k)
        result = []
        for r in range (0, k+1) :
            expr = [ self.R.differentiate (L[j].rhs, self.x**(r-j)) * binomial (q, r-j) for j in range (0, r+1) ]
            expr = Add (*expr)
            result.append (expr)
        return result
    def kra (self, q, evdict) :
        r"""
        Return 

        k, r, A

        where k and r (r <= k) are the integers defined in [Lemma 2.3, 
        Denef and Lipshitz 1984] and A is the polynomial in x and q.
        All these quantities depend on an initial values encoding 
        series ybar which is passed through evdict

        f[n+r] + q*f[n+r-1]' + ... + binomial(q,r)*f[n]^(r)

        which occurs on the left hand side of Formula (2) of Lemma 2.3.

        This polynomial is given in binomial form but can be expanded
        using the sympy/combsimp method
        """
        k = self.k (evdict)
        L = self.binomial_coefficients (k, q)
        M = self.R.evaluate (L, evdict)
        M = [ expand (poly.doit ()).coeff (self.x, 0) for poly in M ]
        r = 0
        while M[r] == 0 :
            r += 1
        return k, r, M[r]
    def beta (self, gamma, evdict) :
        r"""
        Return 2*k + 2 + gamma + r 

        which is the exponent of Formula (3) of [Lemma 2.3, 
        Denef and Lipshitz, 1984]. The integer gamma is supposed
        to be bigger than any positive integer root of polynomial A.

        See the help page of the kra method
        """
        k, r, A = self.kra (var('q'), evdict)
        return 2*k + 2 + gamma + r
    def delta (self, gamma, evdict) :
        r"""
        Return n + 2*k + 2 + gamma

        which is the exponent of Formula (4) of [Lemma 2.3, 
        Denef and Lipshitz, 1984]. The integer gamma is supposed
        to be bigger than any positive integer root of polynomial A

        See the help page of the kra method
        """
        k, r, A = self.kra (var('q'), evdict)
        return self.n + 2*k + 2 + gamma

